/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1020 $
 * $Date: 2020-05-19 16:31:39 +0200 (Tue, 19 May 2020) $
 */
package at.lcm.bu21.general.dtypes;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Hashtable;
import java.util.Map;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.CloneException;

/**
 * Base type of different data representation classes
 * 
 * @author Werner Koppelsttter
 */
public abstract class TData implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * returns current class name. Used during conversion to python objects
	 * 
	 * @return class name
	 */
	public String getClassName() {
		return this.getClass().getSimpleName();
	}

	/**
	 * Read data-structure from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @return data representation of xml-structure
	 * @throws TDataException
	 *             Unable to create data-objects
	 * @throws XMLStreamException
	 *             Exception by reading XML
	 */
	public static TData xml2obj(XMLStreamReader xmlReader) throws TDataException, XMLStreamException {
		String element;
		TData obj;

		element = xmlReader.getLocalName();

		if (element.equals("Null")) {
			obj = new TNull();
		} else if (element.equals("Struct")) {
			obj = new TStruct();
		} else if (element.equals("Array")) {
			obj = new TArray();
		} else if (element.equals("Boolean")) {
			obj = new TBoolean();
		} else if (element.equals("String")) {
			obj = new TString();
		} else if (element.equals("Numeric")) {
			obj = new TNumeric();
		} else if (element.equals("Periodic")) {
			obj = new TPeriodic();
		} else if (element.equals("PeriodicMatrix")) {
			obj = new TPeriodicMatrix();
		} else {
			throw new TDataException("unrecognized Type of data! " + element);
		}
		obj.xml2data(xmlReader);

		return obj;
	}

	/**
	 * Get value of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @return Value of data
	 */
	public TData getValue(String identifier) {
		if ((identifier == null) || (identifier.length() == 0)) {
			identifier = "value__";
		} else {
			identifier += ".value__";
		}

		return this.getData(identifier);
	}

	/**
	 * Get value of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param unit
	 *            Unit of the wanted value
	 * @return Value of data
	 * @throws TDataException
	 *             Unable to get value for the unit
	 */
	public TData getValue(String identifier, TString unit) throws TDataException {
		TData data, value = null;
		TString unit1;

		data = this.getData(identifier);
		if (data != null) {
			unit1 = data.getUnit(null);
			data = data.getValue(null);
			if ((data != null) && (unit1 != null)) {
				if (!unit1.equals(unit)) {
					throw new TDataException("Unit is not equal! (" + identifier + ")");
				}
				value = data;
			}
		}

		return value;
	}

	/**
	 * Get unit of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @return Unit of data
	 */
	public TString getUnit(String identifier) {
		TString unit = null;
		TData data;

		if ((identifier == null) || (identifier.length() == 0)) {
			identifier = "unit__";
		} else {
			identifier += ".unit__";
		}

		data = this.getData(identifier);
		if ((data != null) && (data instanceof TString)) {
			unit = (TString) data;
		}

		return unit;
	}

	/**
	 * Get selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @return Data
	 */
	public TData getData(String identifier) {
		TData data = null;

		if ((identifier == null) || (identifier.length() == 0)) {
			data = this;
		} else {
			data = this.getData(identifier.split("[.\\[]"));
		}

		return data;
	}

	/**
	 * Get selected data
	 * 
	 * @param ident
	 *            List with identifier of data to select
	 * @return Data
	 */
	protected TData getData(String[] ident) {
		TData data = null;

		if (ident.length == 0) {
			data = this;
		}

		return data;
	}

	/**
	 * Read a map of TData
	 * 
	 * @param is
	 *            InputStream
	 * @return Map of TData
	 * @throws TDataException
	 *             Unable to create data-objects
	 * @throws XMLStreamException
	 *             Exception by reading from XML
	 * @throws IOException
	 *             Exception by file IO
	 */
	public static Map<String, TData> readMap(InputStream is) throws TDataException, XMLStreamException, IOException {
		Map<String, TData> map;

		map = new Hashtable<String, TData>();
		TData.readMap(is, map);

		return map;
	}

	/**
	 * Read a map of TData
	 * 
	 * @param is
	 *            InputStream
	 * @param map
	 *            Map to append
	 * @throws TDataException
	 *             Unable to create data-objects
	 * @throws XMLStreamException
	 *             Exception by reading from XML
	 * @throws IOException
	 *             Exception by file IO
	 */
	public static void readMap(InputStream is, Map<String, TData> map)
			throws TDataException, XMLStreamException, IOException {
		TData tmp;
		TStruct struct;

		tmp = TData.read(is);
		if (tmp instanceof TStruct) {
			struct = (TStruct) tmp;
			for (String key : struct.getKeys()) {
				map.put(key, struct.get(key));
			}
		} else {
			throw new TDataException("Unable to convert to map");
		}
	}

	/**
	 * Write a map of TData
	 * 
	 * @param os
	 *            OutputStream
	 * @param map
	 *            Map with TData
	 * @throws XMLStreamException
	 *             Exception by writing to XML
	 */
	public static void writeMap(OutputStream os, Map<String, TData> map) throws XMLStreamException {
		TData.writeMap(os, map, true);
	}

	/**
	 * Write a map of TData
	 * 
	 * @param os
	 *            OutputStream
	 * @param map
	 *            Map with TData
	 * @param header
	 *            Add header with version and encoding
	 * @throws XMLStreamException
	 *             Exception by writing to XML
	 */
	public static void writeMap(OutputStream os, Map<String, TData> map, boolean header) throws XMLStreamException {
		TStruct struct;
		struct = new TStruct();

		for (Map.Entry<String, TData> entry : map.entrySet()) {
			struct.put(entry.getKey(), entry.getValue());
		}
		struct.write(os, header);
	}

	/**
	 * Read object from InputStream
	 * 
	 * @param is
	 *            InputStream
	 * @return Read object
	 * @throws TDataException
	 *             Unable to create data-objects
	 * @throws XMLStreamException
	 *             Exception by reading from XML
	 * @throws IOException
	 *             Exception by file IO
	 */
	public static TData read(InputStream is) throws TDataException, XMLStreamException, IOException {
		TData data;

		// Open xml-reader
		XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
		XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is, "utf-8");

		// read file
		while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
		}

		data = TData.xml2obj(xmlReader);

		if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
			throw new IOException("Undefined error by reading file!");
		}

		// end file
		xmlReader.close();

		return data;
	}

	/**
	 * Write object to OutputStream
	 * 
	 * @param os
	 *            OutputStream
	 * @throws XMLStreamException
	 *             Exception by writing to XML
	 */
	public void write(OutputStream os) throws XMLStreamException {
		this.write(os, true);
	}

	/**
	 * Write object to OutputStream
	 * 
	 * @param os
	 *            OutputStream
	 * @param header
	 *            Add header with version and encoding
	 * @throws XMLStreamException
	 *             Exception by writing to XML
	 */
	public void write(OutputStream os, boolean header) throws XMLStreamException {
		XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance().createXMLStreamWriter(os, "utf-8");
		if (header) {
			xmlWriter.writeStartDocument("utf-8", "1.0");
			// xmlWriter.writeComment(Project.getVersion());
		}

		// write Project to File
		this.obj2xml(xmlWriter);

		xmlWriter.writeEndDocument();
		xmlWriter.close();
	}

	/**
	 * Create a deep clone of given data-structure
	 * 
	 * @return Deep clone of structure
	 * @throws CloneException
	 *             Unable to clone object (serialize)
	 */
	public abstract TData deepClone() throws CloneException;

	/**
	 * Write data-structure to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws XMLStreamException
	 *             Exception by writing to XML
	 */
	public void obj2xml(XMLStreamWriter xmlWriter) throws XMLStreamException {
		this.data2xml(xmlWriter);
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws XMLStreamException
	 *             Exception by reading from XML
	 * @throws TDataException
	 *             Unable to create data-objects
	 */
	protected abstract void xml2data(XMLStreamReader xmlReader) throws XMLStreamException, TDataException;

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws XMLStreamException
	 *             Exception by writing to XML
	 */
	protected abstract void data2xml(XMLStreamWriter xmlWriter) throws XMLStreamException;

	/**
	 * Create TData from string
	 * 
	 * @param value
	 *            String value
	 * @return TData object
	 * @throws TDataException
	 *             Unable to create data-objects
	 */
	public static TData createData(String value) throws TDataException {
		TData data = null;

		value = value.trim();
		if (value.isEmpty()) {
			data = new TNull();
		} else if (value.startsWith("'") && value.endsWith("'")) {
			data = new TString(value.substring(1, value.length() - 1));
		} else if (value.startsWith("[") && value.endsWith("]")) {
			if (value.substring(1, value.length() - 1).trim().isEmpty()) {
				data = new TNull();
			} else {
				data = new TNumeric(value);
			}
		} else {
			try {
				data = new TBoolean(value);
			} catch (Exception e1) {
				try {
					data = new TNumeric(value);
				} catch (Exception e2) {
					data = new TString(value);
				}
			}
		}

		return data;
	}
}
