/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2.TDataModel;

import java.util.ArrayList;
import java.util.List;

import javax.swing.JFrame;

import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;
import at.lcm.bu21.general.plot.Plot2Line;
import at.lcm.bu21.general.plot.Plot2XY;
import at.lcm.bu21.general.plot.PlotException;

/**
 * Representation of a SingalBuilderModel - type
 * 
 * @author Johannes Kaisinger
 *
 */
public class TPeriodicSignalBuilderModel {

	/**
	 * List of possible signals
	 * 
	 * @author Johannes Kaisinger
	 *
	 */
	public enum Signal {
		/**
		 * Rectangle signal
		 */
		RECT,
		/**
		 * Sinus signal
		 */
		SIN,
		/**
		 * Triangle signal
		 */
		TRI
	}

	/**
	 * Representation of a TPeriodicSignal in the signal builder.
	 * 
	 * @author Johannes Kaisinger
	 *
	 */
	public class TPeriodicSignal {

		/** Signal type */
		private Signal sig;

		/** offset */
		private double offs;

		/** amplitude */
		private double ampl;

		/** phase */
		private double phas;

		/** order */
		private int order;

		/** phi - ratio */
		private double phi;

		/**
		 * Create a Signal
		 * 
		 * @param sig
		 *            Signal type
		 * @param order
		 *            Maximum order which is used
		 */
		public TPeriodicSignal(Signal sig, int order) {
			this.sig = sig;
			this.offs = 0;
			this.ampl = 1.0;
			this.phas = 0;
			this.order = order;
			this.phi = Math.PI;
		}

		/**
		 * Create a TPeriodic type out of a TPeriodicSignal type
		 * 
		 * @return TPeriodic
		 */
		public TPeriodic buildSignal() {
			TPeriodic tPeriodic = null;

			switch (this.sig) {
			case RECT:
				tPeriodic = this.buildRect();
				break;
			case SIN:
				tPeriodic = this.buildSin();
				break;
			case TRI:
				tPeriodic = this.buildTri();
				break;
			}

			return tPeriodic;
		}

		/**
		 * Creates a TPeriodic (Rectangle)
		 * 
		 * @return TPeriodic
		 */
		private TPeriodic buildRect() {

			TPeriodic rect = new TPeriodic();

			double phi = this.phi;
			double phas = this.phas;

			phi /= 2;
			phas -= phi;

			final double[] coeffsA = new double[this.order];
			final double[] coeffsB = new double[this.order];

			final double mean = (this.offs + (((4 / Math.PI) * phi) / 2)) - 1.0;

			for (int i = 0; i < this.order; i++) {

				coeffsA[i] = ((4 / Math.PI) * Math.sin((i + 1) * phi)) / (i + 1);
				coeffsB[i] = 0;

			}

			rect.setCoeffs(mean, coeffsA, coeffsB);
			rect = rect.scale(this.ampl);
			rect = rect.shift(phas);

			return rect;
		}

		/**
		 * Creates a TPeriodic (Sin)
		 * 
		 * @return TPeriodic
		 */
		private TPeriodic buildSin() {
			TPeriodic sinus = new TPeriodic(this.order, this.ampl, (this.phas - (Math.PI / 2)));
			sinus = sinus.offset(this.offs);
			return sinus;
		}

		/**
		 * Creates a TPeriodic (Triangle)
		 * 
		 * @return TPeriodic
		 */
		private TPeriodic buildTri() {
			TPeriodic triangle = new TPeriodic();

			double phi = this.phi;
			double phas = this.phas;

			phi /= 2;
			phas -= phi;

			final double[] coeffsA = new double[this.order];
			final double[] coeffsB = new double[this.order];

			final double mean = (phi / (2 * Math.PI)) + this.offs;
			for (int i = 0; i < this.order; i++) {
				coeffsA[i] = ((1 - Math.cos(phi * (i + 1))) / ((i + 1) * (i + 1))) * (2 / (Math.PI * phi));
				coeffsB[i] = 0.0;
			}

			triangle.setCoeffs(mean, coeffsA, coeffsB);
			triangle = triangle.scale(this.ampl);
			triangle = triangle.shift(phas);

			return triangle;
		}
	}

	/** List of TPeriodicSignal defined in the signal builder */
	private List<TPeriodicSignal> signals;

	/** Data variable */
	private TPeriodic tPeriodic;

	/** Maximum of order which is possible to use */
	private int maxOrder;

	/**
	 * Create a TPeriodicSignalBuilderModel
	 */
	public TPeriodicSignalBuilderModel() {

		this.tPeriodic = null;
		this.maxOrder = 20;

		this.signals = new ArrayList<TPeriodicSignal>();

		this.signals.add(new TPeriodicSignal(Signal.RECT, this.maxOrder));
	}

	/**
	 * Get the signal tpye
	 * 
	 * @param row
	 *            Index of the row
	 * @return signal
	 */
	public Signal getSignal(int row) {
		return this.signals.get(row).sig;
	}

	/**
	 * Get the Offset of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @return offset
	 */
	public double getOffs(int row) {
		return this.signals.get(row).offs;
	}

	/**
	 * Get the amplitude of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @return amplitude
	 */
	public double getAmpl(int row) {
		return this.signals.get(row).ampl;
	}

	/**
	 * Get the phase of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @return phase
	 */
	public double getPhas(int row) {
		return this.signals.get(row).phas;
	}

	/**
	 * Get the order of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @return order
	 */
	public int getOrder(int row) {
		return this.signals.get(row).order;
	}

	/**
	 * Get the phi-value of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @return phi
	 */
	public double getPhi(int row) {
		return this.signals.get(row).phi;
	}

	/**
	 * Get the maximum order
	 * 
	 * @return maxOrder
	 */
	public int getMaxOrder() {
		return this.maxOrder;
	}

	/**
	 * Get the number of signals
	 * 
	 * @return length
	 */
	public int getLength() {
		return this.signals.size();
	}

	/**
	 * Set the signal type
	 * 
	 * @param row
	 *            Index of the row
	 * @param sig
	 *            signal type
	 */
	public void setSignal(int row, Signal sig) {
		this.signals.get(row).sig = sig;
	}

	/**
	 * Set the offset of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @param offs
	 *            New value for the offset
	 */
	public void setOffs(int row, double offs) {
		this.signals.get(row).offs = offs;
	}

	/**
	 * Set the amplitude of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @param ampl
	 *            New value for the amplitude
	 */
	public void setAmpl(int row, double ampl) {
		this.signals.get(row).ampl = ampl;
	}

	/**
	 * Set the phase of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @param phas
	 *            New value for the phase
	 */
	public void setPhas(int row, double phas) {
		this.signals.get(row).phas = phas;
	}

	/**
	 * Set the order of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @param order
	 *            New value for the order
	 */
	public void setOrder(int row, int order) {
		if (order <= this.maxOrder) {
			this.signals.get(row).order = order;
		} else {
			this.signals.get(row).order = this.maxOrder;
		}
	}

	/**
	 * Set phi ratio of the signal
	 * 
	 * @param row
	 *            Index of the row
	 * @param phi
	 *            New value for the phi ratio
	 */
	public void setPhi(int row, double phi) {
		this.signals.get(row).phi = phi;
	}

	/**
	 * Set max order of all signals
	 * 
	 * @param maxOrder
	 *            New value for the max order
	 */
	public void setMaxOrder(int maxOrder) {
		this.maxOrder = maxOrder;
	}

	/**
	 * Insert a new signal
	 * 
	 * @param row
	 *            Index of the row
	 * @param sig
	 *            Type of the new signal
	 */
	public void insertSignal(int row, Signal sig) {
		this.signals.add(row, (new TPeriodicSignal(sig, this.maxOrder)));
	}

	/**
	 * Remove a signal
	 * 
	 * @param row
	 *            Index of the row
	 */
	public void removeSignal(int row) {
		this.signals.remove(row);
	}

	/**
	 * Show a preview of the signal (Time)
	 * 
	 * @param row
	 *            Index of the row
	 */
	public void showTime(int row) {
		this.tPeriodic = this.signals.get(row).buildSignal();
		try {
			this.tPeriodic.showPreview();
		} catch (Exception e) {
		}
	}

	/**
	 * Show a preview of the signal (Spectrum)
	 * 
	 * @param row
	 *            Index of the row
	 */
	public void showSpectrum(int row) {
		this.tPeriodic = this.signals.get(row).buildSignal();

		try {
			this.tPeriodic.showSpectrum(null);
		} catch (Exception e) {
		}
	}

	/**
	 * Show a preview of all signals (Time)
	 */
	public void showTimeAll() {
		JFrame overview = new JFrame("All Signals");
		Plot2XY overviewPlot = new Plot2Line();

		overviewPlot.setGrid(true);
		overviewPlot.setLegend(true);

		TPeriodic tPeriodic = new TPeriodic();

		int nmbr = 0;
		if (this.getLength() > -1) {
			for (TPeriodicSignal sig : this.signals) {
				tPeriodic = sig.buildSignal();
				try {
					overviewPlot.addSignal("" + nmbr, tPeriodic, 500);
				} catch (PlotException e) {
				}
				nmbr++;
			}
		}

		overview.add(overviewPlot.getPanel());
		overview.setSize(800, 600);
		overview.setVisible(true);
		overview.setIconImage(IconEditor.ICON_XYPLOT.getImage());

	}

	/**
	 * Returns a TPeriodicModel for all defined signals
	 * 
	 * @return TPeriodicModel
	 */
	public TPeriodicModel getAllSignals() {
		TPeriodic val = new TPeriodic();
		TPeriodicModel tPeriodicModel = null;

		if (this.getLength() > -1) {
			for (TPeriodicSignal sig : this.signals) {
				val = val.add(sig.buildSignal());

				// TODO: !!!! It seems to be necessary. Otherwise too many
				// orders! (BUG).
				val.getMaxOrder();
			}
		}

		tPeriodicModel = (TPeriodicModel) TDataModel.createTDataModel(val);

		return tPeriodicModel;
	}

}
