/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1022 $
 * $Date: 2020-06-03 14:13:51 +0200 (Wed, 03 Jun 2020) $
 */
package at.lcm.bu21.general.dtypes.unit;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

/**
 * Node of syntax-tree to evaluate formulas
 *
 * @author Werner Koppelsttter
 */
public class SyntaxNode {

	/** Token of node */
	Token tok;

	/** Children tokens of node */
	SyntaxNode[] childs = null;

	/**
	 * Create node for token
	 * 
	 * @param tok
	 *            Token to represent
	 */
	SyntaxNode(Token tok) {
		this.tok = tok;
	}

	/**
	 * Add child to node
	 * 
	 * @param sn
	 *            Add child to list
	 */
	void addChild(SyntaxNode sn) {
		if (this.childs == null) {
			this.childs = new SyntaxNode[1];
		} else {
			this.childs = Arrays.copyOf(this.childs, this.childs.length + 1);
		}
		this.childs[this.childs.length - 1] = sn;
	}

	/**
	 * Get raw list of units, without modification.
	 * 
	 * @param baseUnit
	 *            Is base-unit. Don't check prefix and not assign base-unit
	 * @return List of units
	 * @throws EvalException
	 * @throws SyntaxException
	 */
	List<Unit> getRawUnits(boolean baseUnit) throws EvalException, SyntaxException {
		List<Unit> units;

		units = new ArrayList<Unit>();

		this.linearize(units, 1, baseUnit);
		return units;
	}

	/**
	 * Get list of units (in exponential-form)
	 * 
	 * @param baseUnit
	 *            Is base-unit. Don't check prefix and not assign base-unit
	 * @return List of units
	 * @throws EvalException
	 * @throws SyntaxException
	 */
	List<Unit> getUnits(boolean baseUnit) throws EvalException, SyntaxException {
		List<Unit> units;
		Unit u;

		units = this.getRawUnits(baseUnit);
		this.expand(units);
		this.compress(units);
		this.toSI(units);
		this.expand(units);
		this.compress(units);

		if (units.size() == 1) {
			u = units.get(0);
			if ((u.exponent == 1) && (u.baseUnit != null) && (u.baseUnit.siUnits != null)
					&& (u.baseUnit.siUnits.size() == 1)) {
				u.offset = u.baseUnit.offset;
				u.gain = u.baseUnit.gain;
				u.unitString = u.baseUnit.siUnits.get(0).unitString;
			}
		}

		return units;

	}

	/**
	 * Get list of units (in exponential-form)
	 * 
	 * @return List of units
	 * @throws EvalException
	 * @throws SyntaxException
	 */
	public List<Unit> getUnits() throws EvalException, SyntaxException {
		return this.getUnits(false);
	}

	/**
	 * Linearlize syntax-node to list of units
	 * 
	 * @param units
	 *            List of units
	 * @param exponent
	 *            Exponent from parent-node
	 * @param baseUnit
	 *            Is base-unit
	 * @throws EvalException
	 * @throws SyntaxException
	 */
	private void linearize(List<Unit> units, int exponent, boolean baseUnit) throws EvalException, SyntaxException {

		switch (this.tok.kind) {
		case APPEND:
			this.childs[0].linearize(units, exponent, baseUnit);
			this.childs[1].linearize(units, exponent, baseUnit);
			break;
		case SLASH:
			this.childs[0].linearize(units, exponent, baseUnit);
			this.childs[1].linearize(units, -exponent, baseUnit);
			break;
		case EXPONENT:
			this.childs[0].linearize(units, exponent * this.childs[1].tok.val, baseUnit);
			break;
		case ELEMENT:
			units.add(new Unit(this.tok.str, exponent, baseUnit));
			break;
		case NUMBER:
			if (this.tok.val == 1) {
				units.add(new Unit("1", 1));
			} else {
				throw new EvalException("Only 1 is allowed", this.tok);
			}
			break;
		case QUESTION:
			units.add(new Unit("?", 1));
			break;
		default:
			throw new EvalException("Undefined token", this.tok);
		}
	}

	/**
	 * Expand list of units. Split unit and gains. Calculate gains with
	 * exponents.
	 * 
	 * @param units
	 *            List of units
	 * @throws SyntaxException
	 */
	private void expand(List<Unit> units) throws SyntaxException {
		int len;
		Unit u, u1;

		len = units.size();
		for (int k = 0; k < len; k++) {
			u = units.get(k);
			if (u.unitString.equals("1")) {
				if (u.offset == 0.0) {
					u.gain = Math.pow(u.gain, u.exponent);
					u.exponent = 1;
				}
			} else {
				if ((u.offset == 0.0) && (u.gain != 1.0)) {
					u1 = new Unit("1", 1);
					u1.gain = Math.pow(u.gain, u.exponent);
					units.add(u1);
					u.gain = 1.0;
				}
			}
		}
	}

	/**
	 * Compress unit-list. Reduce equal units and no unit factors.
	 * 
	 * @param units
	 *            List of units
	 */
	private void compress(List<Unit> units) {
		Unit u1, u2;
		int k = 0, inc;

		// sort list
		units.sort(new Comparator<Unit>() {

			@Override
			public int compare(Unit o1, Unit o2) {
				int comp;

				comp = o1.unitString.compareTo(o2.unitString);
				if (comp == 0) {
					comp = o2.exponent - o1.exponent;
					if (comp == 0) {
						comp = (int) Math.signum(o2.offset - o1.offset);
						if (comp == 0) {
							comp = (int) Math.signum(o2.gain - o1.gain);
						}
					}
				}
				return comp;
			}
		});

		// compress
		while (k < (units.size() - 1)) {
			inc = 0;
			u1 = units.get(k);
			u2 = units.get(k + 1);
			if (u1.unitString.equals("1")) {
				if (u2.unitString.equals("1") && (u1.offset == 0.0) && (u2.offset == 0.0)) {
					u1.gain *= u2.gain;
					units.remove(k + 1);
				} else {
					inc = 1;
				}
				if ((u1.gain == 1.0) && (u1.offset == 0.0)) {
					units.remove(k);
					inc = 0;
				}
			} else if (u1.unitString.equals(u2.unitString) && (u1.offset == 0.0) && (u2.offset == 0.0)) {
				u1.exponent += u2.exponent;
				units.remove(k + 1);
			} else {
				inc = 1;
			}

			if ((u1.exponent == 0) && (u1.gain == 1.0) && (u1.offset == 0.0)) {
				units.remove(k);
			} else {
				k += inc;
			}

		}
	}

	/**
	 * Replace units with SI-units
	 * 
	 * @param units
	 *            List of units
	 * @throws SyntaxException
	 */
	private void toSI(List<Unit> units) throws SyntaxException {
		int len;
		Unit u;

		len = units.size();
		for (int k = 0; k < len; k++) {
			u = units.get(k);
			if ((u.baseUnit != null) && (u.baseUnit.siUnits != null) && (u.baseUnit.offset == 0.0)) {
				for (Unit u1 : u.baseUnit.siUnits) {
					units.add(new Unit(u1.unitString, u1.exponent * u.exponent));
				}
				u.gain *= u.baseUnit.gain;
				u.unitString = "1";
				u.baseUnit = null;
			}
		}

	}

	@Override
	public String toString() {
		StringBuffer buff;

		buff = new StringBuffer();

		switch (this.tok.kind) {
		case ELEMENT:
			buff.append(this.tok.str);
			break;
		case APPEND:
			buff.append(this.childs[0].toString());
			buff.append(" * ");
			buff.append(this.childs[1].toString());
			break;
		case SLASH:
			buff.append(this.childs[0].toString());
			buff.append(" / ");
			buff.append(this.childs[1].toString());
			break;
		case EXPONENT:
			buff.append("[");
			buff.append(this.childs[0].toString());
			buff.append("]^");
			buff.append(this.childs[1].toString());
			break;
		case NUMBER:
			buff.append(this.tok.val);
			break;
		default:
			break;
		}

		return buff.toString();
	}

}
