/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.math.interpolator2d;

import java.util.ArrayList;

import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.BlockRealMatrix;
import org.apache.commons.math3.linear.LUDecomposition;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;

import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.math.delaunay.Point;
import at.lcm.bu21.general.math.delaunay.Triangle;
import at.lcm.bu21.general.math.delaunay.Triangulate;

/**
 * @author gweidenholzer
 * 
 */
public class DelaunayInt2d extends Interpolator2d {

	ArrayList<Triangle> triangles = new ArrayList<Triangle>();
	ArrayList<Point> points = new ArrayList<Point>();

	private ConvexHull convexhull;

	/**
	 * @param xdata
	 * @param ydata
	 * @param zdata
	 * @throws Exception
	 */
	public DelaunayInt2d(double[] xdata, double[] ydata, double[] zdata) throws Exception {
		super(xdata, ydata, zdata);

		this.convexhull = new ConvexHull(this.xnorm, this.ynorm);

		this.rel_tol = 0.0;

		for (int k = 0; k < this.xnorm.length; k++) {
			this.points.add(new Point(this.xnorm[k], this.ynorm[k], this.zdata[k]));
		}

		this.triangles = Triangulate.triangulate(this.points);

	}

	/**
	 * @param xdata
	 * @param ydata
	 * @param zdata
	 * @throws Exception
	 */
	public DelaunayInt2d(TNumeric xdata, TNumeric ydata, TNumeric zdata) throws Exception {
		this(Interpolator2d.dataVectors(xdata, ydata, zdata));
	}

	/**
	 * xyz[0]...xdata xyz[1]...ydata xyz[2]...zdata
	 * 
	 * @param xyz
	 * @throws Exception
	 */
	protected DelaunayInt2d(double[][] xyz) throws Exception {
		this(xyz[0], xyz[1], xyz[2]);
	}

	@Override
	double value_(double xn, double yn) {

		// find corresponding facet
		Triangle t = this.find(xn, yn).t;

		BlockRealMatrix a = new BlockRealMatrix(
				new double[][] { { t.b.x - t.a.x, t.c.x - t.a.x }, { t.b.y - t.a.y, t.c.y - t.a.y } });

		BlockRealMatrix b = new BlockRealMatrix(new double[][] { { xn - t.a.x }, { yn - t.a.y } });
		RealMatrix rs = new LUDecomposition(a).getSolver().solve(b);

		BlockRealMatrix tmp = new BlockRealMatrix(new double[][] { { t.b.val - t.a.val, t.c.val - t.a.val } });

		return t.a.val + tmp.multiply(rs).getEntry(0, 0);

	}

	private DistPair find(double xn, double yn) {

		int n = this.triangles.size();
		Point p = new Point((float) xn, (float) yn, 0.0f);

		DistPair best = null;

		for (int k = 0; k < n; k++) {
			DistPair dp = new DistPair(this.triangles.get(k), p);
			if (dp.compareTo(best) < 0) {
				best = dp;
				if (best.d == 0.0) {
					break;
				}

			}
		}

		return best;
	}

	private double getDerivative(double x, double y, int d) {

		Triangle t = this.find(x, y).t;

		BlockRealMatrix a = new BlockRealMatrix(3, 3);
		a.setEntry(0, 0, t.a.x);
		a.setEntry(0, 1, t.a.y);
		a.setEntry(0, 2, 1.0);

		a.setEntry(1, 0, t.b.x);
		a.setEntry(1, 1, t.b.y);
		a.setEntry(1, 2, 1.0);

		a.setEntry(2, 0, t.c.x);
		a.setEntry(2, 1, t.c.y);
		a.setEntry(2, 2, 1.0);

		ArrayRealVector b = new ArrayRealVector(3);
		b.setEntry(0, t.a.val);
		b.setEntry(1, t.b.val);
		b.setEntry(2, t.c.val);

		RealVector res = new LUDecomposition(a).getSolver().solve(b);
		return res.getEntry(d);
	}

	@Override
	double dz_dx_(double x, double y) {
		return this.getDerivative(x, y, 0);
	}

	@Override
	double dz_dy_(double x, double y) {
		return this.getDerivative(x, y, 1);
	}

	@Override
	double dz_dx2_(double x, double y) {
		return 0;
	}

	@Override
	double dz_dy2_(double x, double y) {
		return 0;
	}

	@Override
	double dz_dxy_(double x, double y) {
		return 0;
	}

	@Override
	boolean contains_(double x, double y) {
		return this.convexhull.contains(x, y);
	}

}

class DistPair implements Comparable<DistPair> {

	double d;
	Triangle t;

	public DistPair(Triangle t, Point p) {

		this.t = t;

		// Abstandsma berechnen
		double a = this.area(t.a, t.b, t.c);

		double asum = 0;
		asum += this.area(p, t.a, t.b);
		asum += this.area(p, t.b, t.c);
		asum += this.area(p, t.a, t.c);

		this.d = Math.abs((asum - a) / a);

		if (this.d < 1e-5) {
			this.d = 0.0;
		}
	}

	// calculate enclosed area in triangle
	public double area(Point p1, Point p2, Point p3) {

		return Math.abs(((p1.x * p2.y) + (p2.x * p3.y) + (p3.x * p1.y)) - (p1.x * p3.y) - (p3.x * p2.y) - (p2.x * p1.y))
				/ 2.0;
	}

	@Override
	public int compareTo(DistPair o) {
		if (o == null) {
			return -1;
		} else {
			return (int) Math.signum(this.d - o.d);
		}
	}

}
