# -*- coding: utf-8 -*-
"""
***********************************************************************
*    ©  UNIVERSITY - L I N Z / LCM - Linz Center of Mechatronics GmbH
*       Institute for Electrical Drives and Power Electronics
***********************************************************************

Module tperiodic contains the class TPeriodic. It is used to work with
harmonic signals
  
authors: gweidenholzer
"""

from lxml import etree
import math
from math import pi
import numpy as np
import copy
import numbers
from dtypes import TData
import base64

from scipy.fftpack import rfft, irfft, fftfreq, ifft


class TPeriodic(TData):
    """
    Instanceses of TPeriodic are used to represent harmonic signals
    
    Attributes:
        rfft.... vector of coeffs according to definition of scipy.fftpack.rfft
                 Warning: The definition of numpy.fft.rfft differs form scipy!
    """

    previewSamples = 1000
    maxorder = 100
    powerTreshold = 1.e-8
    resolution = 10.

    def __init__(self, *args, **kwargs):
        """
        Constructor to make a deep copy
            TPeriodic(tp)   
            
        Construct using equidistant samples of exactly one period (fft)
            TPeriodic(ydata) where ydata.dtype != np.complex
        
        Construct using time samples:
            TPeriodic(phi, ydata)
        
        Construct single harmonic signal
            TPeriodic(order, magnitude, phase)        
        
        Extra arguments:
            TPeriodic(..., maxorder=20)
        """

        self.rfft = np.zeros(1)
        self.rfft[0] = 0.

        maxorder = TPeriodic.maxorder

        if kwargs:
            if kwargs["maxorder"]:
                self.maxorder = kwargs["maxorder"]

        if args:
            nargs = len(args)

            if nargs == 1:
                try:
                    if isinstance(args[0], TPeriodic):
                        """
                        Make a deep copy of given TPeriodic
                        """
                        self.rfft = copy.deepcopy(args[0].rfft)

                    elif np.iscomplexobj(args[0]):
                        """
                        Take complex fft array according to scipy.fftpack.fft
                        """
                        y = args[0]
                        x = ifft(y)
                        self.rfft = rfft(np.real(x))

                    elif isinstance(args[0], np.ndarray):
                        """
                        Take equidistant samples
                        """
                        self.rfft = rfft(args[0])

                except:
                    raise TypeError()

            if nargs == 2:
                """
                Take sample points at angles phi
                """

                # convert to numpy array
                phi = np.array(args[0], np.float)
                ydata = np.array(args[1], np.float)

                # find max. distance btw. samples
                wrapped = ((phi) % (2. * pi))
                hlp = np.sort(wrapped)
                maxdist = np.max(hlp[2:] - hlp[1:-1])
                maxdist = max([hlp[0] - (hlp[-1] - 2 * pi), maxdist])

                mindist = np.min(hlp[2:] - hlp[1:-1])
                mindist = min([hlp[0] - (hlp[-1] - 2 * pi), mindist])

                # even for non-equidistant sampling, it is sufficient when
                # the highest harmonic satisfies the shannon theoreme.
                # A safty factor 0.9 is applied here
                maxorder = min(maxorder, int(0.9 * pi / maxdist));

                if maxorder < 2:
                    raise Exception("No periodic fitting possible!")

                if maxdist / mindist < 1.001:
                    # option: calculate dft for higher performance
                    pass

                A = np.matrix(np.ones_like(phi)).T

                for o in range(1, maxorder + 1):
                    vec1 = np.matrix(2.0 * np.cos(o * phi))
                    vec2 = np.matrix(-2.0 * np.sin(o * phi))
                    A = np.concatenate((A, vec1.T, vec2.T), axis=1)

                x = np.linalg.lstsq(A, np.matrix(ydata).T)[0]
                x = np.array(x)[:, 0]

                # convert vector of coefficients to rfft vector
                self.rfft = x * len(x)

            elif nargs == 3:
                """
                construct using order, mag, phase
                """
                order = args[0]
                mag = args[1]
                phase = args[2]

                n = order * 2 + 1
                re = mag * math.cos(phase)
                im = mag * math.sin(phase)

                self.rfft = np.zeros(n)
                self.rfft[-2] = re * n / 2
                self.rfft[-1] = im * n / 2

        self.pack()

    def pack(self):
        """
        remove close to zero coefficient and truncate trailing zeros
        """
        rms = self.rms()

        if rms > 0:
            tres = self.rms() * math.sqrt(TPeriodic.powerTreshold)
            # ?? tres *= len(self.rfft)  # magnitude treshold

            idx = np.where(abs(self.rfft) < tres)[0]
            self.rfft[idx] = 0.  # zero out

            idx = np.where(abs(self.rfft) >= tres)[0]

            n = len(self.rfft)
            nnew = max(idx) + 1
            self.rfft = self.rfft[:nnew] * nnew / n

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '[periodic: mean=%g, pp=%g]' % (self.mean(), self.pp());

    def getFFT(self):

        n = len(self.rfft)
        res = np.zeros(n, dtype=np.complex)

        res[0] = self.rfft[0]

        if n % 2:
            # odd number
            tmp = self.rfft[1:].view(np.complex)
            res[1:(n + 1) / 2] = tmp
            res[(n + 1) / 2:] = tmp[::-1].conj()
        else:
            # even number
            tmp = self.rfft[1:-1].view(np.complex)
            res[1:n / 2] = tmp
            res[n / 2] = self.rfft[-1]
            res[n / 2 + 1:] = tmp[::-1].conj()
        return res

    def getMaxOrder(self):
        return len(self.rfft) - 1

    def mean(self):
        return self.rfft[0] / len(self.rfft)

    def getSamples(self, samples):
        n = len(self.rfft)
        return irfft(self.rfft, samples) / n * samples

    def min(self):
        samples = len(self.rfft) * TPeriodic.resolution
        return np.min(self.getSamples(samples))

    def max(self):
        samples = len(self.rfft) * TPeriodic.resolution
        return np.max(self.getSamples(samples))

    def pp(self):
        samples = len(self.rfft) * TPeriodic.resolution
        tmp = self.getSamples(samples)
        return np.max(tmp) - np.min(tmp)

    def valueAt(self, phi):
        """
        Get sample points at angles phi
        """

        fftvec = self.getFFT()
        n = len(fftvec)
        orders = fftfreq(n, 1. / n)

        mat = np.exp(np.outer(phi, orders) * 1j)
        val = np.real(np.dot(mat, fftvec)) / n
        return val

    def __offset(self, scale):

        theClone = copy.deepcopy(self)
        theClone.rfft[0] += scale * len(self.rfft)
        return theClone

    def __scale(self, scale):
        theClone = copy.deepcopy(self)
        theClone.rfft *= scale
        return theClone

    def removeDC(self):
        self.rfft[0] = 0.

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):

        if isinstance(other, numbers.Real):
            return self.__offset(-other)
        else:
            return self.__add__(other.__scale(-1.0))

    def __rsub__(self, other):
        return self.__sub__(other).__scale(-1.)

    def __neg__(self):
        return self.__scale(-1.)

    def __pos__(self):
        return copy.deepcopy(self)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.__div__(other)

    def __invert__(self):
        return (TPeriodic() + 1.0) / self

    def __rdiv__(self, other):
        """
        Right side division.
        """
        return (TPeriodic() + other) / self

    def __rtruediv__(self, other):
        return self.__rdiv__(other)

    def getCoeffs(self):
        """
        Return complex coeffs by value        
        """

        n = len(self.rfft)
        tmp = np.zeros(n / 2, dtype=np.complex)

        dc = np.real(self.rfft[0]) / n

        if n % 2:
            # odd number
            tmp = self.rfft[1:].view(np.complex)
        else:
            # even number
            tmp[:-1] = self.rfft[1:-1].view(np.complex)
            tmp[-1] = self.rfft[-1]

        coeffs = np.copy(tmp) * 2. / n

        return (dc, coeffs)

    def rms(self):
        tmp = copy.deepcopy(self.rfft) / len(self.rfft) * 2.
        tmp[0] /= math.sqrt(2.)
        return np.linalg.norm(tmp) / math.sqrt(2.)

    def thd(self):
        p_base = np.linalg.norm(self.rfft[1:3])
        p_harm = np.linalg.norm(self.rfft[3:])
        return p_harm / p_base

    def deriv(self, omega):

        theClone = copy.deepcopy(self)
        theClone.rfft[0] = 0.

        n = len(theClone.rfft)
        tmp = np.zeros(n / 2, dtype=np.complex)

        if n % 2:
            # odd number
            tmp = theClone.rfft[1:].view(np.complex)
        else:
            # even number
            tmp[:-1] = theClone.rfft[1:-1].view(np.complex)
            tmp[-1] = theClone.rfft[-1]

        freqs = np.arange(1, len(tmp) + 1) * omega
        tmp *= 1.j * freqs
        return theClone

    def integ(self, omega):

        n = len(self.rfft)

        y = self.getFFT()

        orders = fftfreq(n, 1. / n)

        # discard dc offset
        y = y[1:]
        orders = orders[1:]

        # calculate 1/s
        y *= 1 / (1j * omega * orders)

        # add zero dc coefficient
        y = np.concatenate((np.array([0]), y))

        # create TPeriodic from fft array   
        return TPeriodic(y)

    def __add__(self, other):

        if isinstance(other, numbers.Real):
            return self.__offset(other)
        else:

            a = self.rfft / len(self.rfft)
            b = other.rfft / len(other.rfft)

            if len(a) < len(b):
                c = b.copy()
                c[:len(a)] += a
            else:
                c = a.copy()
                c[:len(b)] += b

            res = TPeriodic()
            res.rfft = c * len(c)

            return res

    def __mul__(self, other):
        if isinstance(other, numbers.Real):
            return self.__scale(other)
        else:

            n = len(self.rfft) + len(other.rfft)

            ya = self.getSamples(n)
            yb = other.getSamples(n)

            res = TPeriodic(ya * yb)
            res.pack()

            return res

    def __div__(self, other):
        """
        Divide pointwise
        """
        if isinstance(other, numbers.Real):
            return self.__scale(1. / other)
        else:
            n = max(len(self.rfft) + len(other.rfft), TPeriodic.maxorder)
            phi = np.linspace(0, 2 * pi, n, endpoint=False)

            ya = self.getSamples(n)
            yb = other.getSamples(n)

            y = ya / yb

            idx = np.where(y != float('inf'))
            return TPeriodic(phi[idx], y[idx])

    def __pow__(self, p):

        theClone = copy.deepcopy(self)

        if (p % 1.) == 0.:
            if p == 0.:
                return TPeriodic() + 1.
            elif p == 1.:
                return theClone
            elif p > 1.:
                for k in range(1, p):
                    theClone *= self
                return theClone
        else:
            # calc pointwise power
            n = len(self.rfft)
            n = max(n, n * p, n / p, TPeriodic.maxorder)
            phi = np.linspace(0, 2 * pi, n, endpoint=False)

            y = np.power(self.getSamples(n), p)

            idx = np.where(y != float('inf'))
            return TPeriodic(phi[idx], y[idx])

    def showPreview(self):
        """
        Show simple preview of the signal
        """
        import matplotlib.pyplot as plt
        from matplotlib.ticker import FixedLocator, FormatStrFormatter

        phi = np.linspace(0, 360., TPeriodic.previewSamples, endpoint=False)
        val = self.getSamples(TPeriodic.previewSamples)
        plt.plot(phi, val)
        plt.grid(True)

        xax = plt.axes().xaxis
        xax.grid(True, 'minor')
        xax.set_major_locator(FixedLocator(np.linspace(0, 360, 5, endpoint=True)))
        xax.set_major_formatter(FormatStrFormatter(u'%d°'))
        xax.set_minor_locator(FixedLocator(np.linspace(0, 360, 13, endpoint=True)))

        plt.ion()
        plt.draw()

    def _repr_png_(self, title=None, width=400, height=300):

        from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
        from matplotlib.figure import Figure
        import io
        from matplotlib.ticker import FixedLocator, FormatStrFormatter

        phi = np.linspace(0, 360., TPeriodic.previewSamples, endpoint=False)
        val = self.getSamples(TPeriodic.previewSamples)

        fig = Figure(figsize=(width / 100., height / 100.), dpi=100.)
        canvas = FigureCanvas(fig)
        ax = fig.add_subplot(111)

        ax.plot(phi, val)
        if title:
            ax.set_title(title)
        ax.grid(True, 'minor')
        ax.set_xlim([0, 360])
        ax.xaxis.grid(True, 'minor')
        ax.yaxis.grid(True, 'major')
        ax.xaxis.set_major_locator(FixedLocator(np.linspace(0, 360, 5, endpoint=True)))
        ax.xaxis.set_major_formatter(FormatStrFormatter(u'%d°'))
        ax.xaxis.set_minor_locator(FixedLocator(np.linspace(0, 360, 13, endpoint=True)))

        buf = io.BytesIO()
        canvas.print_figure(buf)

        buf.seek(0)
        png = base64.b64encode(buf.getvalue())
        buf.close()

        return png.decode("utf-8")



    @classmethod
    def fromXML(cls, node):

        obj = cls()  # create empty TPeriodic

        for child in node:
            tag = child.tag.lower()
            s = child.text
            if tag == "dcoffset":
                obj += float(s)
            elif tag == "coeffs":
                if not s.startswith("[") or not s.endswith("]"):
                    raise Exception("Invalid Periodic data")
                if not s == "[]":
                    s = s[1:-1]
                    parts = s.split(";")
                    for p in parts:
                        subparts = p.split(",")
                        order = int(subparts[0])
                        phase = 0.
                        if int(subparts[1]) == 1:  # type sin or cos
                            phase = -pi / 2.
                        mag = float(subparts[2])
                        obj += cls(order, mag, phase)
            else:
                raise Exception("Undefined tag: " + tag)

        return obj

    def data2xml(self):

        self.pack()
        (dc, coeffs) = self.getCoeffs()

        node = etree.Element("Periodic")

        dcoffset = etree.Element("DCOffset")
        dcoffset.text = repr(dc)
        node.append(dcoffset)

        content = etree.Element("Coeffs")
        if len(coeffs) > 0:
            tmp = "["
            for idx, c in enumerate(coeffs):
                tmp = tmp + "%d,%d,%1.17g;" % (idx + 1, 0, np.real(c))
                tmp = tmp + "%d,%d,%1.17g;" % (idx + 1, 1, -np.imag(c))
            tmp = tmp[:-1] + "]"
        else:
            tmp = "[]"
        content.text = tmp
        node.append(content)

        return node


if __name__ == "__main__":

    import numpy as np
    import matplotlib.pyplot as plt    
    
    t1=TPeriodic(1,1,0)
    t2=TPeriodic(1,1,np.pi/2)
    
    phi=np.linspace(0,2*np.pi, 100, endpoint=False)

    n=20
    phi=np.linspace(0,2*np.pi, n, endpoint=False)
    s1=t1.getSamples(n)
    s2=t1.valueAt(phi)
    
    plt.plot(phi, s1)
    plt.plot(phi, s2)    
    plt.show()