# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 1024 $
$LastChangedDate:: 2020-06-25 18:45:41 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import unittest
import numpy as np
import numpy.testing as npt
import lcm.math.rbf as rbf
import lcm.math.rbf.kernels as kernels
import lcm.utils.pytperiodic as pm
from lcm.math.rbf.utils.analytic_PeriodicMultiRBF import AnalyticFunction1
from lcm.math.rbf.utils.analytic_PeriodicMultiRBF import AnalyticFunction2


class TestPeriodicMultiRBF(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.plot = False
        n_phi = 1000
        phi = np.linspace(0, 2 * np.pi, n_phi, endpoint=False)

        xi1 = np.array([0.5 + 0.5 * np.sin(phi), 0.2 * np.cos(3. * phi)])
        xi2 = np.array([0.5 * np.ones_like(phi), 0.2 * np.ones_like(phi)])
        tper1 = None
        tper2 = pm.PyTPeriodicMatrix.fromDC(np.array([[2., 0.5], [0.8, 1.5]]))

        md = 0.5

        cls.ref_data = \
            {
                'fun1-xi1-CS': dict(kernelID=1, maxdist=md, xi=xi1, fnc=AnalyticFunction1(10, 10, n_phi), t_per=tper1),
                'fun1-xi2-CS': dict(kernelID=1, maxdist=md, xi=xi2, fnc=AnalyticFunction1(10, 10, n_phi), t_per=tper1),
                'fun1-xi1-TPS': dict(kernelID=2, maxdist=md, xi=xi1, fnc=AnalyticFunction1(10, 10, n_phi), t_per=tper1),
                'fun1-xi2-TPS': dict(kernelID=2, maxdist=md, xi=xi2, fnc=AnalyticFunction1(10, 10, n_phi), t_per=tper1),
                'fun2-xi1-CS': dict(kernelID=1, maxdist=md, xi=xi1, fnc=AnalyticFunction2(10, 10, n_phi), t_per=tper1),
                'fun2-xi2-CS': dict(kernelID=1, maxdist=md, xi=xi2, fnc=AnalyticFunction2(10, 10, n_phi), t_per=tper1),
                'fun2-xi1-TPS': dict(kernelID=2, maxdist=md, xi=xi1, fnc=AnalyticFunction2(10, 10, n_phi), t_per=tper1),
                'fun2-xi2-TPS': dict(kernelID=2, maxdist=md, xi=xi2, fnc=AnalyticFunction2(10, 10, n_phi), t_per=tper1),
                'fun1-xi1-t2-CS': dict(kernelID=1, maxdist=md, xi=xi1, fnc=AnalyticFunction1(10, 10, n_phi),
                                       t_per=tper2),
                'fun1-xi2-t2-CS': dict(kernelID=1, maxdist=md, xi=xi2, fnc=AnalyticFunction1(10, 10, n_phi),
                                       t_per=tper2),
                'fun1-xi1-t2-TPS': dict(kernelID=2, maxdist=md, xi=xi1, fnc=AnalyticFunction1(10, 10, n_phi),
                                        t_per=tper2),
                'fun1-xi2-t2-TPS': dict(kernelID=2, maxdist=md, xi=xi2, fnc=AnalyticFunction1(10, 10, n_phi),
                                        t_per=tper2),
                'fun2-xi1-t2-CS': dict(kernelID=1, maxdist=md, xi=xi1, fnc=AnalyticFunction2(10, 10, n_phi),
                                       t_per=tper2),
                'fun2-xi2-t2-CS': dict(kernelID=1, maxdist=md, xi=xi2, fnc=AnalyticFunction2(10, 10, n_phi),
                                       t_per=tper2),
                'fun2-xi1-t2-TPS': dict(kernelID=2, maxdist=md, xi=xi1, fnc=AnalyticFunction2(10, 10, n_phi),
                                        t_per=tper2),
                'fun2-xi2-t2-TPS': dict(kernelID=2, maxdist=md, xi=xi2, fnc=AnalyticFunction2(10, 10, n_phi),
                                        t_per=tper2)
            }

        key = ''
        try:
            for key, data in cls.ref_data.items():
                fnc = data['fnc']
                xc_per = fnc.getXcPer()
                zc_per = fnc.getZcPer()
                maxdist = data['maxdist']
                t_per = data['t_per']
                kernel = kernels.AbstractKernel.fromID(data['kernelID'])
                data['rbf'] = rbf.PeriodicMultiRBF(xc_per, zc_per, kernel, max_dist=maxdist, t_per=t_per)
        except Exception as e:
            raise unittest.SkipTest("Constructor failed:\n{}: {}".format(key, e))

    def test_fromCBOR(self):
        for key, data in self.ref_data.items():
            for f in [1, 2, 6]:
                with self.subTest(msg='{}-f={}'.format(key, f)):

                    if f in [6, ]:
                        self.skipTest('subtest {}-f={}'.format(key, f))

                    rbf_interp_src = data['rbf']  # type: rbf.PeriodicMultiRBF

                    cbor_obj = rbf_interp_src.toCBOR(f)

                    self.assertEqual(cbor_obj['xi_per']['f'], f)
                    self.assertEqual(cbor_obj['t_per']['f'], f)
                    self.assertEqual(cbor_obj['c_per']['f'], f)
                    self.assertEqual(cbor_obj['d_per']['f'], f)

                    rbf_interp = rbf.PeriodicMultiRBF.fromCBOR(cbor_obj)

                    xi = data['xi']
                    fnc = data['fnc']
                    # xi_per = fnc.getXiPer(xi)
                    phi = fnc.phi

                    ''' rbf result'''
                    zi = rbf_interp(xi[:, np.newaxis, :], phi)
                    # zi_per = pm.PyTPeriodicMatrix.fromSamples(zi)

                    ''' analytic result'''
                    zi_analytic = fnc.getZi(xi[0, :], xi[1, :], phi)
                    # zi_analytic_per = fnc.getZiPer(xi[0, :], xi[1, :], phi)

                    '''error'''
                    if key in ['fun1-xi1-TPS', 'fun1-xi1-t2-TPS']:
                        rtol = 2e-2
                        atol = 1e-3
                    else:
                        rtol = 1e-2
                        atol = 1e-3

                    if self.plot:
                        fnc.plotComparison(zi, zi_analytic, atol=atol, rtol=rtol)

                    error = np.max(np.abs(zi - zi_analytic))
                    max_error = np.linalg.norm(zi_analytic)
                    npt.assert_array_less(error, rtol * max_error)

                    '''compare signals'''
                    npt.assert_allclose(zi, zi_analytic, atol=atol, rtol=rtol)

    @unittest.skip
    def test_fromMatrices(self):
        self.fail()

    @unittest.skip
    def test_xdim(self):
        self.fail()

    @unittest.skip
    def test_zdim(self):
        self.fail()

    @unittest.skip
    def test_initFromSamples(self):
        self.fail()

    @unittest.skip
    def test_initFromMatrices(self):
        self.fail()

    @unittest.skip
    def test_jacobian(self):
        self.fail()

    def test_discrete_angle_interpolation(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                phi = fnc.phi
                rbf_interp = data['rbf']  # type: rbf.PeriodicMultiRBF

                ''' rbf result'''
                zi = rbf_interp(xi[:, np.newaxis, :], phi)
                # zi_per = pm.PyTPeriodicMatrix.fromSamples(zi)

                ''' analytic result'''
                zi_analytic = fnc.getZi(xi[0, :], xi[1, :], phi)
                # zi_analytic_per = fnc.getZiPer(xi[0, :], xi[1, :], phi)

                '''error'''
                if key in ['fun1-xi1-TPS', 'fun1-xi1-t2-TPS']:
                    rtol = 2e-2
                    atol = 1e-3
                else:
                    rtol = 1e-2
                    atol = 1e-3
                if self.plot:
                    fnc.plotComparison(zi, zi_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(zi - zi_analytic))
                max_error = np.linalg.norm(zi_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(zi, zi_analytic, atol=atol, rtol=rtol)

    def test_periodic_interpolation(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                xi_per = fnc.getXiPer(xi)
                phi = fnc.phi
                n_phi = fnc.n_phi
                rbf_interp = data['rbf']  # type: rbf.PeriodicMultiRBF

                ''' rbf result'''
                zi_per = rbf_interp(xi_per)
                zi = zi_per.get_samples(n_phi)

                ''' analytic result'''
                zi_analytic = fnc.getZi(xi[0, :], xi[1, :], phi)
                # zi_analytic_per = fnc.getZiPer(xi[0, :], xi[1, :], phi)

                '''error'''
                if key in ['fun1-xi1-TPS', 'fun1-xi1-t2-TPS']:
                    rtol = 2e-2
                    atol = 1e-3
                else:
                    rtol = 1e-2
                    atol = 1e-3

                if self.plot:
                    fnc.plotComparison(zi, zi_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(zi - zi_analytic))
                max_error = np.linalg.norm(zi_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(zi, zi_analytic, atol=atol, rtol=rtol)

    @unittest.skip
    def test_discrete_angle_jacobian(self):
        self.fail()

    def test_periodic_jacobian(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):

                if key in ['fun1-xi1-CS', 'fun1-xi1-TPS', 'fun1-xi1-t2-CS', 'fun1-xi1-t2-TPS']:
                    self.skipTest('subtest {} skipped'.format(key))

                xi = data['xi']
                fnc = data['fnc']
                xi_per = fnc.getXiPer(xi)
                phi = fnc.phi
                n_phi = fnc.n_phi
                rbf_interp = data['rbf']  # type: rbf.PeriodicMultiRBF

                ''' get jacobian matrix ( only derivation along x-dimensions, no phi)'''
                dzi_dx_per = rbf_interp.jacobian(xi_per)

                ''' analytic result'''
                dzi_dx_analytic_per = fnc.get_dzi_dxPer(xi[0, :], xi[1, :], phi)

                '''get samples'''
                dzi_dx = dzi_dx_per.get_samples(n_phi)
                dzi_dx_analytic = dzi_dx_analytic_per.get_samples(n_phi)

                '''error'''
                if key in ['fun1-xi1-CS', 'fun1-xi1-TPS', 'fun1-xi1-t2-CS', 'fun1-xi1-t2-TPS']:
                    rtol = 2e-2
                    atol = 1e-3
                else:
                    rtol = 1e-2
                    atol = 1e-3
                if self.plot:
                    fnc.plotComparison(dzi_dx, dzi_dx_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(dzi_dx - dzi_dx_analytic))
                max_error = np.linalg.norm(dzi_dx_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(dzi_dx, dzi_dx_analytic, atol=atol, rtol=rtol)

    @unittest.skip('Skipped due to unreliable results')
    def test_periodic_angle_derivative(self):
        # Todo: make test reliable so it does not even fail once in 100 runs
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                xi_per = fnc.getXiPer(xi)
                phi = fnc.phi
                n_phi = fnc.n_phi
                rbf_interp = data['rbf']  # type: rbf.PeriodicMultiRBF

                ''' get angle derivative (dz/dphi)'''
                omega = 4
                dz_dphi_per = rbf_interp.periodic_angle_derivative(xi_per, omega)

                ''' analytic result'''
                dz_dphi_analytic_per = fnc.get_dzi_dphiPer(xi[0, :], xi[1, :], phi, omega)

                ''' rbf result'''
                zi_per = rbf_interp(xi_per)
                dz_dphi_tp_per = zi_per.deriv(omega)

                '''get samples'''
                dz_dphi = dz_dphi_per.get_samples(n_phi)
                dz_dphi_analytic = dz_dphi_analytic_per.get_samples(n_phi)
                dz_dphi_per = dz_dphi_tp_per.get_samples(n_phi)

                '''error'''
                rtol = 1.01e-2
                atol = 1e-3

                if self.plot:
                    fnc.plotComparison(dz_dphi, dz_dphi_analytic, atol=atol, rtol=rtol)

                # error = np.max(np.abs(dz_dphi - dz_dphi_analytic))
                # max_error = np.linalg.norm(dz_dphi_analytic)
                # npt.assert_array_less(error, rtol * max_error)
                #
                '''compare signals'''
                npt.assert_allclose(dz_dphi, dz_dphi_analytic, atol=atol, rtol=rtol)

    @unittest.skip
    def test_transform_order(self):
        self.fail()

    @unittest.skip
    def test_to_msgpack(self):
        self.fail()

    @unittest.skip
    def test_write_msgpack(self):
        self.fail()

    def test_toCBOR(self):
        for key, data in self.ref_data.items():
            for f in [1, 2, 6]:
                with self.subTest(msg='{}-f={}'.format(key, f)):
                    rbf_interp = data['rbf']

                    cbor_obj = rbf_interp.toCBOR(f)

                    self.assertEqual(cbor_obj['type'], '__periodicRBF__')
                    self.assertEqual(cbor_obj['max_dist'], rbf_interp.max_dist)
                    self.assertEqual(cbor_obj['kernel'], rbf_interp.kernel.id)
                    npt.assert_array_equal(cbor_obj['params'], rbf_interp.kernel.params)

                    self.assertEqual(cbor_obj['xi_per']['f'], f)
                    self.assertEqual(cbor_obj['t_per']['f'], f)
                    self.assertEqual(cbor_obj['c_per']['f'], f)
                    self.assertEqual(cbor_obj['d_per']['f'], f)

                    self.assertEqual(cbor_obj['xi_per']['f'], rbf_interp.xc_t_per.toCBOR(f)['f'])
                    self.assertEqual(cbor_obj['t_per']['f'], rbf_interp.t_per.toCBOR(f)['f'])
                    self.assertEqual(cbor_obj['c_per']['f'], rbf_interp.c_per.toCBOR(f)['f'])
                    self.assertEqual(cbor_obj['d_per']['f'], rbf_interp.d_per.toCBOR(f)['f'])

                    npt.assert_array_equal(cbor_obj['xi_per']['bins'], rbf_interp.xc_t_per.toCBOR(f)['bins'])
                    npt.assert_array_equal(cbor_obj['t_per']['bins'], rbf_interp.t_per.toCBOR(f)['bins'])
                    npt.assert_array_equal(cbor_obj['c_per']['bins'], rbf_interp.c_per.toCBOR(f)['bins'])
                    npt.assert_array_equal(cbor_obj['d_per']['bins'], rbf_interp.d_per.toCBOR(f)['bins'])

    def test_toRBF(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                xi_per = fnc.getXiPer(xi)
                phi = fnc.phi
                rbf_interp = data['rbf']  # type: rbf.PeriodicMultiRBF
                rbf_mean = rbf_interp.toRBF()

                # ''' rbf result'''
                # zi_per = rbf_interp(xi_per)
                # zi_samples = zi_per.get_samples(fnc.n_phi)

                ''' rbf mean result'''
                zi = rbf_mean(xi_per.mean())

                ''' analytic result'''
                xi_mean = xi.mean(axis=1, keepdims=True) * np.ones_like(xi)
                zi_analytic = fnc.getZi(xi_mean[0, :], xi_mean[1, :], phi)

                '''error'''
                rtol = 1e-3
                atol = 0.0
                if self.plot:
                    zi_mean = zi[:, :, np.newaxis] * np.ones_like(zi_analytic)
                    zi_analytic_mean = zi_analytic.mean(keepdims=True, axis=2) * np.ones_like(zi_analytic)
                    fnc.plotComparison(zi_mean, zi_analytic_mean, atol=atol, rtol=rtol)
                    # fnc.plotComparison(zi_mean, zi_analytic, atol=atol, rtol=rtol)
                    # fnc.plotComparison(zi_mean, zi_analytic, atol=atol, rtol=rtol)
 
                '''compare signals'''
                npt.assert_allclose(zi, zi_analytic.mean(axis=2), atol=atol, rtol=rtol)

    @unittest.skip
    def test_select(self):
        self.fail()

    @unittest.skip
    def test_getMaxOrder(self):
        self.fail()

    @unittest.skip
    def test_sample(self):
        self.fail()
