# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 1024 $
$LastChangedDate:: 2020-06-25 18:45:41 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import unittest
import numpy as np
import numpy.testing as npt
import lcm.utils.pytperiodic as pm
from lcm.tdata_interface.pytperiodic_interface import PyTPeriodicInterface
from lcm.tdata_interface.tdata_testing import TDataTesting


class TestPyTPeriodicInterface(unittest.TestCase):

    @staticmethod
    def getCoeffs(jobj, pyobj: pm.PyTPeriodicMatrix):

        nrows = pyobj.nrows()
        ncols = pyobj.ncols()
        max_order = pyobj.get_max_order()
        # max_order = jobj.getMaxOrder()

        jdc = np.zeros((nrows, ncols))
        jcos_coeffs = np.zeros((nrows, ncols, max_order))
        jsin_coeffs = np.zeros((nrows, ncols, max_order))

        for row in range(nrows):
            for col in range(ncols):

                if jobj.getClassName() == "TPeriodic":
                    jcur = jobj
                else:
                    jcur = jobj.get(row, col)

                jdc[row, col] = jcur.mean()
                for order in range(max_order):
                    jcos_coeffs[row, col, order] = jcur.getCoeffA(order + 1)
                    jsin_coeffs[row, col, order] = jcur.getCoeffB(order + 1)

        jcoeffs = [jdc, jcos_coeffs, jsin_coeffs]
        pycoeffs = pyobj.get_coeffs(range(max_order + 1))

        return jcoeffs, pycoeffs

    @staticmethod
    def getMaxOrder(jobj, pyobj: pm.PyTPeriodicMatrix):

        nrows = pyobj.nrows()
        ncols = pyobj.ncols()

        jmax_order = np.zeros((nrows, ncols))

        for row in range(nrows):
            for col in range(ncols):

                if jobj.getClassName() == "TPeriodic":
                    jcur = jobj
                else:
                    jcur = jobj.get(row, col)

                jmax_order[row, col] = jcur.getMaxOrder()

        pymax_order = pyobj.get_max_order()

        return np.max(jmax_order), pymax_order

    @unittest.skipIf(PyTPeriodicInterface.getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_convert_J2Py(self):

        for order in range(1, 10, 2):
            for magnitude in np.linspace(0, 1, 10):
                for angle in np.linspace(-np.pi, np.pi, 10):
                    jobj = TDataTesting.getTPeriodic(magnitude, order, angle)
                    pyobj = PyTPeriodicInterface.convert_J2Py(jobj)

                    jcoeffs, pycoeffs = self.getCoeffs(jobj, pyobj)

                    npt.assert_almost_equal(pycoeffs[0], jcoeffs[0])
                    npt.assert_almost_equal(pycoeffs[1], jcoeffs[1])
                    npt.assert_almost_equal(pycoeffs[2], jcoeffs[2])

                    jmax_order, pymax_order = self.getMaxOrder(jobj, pyobj)

                    npt.assert_almost_equal(jmax_order, pymax_order)

    @unittest.skipIf(PyTPeriodicInterface.getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_convert_Py2J(self):

        for order in range(1, 10, 2):
            for magnitude in np.linspace(0, 1, 10):
                for angle in np.linspace(-np.pi, np.pi, 10):
                    pyobj = TDataTesting.getPyTPeriodic(magnitude, order, angle)
                    jobj = PyTPeriodicInterface.convert_Py2J(pyobj)

                    jcoeffs, pycoeffs = self.getCoeffs(jobj, pyobj)

                    npt.assert_almost_equal(pycoeffs[0], jcoeffs[0])
                    npt.assert_almost_equal(pycoeffs[1], jcoeffs[1])
                    npt.assert_almost_equal(pycoeffs[2], jcoeffs[2])

                    jmax_order, pymax_order = self.getMaxOrder(jobj, pyobj)

                    npt.assert_almost_equal(jmax_order, pymax_order)

    @unittest.skipIf(PyTPeriodicInterface.getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_convert_J2PyMatrix(self):
        nrows = 2
        ncols = 3

        for order in range(1, 10, 2):
            for magnitude in np.linspace(1, 10, 5):
                for angle in np.linspace(-np.pi, np.pi, 5):
                    jobj = TDataTesting.getTPeriodicMatrix(nrows, ncols, order, magnitude, angle)
                    pyobj = PyTPeriodicInterface.convert_J2Py(jobj)

                    jcoeffs, pycoeffs = self.getCoeffs(jobj, pyobj)

                    npt.assert_almost_equal(pycoeffs[0], jcoeffs[0])
                    npt.assert_almost_equal(pycoeffs[1], jcoeffs[1])
                    npt.assert_almost_equal(pycoeffs[2], jcoeffs[2])

                    jmax_order, pymax_order = self.getMaxOrder(jobj, pyobj)

                    npt.assert_almost_equal(jmax_order, pymax_order)

    @unittest.skipIf(PyTPeriodicInterface.getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_convert_Py2JMatrix(self):
        nrows = 2
        ncols = 3

        for order in range(1, 10, 2):
            for magnitude in np.linspace(1, 10, 5):
                for angle in np.linspace(-np.pi, np.pi, 5):
                    pyobj = TDataTesting.getPyTPeriodicMatrix(nrows, ncols, order, magnitude, angle)
                    jobj = PyTPeriodicInterface.convert_Py2J(pyobj)

                    jcoeffs, pycoeffs = self.getCoeffs(jobj, pyobj)

                    npt.assert_almost_equal(pycoeffs[0], jcoeffs[0])
                    npt.assert_almost_equal(pycoeffs[1], jcoeffs[1])
                    npt.assert_almost_equal(pycoeffs[2], jcoeffs[2])

                    jmax_order, pymax_order = self.getMaxOrder(jobj, pyobj)

                    npt.assert_almost_equal(jmax_order, pymax_order)


if __name__ == "__main__":
    from lcm.utils.testing import runUnittest

    runUnittest(PyTPeriodicInterface.getInterpreter())
