/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.utils;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;

import javax.swing.AbstractListModel;
import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileNameExtensionFilter;

import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;
import at.lcm.bu21.magopt.structure.extra.Analysis;

/**
 * Anaylysis dialog
 *
 * @author Werner Koppelsttter
 */
public class AnalysisDialog extends JDialog {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Settings-panel */
	private JPanel settings;
	/** Layout for settings */
	private CardLayout cards;
	/** Index of layout */
	private int cardIndex = 0;

	/** Panel for unused/used selection */
	private JPanel useSelection;
	/** List for unused-parameters */
	private JList<IParameter> unusedList;
	/** List for used-parameters */
	private JList<IParameter> usedList;
	/** Model for unused-list */
	private AnalysisUnusedListModel unusedModel;
	/** Model for used-list */
	private AnalysisUsedListModel usedModel;
	/** Scroll-pane for unused list */
	private JScrollPane unusedScroll;
	/** Scroll-pane for used list */
	private JScrollPane usedScroll;
	/** Select parameter to used list */
	private JButton selectUsed;
	/** Deselect parameter from used list */
	private JButton deselectUsed;

	/** Panel with progress-bar */
	private JPanel progress;

	/** Create analysis */
	private JButton create;
	/** Cancel dialog */
	private JButton cancel;
	/** Load settings */
	private JButton load;
	/** Save settings */
	private JButton save;

	/** Analysis-object */
	private Analysis analysis;

	/** Based project */
	private IProject project;

	/** Used settings-file */
	private File settingFile = null;

	/** Size of dialog (used for resize) */
	private Dimension size;

	/**
	 * List-model for unused parameters
	 * 
	 * @author Werner Koppelsttter
	 */
	private class AnalysisUnusedListModel
			extends AbstractListModel<IParameter> {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getSize() {
			return AnalysisDialog.this.analysis.getUnusedSize();
		}

		@Override
		public IParameter getElementAt(int index) {
			return AnalysisDialog.this.analysis.getUnused(index);
		}

		/**
		 * Fire parameter added to the end of list
		 */
		public void fireAdded() {
			super.fireIntervalAdded(this,
					AnalysisDialog.this.analysis.getUnusedSize() - 1,
					AnalysisDialog.this.analysis.getUnusedSize() - 1);
		}

		/**
		 * Fire one parameter removed from list
		 */
		public void fireRemoved() {
			super.fireIntervalRemoved(this,
					AnalysisDialog.this.analysis.getUnusedSize(),
					AnalysisDialog.this.analysis.getUnusedSize());
			super.fireContentsChanged(this, 0,
					AnalysisDialog.this.analysis.getUnusedSize() - 1);
		}

		/**
		 * Fire content of list changed
		 */
		public void contentChanged() {
			super.fireContentsChanged(this, 0,
					AnalysisDialog.this.analysis.getUnusedSize() - 1);
		}

		/**
		 * Is list empty
		 * 
		 * @return Is empty
		 */
		public boolean isEmpty() {
			return AnalysisDialog.this.analysis.getUnusedSize() == 0;
		}
	}

	/**
	 * List-model for used parameters
	 * 
	 * @author Werner Koppelsttter
	 */
	public class AnalysisUsedListModel extends AbstractListModel<IParameter> {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getSize() {
			return AnalysisDialog.this.analysis.getUsedSize();
		}

		@Override
		public IParameter getElementAt(int index) {
			return AnalysisDialog.this.analysis.getUsed(index);
		}

		/**
		 * Fire parameter added to the end of list
		 */
		public void fireAdded() {
			super.fireIntervalAdded(this,
					AnalysisDialog.this.analysis.getUsedSize() - 1,
					AnalysisDialog.this.analysis.getUsedSize() - 1);
		}

		/**
		 * Fire one parameter removed from list
		 */
		public void fireRemoved() {
			super.fireIntervalRemoved(this,
					AnalysisDialog.this.analysis.getUsedSize(),
					AnalysisDialog.this.analysis.getUsedSize());
			super.fireContentsChanged(this, 0,
					AnalysisDialog.this.analysis.getUsedSize() - 1);
		}

		/**
		 * Fire content of list changed
		 */
		public void contentChanged() {
			super.fireContentsChanged(this, 0,
					AnalysisDialog.this.analysis.getUsedSize() - 1);
		}

		/**
		 * Is list empty
		 * 
		 * @return Is empty
		 */
		public boolean isEmpty() {
			return AnalysisDialog.this.analysis.getUsedSize() == 0;
		}
	}

	/**
	 * Cell render for unused/used-list
	 * 
	 * @author Werner Koppelsttter
	 */
	public class IParameterCellRenderer extends JLabel
			implements ListCellRenderer<IParameter> {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public Component getListCellRendererComponent(
				JList<? extends IParameter> list, IParameter value, int index,
				boolean isSelected, boolean cellHasFocus) {

			this.setBorder(new EmptyBorder(2, 10, 2, 10));

			try {
				this.setText(value.getName());
				this.setToolTipText(value.getIdentifier());
			} catch (RemoteException e) {
				this.setText("???");
				this.setToolTipText("");
			}
			if (isSelected) {
				this.setBackground(list.getSelectionBackground());
				this.setForeground(list.getSelectionForeground());
			} else {
				this.setBackground(list.getBackground());
				this.setForeground(list.getForeground());
			}
			this.setEnabled(list.isEnabled());
			this.setFont(list.getFont());
			this.setOpaque(true);

			return this;
		}
	}

	/**
	 * Create analysis dialog
	 * 
	 * @param owner
	 *            Owner of dialog
	 * @param project
	 *            Project
	 */
	public AnalysisDialog(Component owner, IProject project) {
		this.project = project;
		try {
			this.analysis = new Analysis(project);
		} catch (RemoteException e1) {
			System.err.println(e1.getMessage());
			e1.printStackTrace();
		}

		this.setModal(true);
		this.init();

		this.pack();
		this.setLocationRelativeTo(owner);

		// resize
		this.size = this.getSize();
		this.addComponentListener(new ComponentAdapter() {

			@Override
			public void componentResized(ComponentEvent e) {
				Dimension newSize, compSize; // , minSize;
				int width, height;

				newSize = e.getComponent().getSize();
				compSize = AnalysisDialog.this.unusedScroll.getSize();
				width = compSize.width
						+ (newSize.width - AnalysisDialog.this.size.width);
				height = compSize.height
						+ (newSize.height - AnalysisDialog.this.size.height);
				AnalysisDialog.this.size = newSize;

				AnalysisDialog.this.unusedScroll
						.setPreferredSize(new Dimension(width, height));
				AnalysisDialog.this.usedScroll
						.setPreferredSize(new Dimension(width, height));
			}
		});

		this.refreshLists();
		this.refresh();

		this.setVisible(true);
	}

	/**
	 * Init dialog
	 */
	private void init() {
		this.setTitle("Analysis");
		this.setLayout(new BorderLayout());
		this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		GroupLayout layout = new GroupLayout(this.getContentPane());
		this.getContentPane().setLayout(layout);
		layout.setAutoCreateGaps(true);
		layout.setAutoCreateContainerGaps(true);

		this.initButtons();
		this.initSettings();
		this.initProgress();

		layout.setHorizontalGroup(
				layout.createParallelGroup().addComponent(this.settings)
						.addGroup(layout.createSequentialGroup()
								// .addPreferredGap(ComponentPlacement.UNRELATED,
								// 100, 10000)
								.addComponent(this.load).addComponent(this.save)
								.addPreferredGap(ComponentPlacement.UNRELATED,
										50, 10000)
								.addComponent(this.create)
								.addComponent(this.cancel)));
		layout.setVerticalGroup(layout.createSequentialGroup()
				.addComponent(this.settings)
				.addGroup(layout.createParallelGroup().addComponent(this.load)
						.addComponent(this.save).addComponent(this.create)
						.addComponent(this.cancel)));
		layout.linkSize(this.load, this.save, this.create, this.cancel);

	}

	/**
	 * Initialize settings
	 */
	private void initSettings() {
		this.settings = new JPanel();

		this.initUseSelection();
		this.initProgress();

		this.cards = new CardLayout();
		this.settings.setLayout(this.cards);

		this.settings.add(this.useSelection, "0");
		this.settings.add(this.progress, "9");
	}

	/**
	 * Initialize used selection panel
	 */
	private void initUseSelection() {
		GroupLayout layout;

		this.useSelection = new JPanel();
		layout = new GroupLayout(this.useSelection);
		this.useSelection.setLayout(layout);
		this.selectUsed = new JButton(">>");
		this.useSelection.add(this.selectUsed);
		this.selectUsed.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				int index = AnalysisDialog.this.unusedList.getSelectedIndex()
						- 1;

				for (IParameter p : AnalysisDialog.this.unusedList
						.getSelectedValuesList()) {
					AnalysisDialog.this.select(p);
				}

				if (index < 0) {
					index = 0;
				}
				AnalysisDialog.this.unusedList.setSelectedIndex(index);

				AnalysisDialog.this.refresh();
			}
		});
		this.deselectUsed = new JButton("<<");
		this.useSelection.add(this.deselectUsed);
		this.deselectUsed.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				int index = AnalysisDialog.this.usedList.getSelectedIndex() - 1;

				for (IParameter p : AnalysisDialog.this.usedList
						.getSelectedValuesList()) {
					AnalysisDialog.this.deselect(p);
				}

				if (index < 0) {
					index = 0;
				}
				AnalysisDialog.this.usedList.setSelectedIndex(index);

				AnalysisDialog.this.refresh();
			}
		});

		// this.nonSelectModel = new DefaultListModel<IParameter>();
		this.unusedModel = new AnalysisUnusedListModel();
		this.unusedList = new JList<IParameter>(this.unusedModel);
		this.unusedList.setCellRenderer(new IParameterCellRenderer());
		this.unusedList.setSelectionMode(
				ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		this.unusedList.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() == 2) {
					AnalysisDialog.this.select(
							AnalysisDialog.this.unusedList.getSelectedValue());
				}
			}
		});
		this.unusedScroll = new JScrollPane(this.unusedList);
		this.unusedScroll.setPreferredSize(new Dimension(300, 400));
		this.useSelection.add(this.unusedScroll);

		// this.selectModel = new DefaultListModel<IParameter>();
		this.usedModel = new AnalysisUsedListModel();
		this.usedList = new JList<IParameter>(this.usedModel);
		this.usedList.setCellRenderer(new IParameterCellRenderer());
		this.usedList.setSelectionMode(
				ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		this.usedList.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() == 2) {
					AnalysisDialog.this.deselect(
							AnalysisDialog.this.usedList.getSelectedValue());
				}
			}
		});
		this.usedScroll = new JScrollPane(this.usedList);
		this.usedScroll.setPreferredSize(new Dimension(300, 400));
		this.useSelection.add(this.usedScroll);

		layout.setHorizontalGroup(
				layout.createSequentialGroup().addComponent(this.unusedScroll)
						.addGroup(layout.createParallelGroup()
								.addComponent(this.selectUsed)
								.addComponent(this.deselectUsed))
						.addComponent(this.usedScroll));
		layout.setVerticalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.CENTER)
						.addComponent(this.unusedScroll)
						.addGroup(layout.createSequentialGroup()
								.addComponent(this.selectUsed)
								.addComponent(this.deselectUsed))
						.addComponent(this.usedScroll));
		layout.linkSize(this.selectUsed, this.deselectUsed);

	}

	/**
	 * Initialize progress panel
	 */
	private void initProgress() {
		JProgressBar progressBar;
		GroupLayout group;

		this.progress = new JPanel();
		group = new GroupLayout(this.progress);
		this.progress.setLayout(group);

		progressBar = new JProgressBar();
		progressBar.setMinimum(0);
		progressBar.setIndeterminate(true);
		progressBar.setStringPainted(false);
		this.progress.add(progressBar);

		group.setHorizontalGroup(
				group.createSequentialGroup().addComponent(progressBar));
		group.setVerticalGroup(group.createSequentialGroup()
				.addPreferredGap(ComponentPlacement.UNRELATED, 1, 10000)
				.addComponent(progressBar)
				.addPreferredGap(ComponentPlacement.UNRELATED, 1, 10000));
	}

	/**
	 * Select used parameter
	 * 
	 * @param param
	 *            Parameter
	 */
	private void select(IParameter param) {
		if (this.analysis.select(param)) {
			this.unusedModel.fireRemoved();
			this.usedModel.fireAdded();
		}
	}

	/**
	 * Deselect used parameter
	 * 
	 * @param param
	 *            Parameter
	 */
	private void deselect(IParameter param) {
		if (this.analysis.deselect(param)) {
			this.unusedModel.fireAdded();
			this.usedModel.fireRemoved();
		}
	}

	/**
	 * Initialize buttons
	 */
	private void initButtons() {

		this.load = new JButton("Load");
		this.add(this.load);
		this.load.setToolTipText("Load settings");
		this.load.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser chooser;
				File dir = null;

				chooser = new JFileChooser();
				if (AnalysisDialog.this.settingFile != null) {
					dir = AnalysisDialog.this.settingFile.getParentFile();
				} else {
					try {
						dir = AnalysisDialog.this.project.getMopFile()
								.getParentFile();
					} catch (RemoteException e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}

				}
				chooser.setCurrentDirectory(dir);

				// chooser.addChoosableFileFilter(new FileNameExtensionFilter(
				// "Analysis-file (*.acf)", "acf"));
				chooser.setFileFilter(new FileNameExtensionFilter(
						"Analysis-file (*.acf)", "acf"));

				chooser.setAcceptAllFileFilterUsed(false);
				chooser.setDialogTitle("Choose Analysis-file");
				chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				chooser.setMultiSelectionEnabled(false);
				chooser.setFileHidingEnabled(true);

				if (chooser.showOpenDialog(
						AnalysisDialog.this) == JFileChooser.APPROVE_OPTION) {

					AnalysisDialog.this.settingFile = chooser.getSelectedFile();
					try {
						AnalysisDialog.this.analysis
								.loadSettings(AnalysisDialog.this.settingFile);
						AnalysisDialog.this.refreshLists();
						AnalysisDialog.this.refresh();
					} catch (Exception e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}
				}
			}
		});

		this.save = new JButton("Save");
		this.add(this.save);
		this.save.setToolTipText("Save settings");
		this.save.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser chooser;
				File dir = null, file;

				chooser = new JFileChooser();
				if (AnalysisDialog.this.settingFile != null) {
					dir = AnalysisDialog.this.settingFile.getParentFile();
				} else {
					try {
						dir = AnalysisDialog.this.project.getMopFile()
								.getParentFile();
					} catch (RemoteException e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}

				}
				chooser.setCurrentDirectory(dir);

				// chooser.addChoosableFileFilter(new FileNameExtensionFilter(
				// "Analysis-file (*.acf)", "acf"));
				chooser.setFileFilter(new FileNameExtensionFilter(
						"Analysis-file (*.acf)", "acf"));

				chooser.setAcceptAllFileFilterUsed(false);
				chooser.setDialogTitle("Choose Analysis-file");
				chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				chooser.setMultiSelectionEnabled(false);
				chooser.setFileHidingEnabled(true);

				if (chooser.showSaveDialog(
						AnalysisDialog.this) == JFileChooser.APPROVE_OPTION) {

					file = chooser.getSelectedFile();
					if (!file.getName().toLowerCase().endsWith(".acf")) {
						file = new File(file.getParentFile(),
								file.getName() + ".acf");
					}
					AnalysisDialog.this.settingFile = file;

					try {
						AnalysisDialog.this.analysis
								.storeSettings(AnalysisDialog.this.settingFile);
					} catch (Exception e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}
				}

			}
		});

		this.create = new JButton("Create Excel");
		this.add(this.create);
		this.create.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				final JFileChooser chooser;
				File dir = null;

				chooser = new JFileChooser();
				if (AnalysisDialog.this.settingFile != null) {
					dir = AnalysisDialog.this.settingFile.getParentFile();
				} else {
					try {
						dir = AnalysisDialog.this.project.getMopFile()
								.getParentFile();
					} catch (RemoteException e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}

				}
				chooser.setCurrentDirectory(dir);

				// chooser.addChoosableFileFilter(new FileNameExtensionFilter(
				// "Excel files (*.xls, *.xlsx)", "xls", "xlsx"));
				chooser.setFileFilter(new FileNameExtensionFilter(
						"Excel files (*.xls, *.xlsx)", "xls", "xlsx"));

				chooser.setAcceptAllFileFilterUsed(false);
				chooser.setDialogTitle("Choose Analysis-file");
				chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				chooser.setMultiSelectionEnabled(false);
				chooser.setFileHidingEnabled(true);

				if (chooser.showSaveDialog(
						AnalysisDialog.this) == JFileChooser.APPROVE_OPTION) {
					AnalysisDialog.this.cardIndex = 9;
					AnalysisDialog.this.cards.show(AnalysisDialog.this.settings,
							Integer.toString(AnalysisDialog.this.cardIndex));
					AnalysisDialog.this.refresh();
					new Thread() {
						@Override
						public void run() {
							File file;

							file = chooser.getSelectedFile();
							try {
								file = AnalysisDialog.this.analysis
										.create(file);
							} catch (IOException e2) {
								System.err.println(e2.getMessage());
								e2.printStackTrace();
							}

							if (Desktop.isDesktopSupported()) {
								try {
									Desktop.getDesktop().open(file);
								} catch (IOException e1) {
									System.err.println(e1.getMessage());
									e1.printStackTrace();
								}
							}

							AnalysisDialog.this.setVisible(false);
							AnalysisDialog.this.dispose();
						}
					}.start();
				}
			}
		});

		this.cancel = new JButton("Cancel");
		this.add(this.cancel);
		this.cancel.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				AnalysisDialog.this.setVisible(false);
				AnalysisDialog.this.dispose();
			}
		});
	}

	/**
	 * Refresh unused and used list
	 */
	private void refreshLists() {
		AnalysisDialog.this.unusedModel.contentChanged();
		AnalysisDialog.this.usedModel.contentChanged();
	}

	/**
	 * Refresh dialog
	 */
	private void refresh() {
		this.selectUsed.setEnabled(!this.unusedModel.isEmpty());
		this.deselectUsed.setEnabled(!this.usedModel.isEmpty());
		this.create.setEnabled(!this.usedModel.isEmpty());
		this.cancel.setEnabled(this.cardIndex < 9);
		this.load.setEnabled(this.cardIndex < 9);
		this.save.setEnabled(this.cardIndex < 9);
	}
}
