/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9814 $
 * $Date: 2020-04-30 16:26:03 +0200 (Thu, 30 Apr 2020) $
 */
package at.lcm.symspace.center.views;

import java.rmi.RemoteException;

import javax.swing.JComponent;
import javax.swing.JFrame;

import at.lcm.bu21.magopt.structure.Project;
import at.lcm.symspace.center.CenterController;
import at.lcm.symspace.center.views.indices.IndicesViewPanel;
import at.lcm.symspace.center.views.tree.StructureViewPanel;
import at.lcm.symspace.gui.views.DisablePanel;
import at.lcm.symspace.gui.views.SplitView;
import at.lcm.symspace.gui.views.TabView;
import at.lcm.symspace.gui.views.ViewManager;
import magopt.gui.console.Console;
import magopt.gui.console.PythonConsole;
import magopt.gui.event.EventView;
import magopt.gui.view.FlagsView;
import magopt.gui.view.ParamListener;
import magopt.gui.view.PreviewView;
import magopt.gui.view.component.ComponentView;
import magopt.gui.view.details.DetailView;
import magopt.gui.view.docfiles.DocFilesView;
import magopt.gui.view.port.ImportExportView;
import magopt.gui.view.report.ReportParameterView;
import magopt.gui.view.summary.SummaryView;

public class CenterViewManager extends ViewManager<CenterViewItem> implements ParamListener {

	public CenterViewItem STRUCTURE = new CenterViewItem(CenterViewType.STRUCTURE, "Structure");
	public CenterViewItem DETAIL = new CenterViewItem(CenterViewType.DETAIL, "Details");
	public CenterViewItem FLAGS = new CenterViewItem(CenterViewType.FLAGS, "Flags");
	public CenterViewItem REPORT = new CenterViewItem(CenterViewType.REPORT, "Report");
	public CenterViewItem SUMMARY = new CenterViewItem(CenterViewType.SUMMARY, "Summary");
	public CenterViewItem COMPONENT = new CenterViewItem(CenterViewType.COMPONENT, "Component");
	public CenterViewItem IMPORT_EXPORT = new CenterViewItem(CenterViewType.IMPORT_EXPORT, "Import/Export");
	public CenterViewItem DOCFILES = new CenterViewItem(CenterViewType.DOCFILES, "DocFiles");
	public CenterViewItem PREVIEW = new CenterViewItem(CenterViewType.PREVIEW, "Preview");
	public CenterViewItem INDICES = new CenterViewItem(CenterViewType.INDICES, "Indices");
	public CenterViewItem EVENT = new CenterViewItem(CenterViewType.EVENT, "Log");
	public CenterViewItem CONSOLE = new CenterViewItem(CenterViewType.CONSOLE, "Console");
	public CenterViewItem PYTHON = new CenterViewItem(CenterViewType.PYTHON, "Python Console");

	StructureViewPanel struct;

	DetailView detail;
	FlagsView flags;
	ReportParameterView report;
	SummaryView summary;
	ComponentView comp;
	ImportExportView impExp;
	DocFilesView doc;
	PreviewView preview;

	IndicesViewPanel indices;

	EventView event;
	Console console;

	public PythonConsole pythonConsole;

	public CenterViewManager(CenterController controller) {
		super(controller, new CenterMenuBar(controller), new CenterToolBar(controller),
				new CenterStatusBar(controller));

		SplitView root, split;
		TabView tab;

		// create view-items
		this.viewItems.add(this.STRUCTURE);
		this.viewItems.add(this.DETAIL);
		this.viewItems.add(this.FLAGS);
		this.viewItems.add(this.REPORT);
		this.viewItems.add(this.SUMMARY);
		this.viewItems.add(this.COMPONENT);
		this.viewItems.add(this.IMPORT_EXPORT);
		this.viewItems.add(this.DOCFILES);
		this.viewItems.add(this.PREVIEW);
		this.viewItems.add(this.INDICES);
		this.viewItems.add(this.EVENT);
		this.viewItems.add(this.CONSOLE);
		this.viewItems.add(this.PYTHON);

		this.struct = new StructureViewPanel(this);

		this.detail = new DetailView();
		this.detail.addParamListener(this);
		this.flags = new FlagsView();
		this.flags.addParamListener(this);
		this.report = new ReportParameterView();
		this.report.addParamListener(this);
		this.summary = new SummaryView();
		this.summary.addParamListener(this);
		this.comp = new ComponentView();
		this.comp.addParamListener(this);
		this.impExp = new ImportExportView();
		this.impExp.addParamListener(this);
		this.doc = new DocFilesView();
		this.doc.addParamListener(this);
		this.preview = new PreviewView();
		this.preview.addParamListener(this);

		this.indices = new IndicesViewPanel(this);

		this.event = new EventView();
		this.console = new Console();
		System.setOut(this.console.getOutputPrintStream());
		System.setErr(this.console.getErrorPrintStream());
		try {
			this.pythonConsole = new PythonConsole();
		} catch (Throwable ex) {
		}

		root = new SplitView(this);
		this.views = root;
		root.setHorizontal(false);
		root.setDivider(0.65);

		split = new SplitView(this);
		split.setHorizontal(true);
		split.setDivider(0.4);
		root.setLeftView(split);

		tab = new TabView(this);
		split.setLeftView(tab);
		tab.addView(this.STRUCTURE);

		tab = new TabView(this);
		split.setRightView(tab);
		tab.addView(this.DETAIL);
		tab.addView(this.FLAGS);
		tab.addView(this.REPORT);
		tab.addView(this.SUMMARY);
		tab.addView(this.COMPONENT);
		tab.addView(this.IMPORT_EXPORT);
		tab.addView(this.DOCFILES);
		tab.addView(this.PREVIEW);

		split = new SplitView(this);
		split.setHorizontal(true);
		split.setDivider(0.3);
		root.setRightView(split);

		tab = new TabView(this);
		split.setLeftView(tab);
		tab.addView(this.INDICES);

		tab = new TabView(this);
		split.setRightView(tab);
		tab.addView(this.EVENT);
		tab.addView(this.CONSOLE);
		tab.addView(this.PYTHON);

		this.refresh();
	}

	@Override
	public CenterController getController() {
		return (CenterController) super.getController();
	}

	public void setController(CenterController controller) {
		super.setController(controller);
	}

	@Override
	public JComponent getComponent() {
		return this.views.getComponent();
	}

	@Override
	public void refresh(CenterViewItem item) {
		switch (item.getType()) {
		case COMPONENT:
			if (this.getController() != null) {
				this.comp.setContainer(this.getController().getSelected());
			} else {
				this.comp.setComponent(null);
			}
			break;
		case CONSOLE:
			// this.comp = this.console;
			break;
		case DETAIL:
			if (this.getController() != null) {
				this.detail.setParameter(this.getController().getSelected());
			} else {
				this.detail.setParameter(null);
			}
			break;
		case DOCFILES:
			if (this.getController() != null) {
				this.doc.setParameter(this.getController().getSelected());
			} else {
				this.doc.setParameter(null);
			}
			break;
		case EVENT:
			if (this.getController() != null) {
				// FIXME IEventLog
				if (this.getController().getProject() != null) {
					this.event.setEventLog(((Project) this.getController().getProject()).getEventLog());
				} else {
					this.event.setEventLog(null);
				}
			} else {
				this.event.setEventLog(null);
			}
			break;
		case FLAGS:
			if (this.getController() != null) {
				this.flags.setParameter(this.getController().getSelected());
			} else {
				this.flags.setParameter(null);
			}
			break;
		case HELP:
			if (this.getController() != null) {
				// FIXME
				// help.set
			} else {
				// FIXME
				// help.set
			}
			break;
		case IMPORT_EXPORT:
			if (this.getController() != null) {
				this.impExp.setFunction(this.getController().getSelected());
			} else {
				this.impExp.setFunction(null);
			}
			break;
		case INDICES:
			if (this.getController() != null) {
				// FIXME IProject
				this.indices.setProject((Project) this.getController().getProject());
			} else {
				this.indices.setProject(null);
			}
			break;
		case PREVIEW:
			if (this.getController() != null) {
				this.preview.setFunction(this.getController().getSelected());
			} else {
				this.preview.setFunction(null);
			}
			break;
		case PYTHON:
			// this.comp = this.pythonConsole;
			break;
		case REPORT:
			if (this.getController() != null) {
				this.report.setProject(this.getController().getSelected());
			} else {
				this.report.setProject(null);
			}
			break;
		case STRUCTURE:
			if (this.getController() != null) {
				// FIXME IProject
				this.struct.setProject((Project) this.getController().getProject());
				this.struct.refresh();
			} else {
				this.struct.setProject(null);
			}
			break;
		case SUMMARY:
			if (this.getController() != null) {
				this.summary.setContainer(this.getController().getSelected());
			} else {
				this.summary.setContainer(null);
			}
			break;
		}
	}

	@Override
	public void refresh() {
		String title;
		if (this.getController().getGuiComponent() instanceof JFrame) {
			title = "SyMSpace Center";
			if (this.getController().getProject() != null) {
				title += " - ";
				try {
					title += this.getController().getProject().getMopFile().getName() + " ("
							+ this.getController().getProject().getMopFile().getParent() + ")";
				} catch (RemoteException e) {
				}
			}
			((JFrame) this.getController().getGuiComponent()).setTitle(title);
		}

		super.refresh();
	}

	@Override
	public JComponent getComponent(CenterViewItem item) {
		JComponent comp = null;

		switch (item.getType()) {
		case COMPONENT:
			comp = new DisablePanel(this.comp);
			break;
		case CONSOLE:
			comp = this.console;
			break;
		case DETAIL:
			comp = new DisablePanel(this.detail);
			break;
		case DOCFILES:
			comp = new DisablePanel(this.doc);
			break;
		case EVENT:
			comp = this.event;
			break;
		case FLAGS:
			comp = new DisablePanel(this.flags);
			break;
		case INDICES:
			comp = new DisablePanel(this.indices);
			break;
		case IMPORT_EXPORT:
			comp = new DisablePanel(this.impExp);
			break;
		case PREVIEW:
			comp = new DisablePanel(this.preview);
			break;
		case PYTHON:
			comp = new DisablePanel(this.pythonConsole);
			break;
		case REPORT:
			comp = new DisablePanel(this.report);
			break;
		case STRUCTURE:
			comp = new DisablePanel(this.struct);
			break;
		case SUMMARY:
			comp = new DisablePanel(this.summary);
			break;
		}

		return comp;
	}

	@Override
	public void parameterChanged() {
		this.refresh();
	}
}
