/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.optimizer.views.compare;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.File;
import java.util.LinkedList;

import javax.swing.DefaultCellEditor;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import at.lcm.bu21.general.plot.HOLD;
import at.lcm.bu21.general.plot.LinkPlot;
import at.lcm.bu21.general.plot.SubPlot;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.core.Archive;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.gui.project.PanelPlots;
import jobmanager.server.Job;
import magopt.gui.helpers.JProjectChooser;
import magopt.utils.MopFile;

/**
 * Compare different Problem with the same objectives
 * 
 * @author Werner Koppelsttter
 * 
 */
public class PanelCompare extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Project of based compare-object */
	private Project project = null;

	/** Sub-plot for objective-plot */
	private SubPlot sub;

	private LinkPlot link;

	/** Table-model of settings */
	private CompareModel model;

	/** Settings-table of selected problems */
	private JTable table;

	/** Tabbed pane for settings and plot */
	private JTabbedPane tabbedPane;

	/** List of selected problems */
	private LinkedList<CompareObject> compare = new LinkedList<PanelCompare.CompareObject>();

	/** Based compare (for plot and validation) */
	CompareObject baseCompare = null;

	/**
	 * Connects column indices with column header names.
	 * 
	 * @author Werner Koppelsttter
	 */
	public enum CompareColumn {
		/** Optimization project */
		MOP(0, "Project", String.class),
		/** Optimization archive */
		ARCHIVE(1, "Archive", String.class),
		/** Optimization problem */
		PROBLEM(2, "Problem", String.class),
		/** Valid selection */
		VALID(3, "Valid", Boolean.class),
		/** Plot color */
		COLOR(4, "Color", ColorSelect.class),
		/** Plot order */
		ORDER(5, "Order", Integer.class),
		/** Plot base */
		BASE(6, "Base", Boolean.class);

		/** Column number */
		private final int column;

		/** Column name */
		private final String name;

		/** Column type */
		private final Class<?> type;

		/**
		 * Create enumeration
		 * 
		 * @param column
		 *            Column number
		 * @param name
		 *            Column name
		 * @param type
		 *            Column type
		 */
		private CompareColumn(int column, String name, Class<?> type) {
			this.column = column;
			this.name = name;
			this.type = type;
		}

		/**
		 * Get column number
		 * 
		 * @return Number
		 */
		protected int getValue() {
			return this.column;
		}

		/**
		 * Get column name
		 * 
		 * @return Name
		 */
		protected String getName() {
			return this.name;
		}

		/**
		 * Get column type
		 * 
		 * @return Type
		 */
		protected Class<?> getType() {
			return this.type;
		}

		/**
		 * Get element of enumeration with to the index
		 * 
		 * @param index
		 *            Index of column
		 * @return Enumeration element
		 */
		static CompareColumn getColumn(int index) {
			CompareColumn comp = null;

			for (CompareColumn c : CompareColumn.values()) {
				if (c.getValue() == index) {
					comp = c;
				}
			}

			return comp;
		}
	}

	/**
	 * Enumeration of possible colors
	 * 
	 * @author Werner Koppelsttter
	 */
	public enum ColorSelect {
		/** blue */
		BLUE("blue", Color.blue),
		/** green */
		GREEN("green", Color.green.darker()),
		/** cyan */
		CYAN("cyan", Color.cyan),
		/** magenta */
		MAGENTA("magenta", Color.magenta),
		/** yellow */
		YELLOW("yellow", Color.yellow.darker()),
		/** black */
		BLACK("black", Color.black),
		/** red */
		RED("red", Color.red);

		/** Name of color */
		private final String name;

		/** Color value */
		private final Color color;

		/**
		 * Create color enumeration
		 * 
		 * @param name
		 *            Name of color
		 * @param color
		 *            Color object
		 */
		private ColorSelect(String name, Color color) {
			this.name = name;
			this.color = color;
		}

		/**
		 * Get name of color
		 * 
		 * @return Name
		 */
		protected String getName() {
			return this.name;
		}

		/**
		 * Get color objext
		 * 
		 * @return Color
		 */
		protected Color getColor() {
			return this.color;
		}

		@Override
		public String toString() {
			return this.name;
		}
	}

	/**
	 * Compare object with selected problem
	 * 
	 * @author Werner Koppelsttter
	 */
	private class CompareObject {

		/** Mop-file */
		private MopFile mop = null;

		/** Archive */
		private Archive arch = null;

		/** Problem */
		private Problem prob = null;

		/** Selected color */
		private ColorSelect color = null;

		/**
		 * Get selected mop-file
		 * 
		 * @return Mop-file
		 */
		public MopFile getMopFile() {
			MopFile mop;
			Archive arch;

			arch = this.getArchive();
			if (arch != null) {
				mop = arch.getMop();
			} else {
				mop = this.mop;
			}

			return mop;
		}

		/**
		 * Select mop-file
		 * 
		 * @param mop
		 *            Mop-file
		 */
		public void setMopFile(MopFile mop) {
			if ((this.mop == null) || (mop == null)
					|| ((this.mop != mop) && (this.mop.getAbsolutePath().compareTo(mop.getAbsolutePath()) != 0))) {
				// reset archive and problem
				this.arch = null;
				this.prob = null;
			}
			this.mop = mop;
		}

		/**
		 * Get selected archive
		 * 
		 * @return Archive
		 */
		public Archive getArchive() {
			Archive arch;
			Problem prob;

			prob = this.getProblem();
			if (prob != null) {
				arch = prob.getArchive();
			} else {
				arch = this.arch;
			}

			return arch;
		}

		/**
		 * Select archive
		 * 
		 * @param arch
		 *            Archive
		 */
		public void setArchive(Archive arch) {
			if (this.arch != arch) {
				this.arch = arch;
				// reset problem
				this.prob = null;
			}
		}

		/**
		 * Get selected problem
		 * 
		 * @return Problem
		 */
		public Problem getProblem() {
			return this.prob;
		}

		/**
		 * Select problem
		 * 
		 * @param prob
		 *            Problem
		 */
		public void setProblem(Problem prob) {
			this.prob = prob;
		}

		/**
		 * Get selected color
		 * 
		 * @return Color-enumeration
		 */
		public ColorSelect getColor() {
			return this.color;
		}

		/**
		 * Select color
		 * 
		 * @param color
		 *            Color-enumeration
		 */
		public void setColor(ColorSelect color) {
			this.color = color;
		}

		/**
		 * Is selected problem valid. Check if there are equal objectives
		 * 
		 * @param prob
		 *            Problem to check
		 * @return Is valid
		 */
		public boolean isValid(Problem prob) {
			boolean valid = false, val;
			ResultDesc[] res1, res2;
			String name1, name2;

			if (this.getProblem() != null) {
				// check for equal objectives
				if (prob.getNumSavePoints() > 0) {
					if (prob.getNumObjectives() == this.getProblem().getNumObjectives()) {
						valid = true;
						res1 = prob.getObjectives();
						res2 = this.getProblem().getObjectives();
						for (int k1 = 0; k1 < res1.length; k1++) {
							name1 = res1[k1].getName();
							name1 = name1.substring(name1.indexOf('.'));
							val = false;
							for (int k2 = 0; k2 < res2.length; k2++) {
								name2 = res2[k2].getName();
								name2 = name2.substring(name2.indexOf('.'));
								val |= name1.compareTo(name2) == 0;
							}
							valid &= val;
						}
					}
				}
			}

			return valid;
		}
	}

	/**
	 * Table model for compare setting table.
	 * 
	 * @author Werner Koppelsttter
	 */
	public class CompareModel extends AbstractTableModel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getColumnCount() {
			return CompareColumn.values().length;
		}

		@Override
		public int getRowCount() {
			return PanelCompare.this.compare.size() + 1;
		}

		@Override
		public Object getValueAt(int row, int column) {
			CompareObject comp;
			Object obj = null;

			if (row == (this.getRowCount() - 1)) {
				comp = new CompareObject();
				comp.setColor(ColorSelect.values()[row % ColorSelect.values().length]);
			} else {
				comp = PanelCompare.this.compare.get(row);
			}

			if (column == CompareColumn.MOP.getValue()) {
				if (comp.getMopFile() != null) {
					obj = comp.getMopFile().getName();
				}
			} else if (column == CompareColumn.ARCHIVE.getValue()) {
				if (comp.getArchive() != null) {
					obj = comp.getArchive().getName();
				}
			} else if (column == CompareColumn.PROBLEM.getValue()) {
				if (comp.getProblem() != null) {
					obj = comp.getProblem().getName();
				}
			} else if (column == CompareColumn.COLOR.getValue()) {
				obj = comp.getColor();
			} else if (column == CompareColumn.VALID.getValue()) {
				if (!PanelCompare.this.compare.isEmpty()) {
					obj = comp.isValid(PanelCompare.this.compare.getFirst().getProblem());
				} else {
					obj = false;
				}
			} else if (column == CompareColumn.ORDER.getValue()) {
				obj = row + 1;
			} else if (column == CompareColumn.BASE.getValue()) {
				obj = comp == PanelCompare.this.baseCompare;
			} else {
				obj = comp;
			}

			return obj;
		}

		@Override
		public String getColumnName(int column) {
			for (CompareColumn e : CompareColumn.values()) {
				if (e.getValue() == column) {
					return e.getName();
				}
			}
			return null;
		}

		@Override
		public boolean isCellEditable(int row, int column) {
			boolean edit = false;

			// if (row > 0) {
			if (column == CompareColumn.MOP.getValue()) {
				edit = true;
			} else if (column == CompareColumn.ARCHIVE.getValue()) {
				edit = PanelCompare.this.model.getValueAt(row, CompareColumn.MOP.getValue()) != null;
			} else if (column == CompareColumn.PROBLEM.getValue()) {
				edit = PanelCompare.this.model.getValueAt(row, CompareColumn.ARCHIVE.getValue()) != null;
			} else if (column == CompareColumn.COLOR.getValue()) {
				edit = true;
			} else if (column == CompareColumn.ORDER.getValue()) {
				edit = true;
			} else if (column == CompareColumn.BASE.getValue()) {
				edit = true;
			}
			// }

			if ((row >= PanelCompare.this.compare.size()) && (column != CompareColumn.MOP.getValue())) {
				edit = false;
			}

			return edit;
		}

		@Override
		public Class<?> getColumnClass(int column) {
			return CompareColumn.getColumn(column).getType();
		}

		@Override
		public void setValueAt(Object value, int row, int col) {
			CompareObject comp;
			File path = null;
			int val;

			if (row == (this.getRowCount() - 1)) {
				PanelCompare.this.compare.add((CompareObject) this.getValueAt(row, -1));
			}
			comp = PanelCompare.this.compare.get(row);

			if (col == CompareColumn.MOP.getValue()) {
				if (value == null) {
					PanelCompare.this.removeCompareObject(comp);
				} else if (((String) value).compareTo("...") == 0) {
					if (PanelCompare.this.baseCompare != null) {
						path = PanelCompare.this.baseCompare.getMopFile().getParentFile();
					}
					JProjectChooser chooser = new JProjectChooser(JProjectChooser.PROJECT_ONLY, path);
					int returnVal = chooser.showOpenDialog(PanelCompare.this);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						comp.setMopFile(new MopFile(chooser.getSelectedFile()));
						if (PanelCompare.this.baseCompare == null) {
							PanelCompare.this.baseCompare = comp;
						}
						PanelCompare.this.refresh();
					}
				}
			} else if (col == CompareColumn.ARCHIVE.getValue()) {
				try {
					if (value != null) {
						comp.setArchive(Archive.load(comp.getMopFile(), (String) value));
					} else {
						comp.setArchive(null);
					}
					PanelCompare.this.refresh();
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else if (col == CompareColumn.PROBLEM.getValue()) {
				try {
					if (value != null) {
						comp.setProblem(comp.getArchive().readProblem((String) value));
					} else {
						comp.setProblem(null);
					}
					PanelCompare.this.refresh();
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else if (col == CompareColumn.COLOR.getValue()) {
				comp.setColor((ColorSelect) value);
				PanelCompare.this.refresh();
			} else if (col == CompareColumn.ORDER.getValue()) {
				val = (Integer) value;
				if ((val > 0) && (val <= PanelCompare.this.compare.size())) {
					PanelCompare.this.compare.remove(comp);
					PanelCompare.this.compare.add(val - 1, comp);
					PanelCompare.this.refresh();
				}
			} else if (col == CompareColumn.BASE.getValue()) {
				if ((Boolean) value) {
					PanelCompare.this.baseCompare = comp;
					PanelCompare.this.refresh();
				}
			}
		}
	}

	/**
	 * Create a compare panel
	 */
	public PanelCompare() {
		JPanel plots;
		JScrollPane scroll;

		this.setLayout(new BorderLayout(0, 0));

		this.tabbedPane = new JTabbedPane(SwingConstants.TOP);
		this.add(this.tabbedPane);

		scroll = new JScrollPane();
		this.tabbedPane.addTab("Settings", null, scroll, null);

		this.model = new CompareModel();

		this.table = new JTable(this.model);
		scroll.setViewportView(this.table);

		this.table.getTableHeader().setReorderingAllowed(false);
		this.table.setDefaultEditor(String.class, new DefaultCellEditor(new JComboBox()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row,
					int column) {
				DefaultComboBoxModel model = ((DefaultComboBoxModel) ((JComboBox) this.getComponent()).getModel());
				model.removeAllElements();
				model.addElement(null);
				if (column == CompareColumn.MOP.getValue()) {
					if (value != null) {
						model.addElement(value);
					}
					model.addElement("...");
				} else if (column == CompareColumn.ARCHIVE.getValue()) {
					MopFile mop = ((CompareObject) PanelCompare.this.model.getValueAt(row, -1)).getMopFile();
					if (mop != null) {
						try {
							for (String name : Archive.listArchives(mop)) {
								model.addElement(name);
							}
						} catch (Exception e) {
						}
					}
				} else if (column == CompareColumn.PROBLEM.getValue()) {
					Archive arch = ((CompareObject) PanelCompare.this.model.getValueAt(row, -1)).getArchive();
					if (arch != null) {
						for (String name : arch.listProblems()) {
							model.addElement(name);
						}
					}
				}
				return super.getTableCellEditorComponent(table, value, isSelected, row, column);
			}

		});
		this.table.setDefaultRenderer(String.class, new DefaultTableCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
					boolean hasFocus, int row, int column) {

				super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
				if (column == CompareColumn.MOP.getValue()) {
					MopFile mop = ((CompareObject) PanelCompare.this.model.getValueAt(row, -1)).getMopFile();
					if (mop != null) {
						this.setText(mop.getName());
						this.setToolTipText(mop.getPath());
					} else {
						this.setText(null);
						this.setToolTipText(null);
					}
				} else {
					this.setToolTipText(this.getText());
				}

				return this;
			}

		});
		this.table.setDefaultEditor(ColorSelect.class, new DefaultCellEditor(new JComboBox(ColorSelect.values())));
		this.table.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				Job.refreshList();
				try {
					switch (e.getKeyCode()) {
					case KeyEvent.VK_DELETE:
						int row = PanelCompare.this.table.getSelectedRow();
						if (row > 0) {
							PanelCompare.this.removeCompareObject(PanelCompare.this.compare.get(row));
						}
						break;
					}
				} catch (Exception e1) {
					// e1.printStackTrace();
				}
			}
		});

		// min column width
		this.table.getColumnModel().getColumn(CompareColumn.MOP.getValue()).setMinWidth(50);
		this.table.getColumnModel().getColumn(CompareColumn.ARCHIVE.getValue()).setMinWidth(50);
		this.table.getColumnModel().getColumn(CompareColumn.PROBLEM.getValue()).setMinWidth(50);
		this.table.getColumnModel().getColumn(CompareColumn.VALID.getValue()).setMinWidth(40);
		this.table.getColumnModel().getColumn(CompareColumn.COLOR.getValue()).setMinWidth(50);
		this.table.getColumnModel().getColumn(CompareColumn.ORDER.getValue()).setMinWidth(40);
		this.table.getColumnModel().getColumn(CompareColumn.BASE.getValue()).setMinWidth(40);
		// preferred column width
		this.table.getColumnModel().getColumn(CompareColumn.VALID.getValue()).setMaxWidth(60);
		this.table.getColumnModel().getColumn(CompareColumn.COLOR.getValue()).setMaxWidth(80);
		this.table.getColumnModel().getColumn(CompareColumn.ORDER.getValue()).setMaxWidth(50);
		this.table.getColumnModel().getColumn(CompareColumn.BASE.getValue()).setMaxWidth(50);

		plots = new JPanel();
		plots.setLayout(new BorderLayout());
		this.sub = new SubPlot(1, 1);
		plots.add(this.sub.getPanel(), BorderLayout.CENTER);
		this.tabbedPane.addTab("Plots", null, plots, null);

		// prob = this._views.getController().getProblem();
		// try {
		// PanelPlots.initObjectives(this._sub, prob,
		// this._views.getController().getBaseProject());
		// this.refreshPlot();
		// } catch (Exception ex) {
		// ex.printStackTrace();
		// }
		// for (int r = 0; r < this._sub.getSize()[0]; r++) {
		// for (int c = 0; c < this._sub.getSize()[1]; c++) {
		// if (this._sub.getPlot(r, c) != null) {
		// this._sub.getPlot(r, c).setHold(HOLD.ALL);
		// }
		// }
		// }
		this.link = new LinkPlot();

		this.tabbedPane.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent event) {
				PanelCompare.this.refresh();
			}
		});
	}

	/**
	 * Refresh panel
	 */
	private void refresh() {
		this.model.fireTableDataChanged();
		if (this.tabbedPane.getSelectedIndex() > 0) {
			if (this.getProblem() != null) {
				try {
					PanelPlots.initObjectives(this.sub, this.getProblem());
					this.link.rmAllPlots();
					this.link.addPlot(this.sub);
				} catch (Exception ex) {
					ex.printStackTrace();
				}
				for (int r = 0; r < this.sub.getSize()[0]; r++) {
					for (int c = 0; c < this.sub.getSize()[1]; c++) {
						if (this.sub.getPlot(r, c) != null) {
							this.sub.getPlot(r, c).setHold(HOLD.ALL);
						}
					}
				}
				PanelCompare.this.refreshPlot();
			} else {
				this.sub.setSize(1, 0);
			}
		}
	}

	/**
	 * Refresh plots
	 */
	private void refreshPlot() {
		LinkedList<Individual[]> paretos;
		LinkedList<Color> colors;
		Problem baseProb, prob;

		if (this.baseCompare != null) {
			baseProb = this.baseCompare.getProblem();
			if (baseProb != null) {
				paretos = new LinkedList<Individual[]>();
				colors = new LinkedList<Color>();

				// get pareto-fronts
				for (CompareObject comp : PanelCompare.this.compare) {
					if (comp.isValid(baseProb)) {
						prob = comp.getProblem();
						colors.add(comp.getColor().getColor());
						if (prob.getNumSavePoints() > 0) {
							paretos.add(prob.getSavePoint(prob.getNumSavePoints() - 1).getFullPareto());
						} else {
							paretos.add(new Individual[0]);
						}
					}
				}

				// plot paretos
				try {
					// set color-order
					Color[] col = colors.toArray(new Color[0]);
					for (int r = 0; r < this.sub.getSize()[0]; r++) {
						for (int c = 0; c < this.sub.getSize()[1]; c++) {
							if (this.sub.getPlot(r, c) != null) {
								this.sub.getPlot(r, c).setColorOrder(col);
							}
						}
					}
					// plot
					PanelPlots.drawObjectives(this.sub, baseProb, paretos, -1, false);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * Get number of compare-objects
	 * 
	 * @return Size
	 */
	public int getCompareObjectSize() {
		return this.compare.size();
	}

	/**
	 * Get compare-object for index
	 * 
	 * @param index
	 *            Indes in list
	 * @return Compare-object
	 */
	public CompareObject getCompareObject(int index) {
		CompareObject comp = null;

		if ((index >= 0) && (index < this.compare.size())) {
			comp = this.compare.get(index);
		}

		return comp;
	}

	/**
	 * Add problem as compare-object if not exists
	 * 
	 * @param prob
	 *            Problem
	 */
	public void addCompareObject(Problem prob) {
		boolean exist = false;
		CompareObject comp;
		MopFile mop;

		if (prob != null) {
			mop = prob.getArchive().getMop();
			for (CompareObject cmp : this.compare) {
				if (cmp.getMopFile().equals(mop)) {
					if ((cmp.getProblem() != null) && prob.getArchive().getName().equals(cmp.getArchive().getName())
							&& prob.getName().equals(cmp.getProblem().getName())) {
						exist = true;
					}
				}
			}

			if (!exist) {
				comp = new CompareObject();
				comp.setColor(ColorSelect.values()[this.compare.size() % ColorSelect.values().length]);
				comp.setProblem(prob);
				this.compare.add(comp);
				if (this.baseCompare == null) {
					this.baseCompare = comp;
				}
				this.refresh();
			}
		}

	}

	/**
	 * Add mop as compare-object if not exists
	 * 
	 * @param mop
	 *            Mop-file
	 */
	public void addCompareObject(MopFile mop) {
		boolean exist = false;
		CompareObject comp;
		if (mop != null) {
			for (CompareObject cmp : this.compare) {
				if (cmp.getMopFile().equals(mop)) {
					if (cmp.getProblem() == null) {
						exist = true;
					}
				}
			}

			if (!exist) {
				comp = new CompareObject();
				comp.setColor(ColorSelect.values()[this.compare.size() % ColorSelect.values().length]);
				comp.setMopFile(mop);
				this.compare.add(comp);
				if (this.baseCompare == null) {
					this.baseCompare = comp;
				}
				this.refresh();
			}
		}

		// Project baseProj = null;

		// FIXME
		// if ((prob == null) || !prob.equals(this._problem) || (baseProj ==
		// null)
		// || baseProj.equals(this._project)) {
		// this._problem = prob;
		// this._project = baseProj;
		//
		// if (this._problem != null) {
		// if (this.openObj == null) {
		// this.openObj = new CompareObject();
		// this._compare.addFirst(this.openObj);
		// this.openObj.setColor(ColorSelect.BLUE);
		// }
		// this.openObj.setProblem(this._problem);
		//
		// } else {
		// if (this.openObj != null) {
		// this._compare.remove(this.openObj);
		// this.openObj = null;
		// }
		// }
		//
		// this.refresh();
		// }

	}

	/**
	 * Remove compare-object
	 * 
	 * @param index
	 *            Index of compare-object
	 */
	public void removeCompareCompareObject(int index) {
		this.removeCompareObject(this.getCompareObject(index));
	}

	/**
	 * Remove compare-object
	 * 
	 * @param comp
	 *            Compare-object
	 */
	public void removeCompareObject(CompareObject comp) {
		if (comp != null) {
			this.compare.remove(comp);
			if (this.baseCompare == comp) {
				if (!this.compare.isEmpty()) {
					this.baseCompare = this.compare.getFirst();
				} else {
					this.baseCompare = null;
				}
			}
			this.refresh();
		}
	}

	/**
	 * Get based compare-object
	 * 
	 * @return Compare-object
	 */
	public CompareObject getBaseCompareObject() {
		return this.baseCompare;
	}

	/**
	 * Set based compare-object
	 * 
	 * @param index
	 *            Index of compare
	 */
	public void setBaseCompareObject(int index) {
		this.setBaseCompareObject(this.getCompareObject(index));
	}

	/**
	 * Set based compare-object
	 * 
	 * @param comp
	 *            Compare-object
	 */
	public void setBaseCompareObject(CompareObject comp) {
		if (comp != null) {
			this.baseCompare = comp;
			this.refresh();
		}
	}

	/**
	 * Get problem of base compare-object
	 * 
	 * @return Problem
	 */
	private Problem getProblem() {
		Problem prob = null;
		if (this.baseCompare != null) {
			prob = this.baseCompare.getProblem();
		}

		return prob;
	}

	/**
	 * Get project of base compare-object
	 * 
	 * @return Project
	 */
	private Project getProject() {
		Project proj = null;

		if (this.baseCompare != null) {
			if ((this.project == null) || !this.baseCompare.getMopFile().equals(this.project.getMopFile())) {
				try {
					this.project = Project.load(this.baseCompare.getMopFile());
				} catch (Exception e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				}
			}
			proj = this.project;
		} else {
			proj = null;
		}

		return proj;
	}
}
