/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.optimizer.views.pool;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;

import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.project.ComputeProject;

/**
 * Overview over the project pool (Individuals)
 * 
 * @author Gnther Weidenholzer
 */
public class PoolViewPanel extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Number of columns */
	static final int COLUMNS = 15;

	/** Block size */
	static final int BLOCKSIZE = 15;

	/** Color for complete individual */
	static final Color COLOR_OK = new Color(51, 255, 51);

	/** Color for incomplete individual */
	static final Color COLOR_INCOMPLETE = new Color(255, 204, 0);

	/** Color for individual with error */
	static final Color COLOR_ERROR = new Color(255, 0, 0);

	/** Color for invalid individual */
	static final Color COLOR_INVALID = new Color(255, 0, 255);

	/** Color for free individual */
	static final Color COLOR_CREATE = new Color(153, 255, 255);

	/** Color for unused individual */
	static final Color COLOR_UNUSED = new Color(128, 128, 128);

	/** Label for free color */
	private JLabel jLabelNew = new JLabel("New", SwingConstants.CENTER);

	/** Label for complete color */
	private JLabel jLabelOk = new JLabel("OK", SwingConstants.CENTER);

	/** Label for incomplete color */
	private JLabel jLabelIncomplete = new JLabel("In progress",
			SwingConstants.CENTER);

	/** Label for incomplete color */
	private JLabel jLabelInvalid = new JLabel("Invalid", SwingConstants.CENTER);

	/** Label for error color */
	private JLabel jLabelError = new JLabel("Error", SwingConstants.CENTER);

	/** Label for unused color */
	private JLabel jLabelUnused = new JLabel("Unused", SwingConstants.CENTER);

	/** Panel for legend */
	private JPanel jPanelLegend = new JPanel();

	/** Scroll panel for table */
	private JScrollPane jScrollPane1 = new JScrollPane();

	/** Table to display individuals */
	private JTable jTable1 = new JTable();

	/** Project optimization object */
	protected ComputeProject calc = null;

	/**
	 * Set pool to display
	 * 
	 * @param calc
	 *            Calculation object
	 */
	public void set(ComputeProject calc) {

		this.calc = calc;

		this.refresh();
	}

	/**
	 * Create pool overview panel
	 */
	public PoolViewPanel() {

		this.setLayout(new BorderLayout(0, 5));
		this.setBorder(new EmptyBorder(5, 5, 5, 5));

		this.jTable1.setModel(
				new DefaultTableModel(new Object[1][PoolViewPanel.COLUMNS],
						new String[PoolViewPanel.COLUMNS]) {

					/** Serialization version number */
					private static final long serialVersionUID = 1L;

					@Override
					public boolean isCellEditable(int rowIndex,
							int columnIndex) {
						return false;
					}
				});

		this.jTable1.setCellSelectionEnabled(true);
		this.jTable1.setFocusable(false);
		this.jTable1.addMouseListener(new java.awt.event.MouseAdapter() {

			@Override
			public void mouseClicked(java.awt.event.MouseEvent evt) {
				PoolViewPanel.this.jTable1MouseClicked(evt);
			}
		});

		this.jScrollPane1.setViewportView(this.jTable1);
		this.jScrollPane1.getViewport().setBackground(Color.white);
		Dimension dim1 = new Dimension(
				PoolViewPanel.COLUMNS * PoolViewPanel.BLOCKSIZE,
				PoolViewPanel.COLUMNS * PoolViewPanel.BLOCKSIZE);
		this.setPreferredSize(dim1);

		this.jTable1.setDefaultRenderer(Object.class, new MyCellRenderer());
		this.jTable1.setShowGrid(false);

		this.jTable1.setTableHeader(null);

		this.jScrollPane1.setColumnHeaderView(null);
		this.jTable1.setRowHeight(PoolViewPanel.BLOCKSIZE);
		for (int k = 0; k < PoolViewPanel.COLUMNS; k++) {
			this.jTable1.getColumnModel().getColumn(k).setMinWidth(5);
		}

		this.add(this.jScrollPane1, BorderLayout.CENTER);

		this.jPanelLegend.setLayout(new GridLayout(2, 3, 5, 5));

		this.jLabelNew.setBackground(PoolViewPanel.COLOR_CREATE);
		this.jLabelNew.setOpaque(true);
		this.jPanelLegend.add(this.jLabelNew);

		this.jLabelOk.setBackground(PoolViewPanel.COLOR_OK);
		this.jLabelOk.setOpaque(true);
		this.jPanelLegend.add(this.jLabelOk);

		this.jLabelIncomplete.setBackground(PoolViewPanel.COLOR_INCOMPLETE);
		this.jLabelIncomplete.setOpaque(true);
		this.jPanelLegend.add(this.jLabelIncomplete);

		this.jLabelInvalid.setBackground(PoolViewPanel.COLOR_INVALID);
		this.jLabelInvalid.setOpaque(true);
		this.jPanelLegend.add(this.jLabelInvalid);

		this.jLabelError.setBackground(PoolViewPanel.COLOR_ERROR);
		this.jLabelError.setOpaque(true);
		this.jPanelLegend.add(this.jLabelError);

		this.jLabelUnused.setBackground(PoolViewPanel.COLOR_UNUSED);
		this.jLabelUnused.setOpaque(true);
		this.jPanelLegend.add(this.jLabelUnused);

		this.add(this.jPanelLegend, BorderLayout.SOUTH);

	}

	/**
	 * Mouse event for selecting individual
	 * 
	 * @param evt
	 *            Mouse event
	 */
	private void jTable1MouseClicked(java.awt.event.MouseEvent evt) {

		JTable target = (JTable) evt.getSource();
		// int row = target.getSelectedRow();
		// int column = target.getSelectedColumn();
		//
		// int idx = (row * PanelProjectOverview.COLUMNS) + column;
		target.clearSelection();

		// TODO event fr datensatzwechsel erstellen
		// System.out.println("Mouseclick on #" + Integer.toString(idx + 1));

	}

	/**
	 * Refresh display
	 */
	public void refresh() {
		int rows = 0;

		if (this.calc != null) {
			rows = 1 + (this.calc.getProject().getLength()
					/ PoolViewPanel.COLUMNS);
		}

		DefaultTableModel mdl = (DefaultTableModel) (this.jTable1.getModel());
		while (mdl.getRowCount() > rows) {
			mdl.removeRow(mdl.getRowCount() - 1);
		}
		while (mdl.getRowCount() < rows) {
			mdl.addRow((Object[]) null);
		}

		((DefaultTableModel) this.jTable1.getModel()).fireTableDataChanged();
	}

	/**
	 * Table cell renderer to display the individuals
	 * 
	 * @author Gnther Weidenholzer
	 * @author Werner Koppelsttter
	 */
	class MyCellRenderer implements TableCellRenderer {

		@Override
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) {
			int idx, len = 0, selected;
			Color color = Color.white;
			boolean[] used;
			Individual[] ind;
			Problem problem;

			Component renderer = new DefaultTableCellRenderer()
					.getTableCellRendererComponent(table, value, isSelected,
							hasFocus, row, column);

			((JLabel) renderer).setOpaque(true);

			idx = (row * PoolViewPanel.COLUMNS) + column;
			if (PoolViewPanel.this.calc != null) {
				len = PoolViewPanel.this.calc.getProject().getLength();
			}
			if (len > 0) {
				selected = PoolViewPanel.this.calc.getProject().getIndex();
				if (idx < len) {
					used = PoolViewPanel.this.calc.getUsed();
					ind = PoolViewPanel.this.calc.getInd();
					problem = PoolViewPanel.this.calc.getProblem();

					((JLabel) renderer).setBorder(
							BorderFactory.createLineBorder(Color.BLACK, 1));

					if ((used != null) && (idx < used.length) && used[idx]) {
						color = PoolViewPanel.COLOR_INCOMPLETE;
					} else if ((ind != null) && (idx < ind.length)) {
						if (ind[idx] == null) {
							color = PoolViewPanel.COLOR_CREATE;
						} else if (ind[idx].isError()) {
							color = PoolViewPanel.COLOR_ERROR;
						} else if (ind[idx].isValid(problem)) {
							color = PoolViewPanel.COLOR_OK;
						} else {
							color = PoolViewPanel.COLOR_INVALID;
						}
					} else {
						color = PoolViewPanel.COLOR_UNUSED;
					}
				}

				// mark current dataset
				if (idx == selected) {
					((JLabel) renderer).setBorder(
							BorderFactory.createLineBorder(Color.BLACK, 2));
				}

			}

			renderer.setBackground(color);

			return renderer;
		}
	}
}
