/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */

package at.lcm.symspace.optimizer.views.settings;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;

import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.optim.algorithm.Solver;
import at.lcm.bu21.optim.algorithm.SolverList;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.gui.setting.ChooseProblem;
import at.lcm.bu21.optim.gui.setting.ChooseResults;
import at.lcm.bu21.optim.gui.setting.ChooseVariables;
import at.lcm.bu21.optim.gui.setting.GeneralSettings;
import at.lcm.bu21.optim.gui.setting.PanelRefresh;
import at.lcm.bu21.optim.gui.setting.SettingsPanel;
import at.lcm.bu21.optim.gui.setting.SolverSettings;
import at.lcm.symspace.optimizer.icons.opt.OptIcons;
import magopt.utils.MopFile;

/**
 * Define settings for optimizer
 *
 * @author Werner Koppelsttter
 */
public class SettingsDialog extends JFrame implements PanelRefresh {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Are settings finished */
	private boolean finished = false;

	/**
	 * State of settings dialog
	 *
	 * @author Werner Koppelsttter
	 */
	private enum State {
		/** Selcect probelem */
		PROBLEM("Define problem"),
		/** Select variables */
		VARIABLES("Select variables"),
		/** Select results */
		RESULTS("Select results"),
		/** General settings */
		GENERAL("General settings"),
		/** Solver settings */
		SOLVER("Solver settings");

		/** Description of state */
		private String desc;

		/**
		 * Initialize state
		 * 
		 * @param desc
		 *            Description of state
		 */
		State(String desc) {
			this.desc = desc;
		}

		/**
		 * Get description of state
		 * 
		 * @return Description
		 */
		public String getDescription() {
			return this.desc;
		}
	}

	/** Current state */
	private State state = State.PROBLEM;

	/** Label of state */
	private JLabel stateLabel;

	/** Panel for detail panels */
	private JPanel settings;

	/** Current detail panel */
	private SettingsPanel settingsPanel;

	/** Panel with controls and displays */
	private JPanel bottom;
	/** Message label */
	private JLabel message;

	/** Buttons panel */
	private JPanel buttons;
	/** Button for previous state */
	private JButton back;
	/** Button for next state */
	private JButton next;
	/** Button to finish */
	private JButton finish;
	/** Button to cancel */
	private JButton cancel;

	/** Current problem */
	private Problem problem = null;
	/** Initial working directory */
	private File initPath = null;

	private String solverName = null;
	private TStruct solverState = null;

	/**
	 * Initialize dialog
	 */
	private void init() {

		this.setTitle("Optimizer Settings");
		this.setLayout(new BorderLayout());
		this.setAlwaysOnTop(true);

		this.stateLabel = new JLabel();
		this.add(this.stateLabel, BorderLayout.NORTH);
		this.stateLabel.setBorder(new EmptyBorder(10, 10, 10, 10));
		this.stateLabel.setFont(
				new Font(this.stateLabel.getFont().getFontName(), Font.BOLD, this.stateLabel.getFont().getSize()));

		this.settings = new JPanel(new BorderLayout());
		this.add(this.settings, BorderLayout.CENTER);

		this.bottom = new JPanel(new GridLayout(2, 1));
		this.add(this.bottom, BorderLayout.SOUTH);
		this.bottom.setBorder(new EmptyBorder(0, 10, 5, 10));

		this.message = new JLabel("");
		this.bottom.add(this.message);
		this.message
				.setFont(new Font(this.message.getFont().getFontName(), Font.BOLD, this.message.getFont().getSize()));
		this.message.setForeground(Color.BLUE);

		this.buttons = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		this.bottom.add(this.buttons);

		this.back = new JButton("< Back");
		this.buttons.add(this.back);
		this.back.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				SettingsDialog.this.backState();
			}
		});

		this.next = new JButton("Next >");
		this.buttons.add(this.next);
		this.next.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				SettingsDialog.this.nextState();
			}
		});

		this.finish = new JButton("Finish");
		this.buttons.add(this.finish);
		this.finish.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent arg0) {
				SettingsDialog.this.setVisible(false);
				new Thread() {
					@Override
					public void run() {
						SettingsDialog.this.finish();
						SettingsDialog.this.dispose();
					}
				}.start();
			}
		});

		this.cancel = new JButton("Cancel");
		this.buttons.add(this.cancel);
		this.cancel.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				SettingsDialog.this.setVisible(false);
				SettingsDialog.this.dispose();

			}
		});
	}

	@Override
	public void refresh() {
		if (this.state == State.PROBLEM) {
			if ((this.settingsPanel != null) && (this.settingsPanel instanceof ChooseProblem)) {
				ChooseProblem prob = (ChooseProblem) this.settingsPanel;
				this.problem = prob.getProblem();
			}
		}
		this.refreshPanel();
	}

	/**
	 * Refresh panel
	 */
	private void refreshPanel() {
		this.stateLabel.setText(this.state.getDescription());
		this.back.setEnabled(this.isBackEnable());
		this.next.setEnabled(this.isNextEnable());
		this.finish.setEnabled(this.isFinishedEnable());
	}

	/**
	 * Is finish allowed
	 * 
	 * @return Enable finish
	 */
	private boolean isFinishedEnable() {
		boolean enable;

		enable = this.problem != null;
		if (enable) {
			enable &= ChooseProblem.validate(this.problem) == null;
			enable &= ChooseVariables.validate(this.problem) == null;
			enable &= ChooseResults.validate(this.problem) == null;
			enable &= GeneralSettings.validate(this.problem) == null;
			enable &= SolverSettings.validate(this.problem.getSolver()) == null;
		}

		return enable;
	}

	/**
	 * Is next state allowed
	 * 
	 * @return Enable next
	 */
	private boolean isNextEnable() {
		boolean enable = false;
		String validString;

		enable = this.state != State.SOLVER;
		if (enable && (this.settingsPanel != null)) {
			validString = this.settingsPanel.validateContents();
			this.message.setText(validString);
			enable = validString == null;
		}

		return enable;
	}

	/**
	 * Is previous state allowed
	 * 
	 * @return Enable back
	 */
	private boolean isBackEnable() {
		return this.state != State.PROBLEM;
	}

	/**
	 * Finish selection
	 */
	private void finish() {
		Solver solver, sol;
		switch (this.state) {
		case PROBLEM:
		case VARIABLES:
		case RESULTS:
		case GENERAL:
		case SOLVER:
			try {
				solver = this.problem.getSolver();
				// recreate solver
				sol = SolverList.getSolverListEntry(solver.getOptimizerName()).getSolver(this.problem);
				sol.setSettings(solver.getSettings());
				if ((this.solverState != null) && this.solverName.equals(sol.getOptimizerName())) {
					sol.setState(this.solverState);
				}
//				sol.setState(solver.getState());
				this.problem.setSolver(sol);

				if (this.problem.getNumSavePoints() == 0) {
					if (this.problem.getInitSolver()) {
						sol.setArchive(this.problem.getArchive().getMemory().getValidIndividual(this.problem));
					} else {
						sol.setArchive(new Individual[0]);
					}
				}

				// Save problem
				this.problem.getArchive().save();
				SettingsDialog.this.finished = true;
			} catch (Exception ex) {
				JOptionPane.showConfirmDialog(this, ex.getMessage(), "Finish wizard", JOptionPane.DEFAULT_OPTION,
						JOptionPane.ERROR_MESSAGE);

				ex.printStackTrace();
			}
			break;
		}
	}

	/**
	 * Change to next state
	 */
	private void nextState() {
		switch (this.state) {
		case PROBLEM:
			this.state = State.VARIABLES;
			if (this.problem.getSolver() != null) {
				try {
					this.solverName = this.problem.getSolver().getOptimizerName();
					this.solverState = this.problem.getSolver().getState();
				} catch (Exception e) {
					this.solverState = null;
					System.err.println(e.getMessage());
					e.printStackTrace();
				}
			}
			break;
		case VARIABLES:
			this.state = State.RESULTS;
			break;
		case RESULTS:
			this.state = State.GENERAL;
			break;
		case GENERAL:
			this.state = State.SOLVER;
			break;
		case SOLVER:
			break;
		}
		this.changeState();
	}

	/**
	 * Change to previous state
	 */
	private void backState() {
		switch (this.state) {
		case PROBLEM:
			break;
		case VARIABLES:
			try {
				if ((this.solverState != null) && (this.problem.getSolver() != null)
						&& this.solverName.equals(this.problem.getSolver().getOptimizerName())) {
					this.problem.getSolver().setState(this.solverState);
				}
				SettingsDialog.this.problem.getArchive().save();
			} catch (Exception e) {
				e.printStackTrace();
			}
			this.state = State.PROBLEM;
			break;
		case RESULTS:
			this.state = State.VARIABLES;
			break;
		case GENERAL:
			this.state = State.RESULTS;
			break;
		case SOLVER:
			this.state = State.GENERAL;
			break;
		}
		this.changeState();
	}

	/**
	 * Get detail settings panel for state
	 * 
	 * @param state
	 *            Current state
	 * @return Detail panel
	 */
	private SettingsPanel getSettingsPanel(State state) {
		SettingsPanel panel = null;

		switch (state) {
		case PROBLEM:
			if (this.problem != null) {
				panel = new ChooseProblem(this.problem, this);
			} else if (this.initPath != null) {
				if (this.initPath.getName().toLowerCase().endsWith(".mop")) {
					panel = new ChooseProblem(new MopFile(this.initPath), this);
				} else {
					panel = new ChooseProblem(this.initPath, this);
				}
			} else {
				panel = new ChooseProblem("", this);
			}
			break;
		case VARIABLES:
			if (this.problem != null) {
				panel = new ChooseVariables(this.problem, this);
			}
			break;
		case RESULTS:
			if (this.problem != null) {
				panel = new ChooseResults(this.problem, this);
			}
			break;
		case GENERAL:
			if (this.problem != null) {
				panel = new GeneralSettings(this.problem, this);
			}
			break;
		case SOLVER:
			if (this.problem != null) {
				panel = new SolverSettings(this.problem.getSolver(), this);
			}
			break;
		}

		return panel;
	}

	/**
	 * Change state of dialog (change displayed panel)
	 */
	private void changeState() {
		this.settings.removeAll();
		this.settingsPanel = this.getSettingsPanel(this.state);
		if (this.settingsPanel != null) {
			this.settings.add(this.settingsPanel);
		}
		this.repaint();

		SettingsDialog.this.refreshPanel();
	}

	/**
	 * Create dialog
	 * 
	 * @param parent
	 *            Parent gui element
	 */
	public SettingsDialog(Component parent) {
		this.create(parent);
	}

	/**
	 * Create dialog
	 * 
	 * @param parent
	 *            Parent gui element
	 * @param path
	 *            Working directory
	 */
	public SettingsDialog(Component parent, File path) {
		this.initPath = path;
		this.create(parent);
	}

	/**
	 * Create dialog
	 * 
	 * @param parent
	 *            Parent gui element
	 * @param prob
	 *            Optimizer probelem
	 */
	public SettingsDialog(Component parent, Problem prob) {
		this.problem = prob;
		this.create(parent);
	}

	/**
	 * Create dialog
	 * 
	 * @param parent
	 *            Parent gui element
	 */
	private void create(Component parent) {
		int x, y;

		this.setIconImages(OptIcons.getIcons());

		this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		this.setSize(700, 600);

		// set position of window
		if (parent != null) {
			x = parent.getLocation().x + ((parent.getWidth() - this.getWidth()) / 2);
			y = parent.getLocation().y + ((parent.getHeight() - this.getHeight()) / 2);

		} else {
			// screen insets
			Insets insets = Toolkit.getDefaultToolkit().getScreenInsets(GraphicsEnvironment
					.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDefaultConfiguration());
			// screen area
			Rectangle rect = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice()
					.getDefaultConfiguration().getBounds();
			// working area
			Rectangle work = new Rectangle(rect.x + insets.left, rect.y + insets.top,
					rect.width - insets.left - insets.right, rect.height - insets.top - insets.bottom);

			x = work.x + ((work.width - this.getWidth()) / 2);
			y = work.y + ((work.height - this.getHeight()) / 2);
		}
		if (x < 0) {
			x = 0;
		}
		if (y < 0) {
			y = 0;
		}
		this.setLocation(x, y);

		this.init();
		this.refreshPanel();
		this.setVisible(true);
		this.changeState();

		if (this.finish.isEnabled()) {
			this.finish.requestFocus();
		}
	}

	/**
	 * Get selected problem
	 * 
	 * @return Optimizer problem
	 */
	public Problem getProblem() {
		Problem prob = null;

		if (this.finished) {
			prob = this.problem;
		}

		return prob;
	}
}
