/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.Parameter.ParameterPath;
import at.lcm.bu21.magopt.structure.client.CDepend;
import at.lcm.bu21.magopt.structure.rmi.RDepend;

/**
 * Implementation of dependency structure for Parameter
 * 
 * @author Werner Koppelsttter
 */
public class Depend implements IDepend {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** List of imports */
	private ArrayList<DependItem> imports = new ArrayList<DependItem>(0);

	/** List of exports */
	private ArrayList<DependItem> exports = new ArrayList<DependItem>(0);

	/** Reference to the parent parameter */
	private Parameter parent;

	/**
	 * Constructor to create a dependency object for given parameter
	 * 
	 * @param parent
	 *            Parameter of dependency
	 */
	public Depend(Parameter parent) {
		this.parent = parent;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		Depend other = (Depend) obj;
		if (this.imports == null) {
			if (other.imports != null) {
				return false;
			}
		} else if (!this.imports.equals(other.imports)) {
			return false;
		}
		if (this.exports == null) {
			if (other.exports != null) {
				return false;
			}
		} else if (!this.exports.equals(other.exports)) {
			return false;
		}
		return true;
	}

	/**
	 * Create a deep clone of given parameter
	 * 
	 * @return Deep clone of parameter
	 * @throws Exception
	 */
	public Depend deepClone() throws Exception {
		return (Depend) Clone.deepClone(this);
	}

	/**
	 * Clone Parameter
	 * 
	 * @param clone
	 *            The clone depend
	 * @throws CloneException
	 */
	void cloneDepend(Depend clone) throws CloneException {

		// imports
		clone.clearImport();
		for (int k = 0; k < this.imports.size(); k++) {
			clone.imports.add(this.imports.get(k).cloneDepend());
		}

		// exports
		clone.clearExport();
		for (int k = 0; k < this.exports.size(); k++) {
			clone.exports.add(this.exports.get(k).cloneDepend());
		}
	}

	@Override
	public int getDependsLength(DependDirection direction) {
		AbstractList<DependItem> items;

		if (direction == DependDirection.IMPORT) {
			items = this.imports;
		} else {
			items = this.exports;
		}

		return items.size();
	}

	@Override
	public DependItem getDepend(DependDirection direction, int index) {
		AbstractList<DependItem> items;
		DependItem item = null;

		if (direction == DependDirection.IMPORT) {
			items = this.imports;
		} else {
			items = this.exports;
		}

		try {
			item = items.get(index);
		} catch (Exception e) {
		}

		return item;
	}

	/**
	 * Get depends of parameter
	 * 
	 * @param direction
	 *            Direction of depends
	 * @return imports of parameter
	 */
	public DependItem[] getDepends(DependDirection direction) {
		AbstractList<DependItem> items;
		DependItem[] list = new DependItem[0];

		if (direction == DependDirection.IMPORT) {
			items = this.imports;
		} else {
			items = this.exports;
		}

		if (items.size() > 0) {
			list = items.toArray(list);
			// try {
			// list = (DependItem[]) Clone.deepClone(items.toArray(list));
			// } catch (CloneException e) {
			// }
		}
		return list;
	}

	@Override
	public DependItem[] getImports() {
		return this.getDepends(DependDirection.IMPORT);
	}

	@Override
	public DependItem[] getExports() {
		return this.getDepends(DependDirection.EXPORT);
	}

	/**
	 * Replace object in import
	 * 
	 * @param oldObj
	 *            Object to replace
	 * @param newObj
	 *            Object that replace the object
	 */
	void replaceImport(Parameter oldObj, Parameter newObj) {
		this.replaceDepend(oldObj, newObj, DependDirection.IMPORT);
	}

	/**
	 * Replace object in export
	 * 
	 * @param oldObj
	 *            Object to replace
	 * @param newObj
	 *            Object that replace the object
	 */
	void replaceExport(Parameter oldObj, Parameter newObj) {
		this.replaceDepend(oldObj, newObj, DependDirection.EXPORT);
	}

	/**
	 * Replace object in dependency
	 * 
	 * @param oldObj
	 *            Object to replace
	 * @param newObj
	 *            Object that replace the object
	 * @param direction
	 *            flag for import/export
	 */
	protected void replaceDepend(Parameter oldObj, Parameter newObj,
			DependDirection direction) {
		DependItem item;
		ArrayList<DependItem> items;

		if (direction == DependDirection.IMPORT) {
			items = this.imports;
		} else {
			items = this.exports;
		}

		for (int i = 0; i < items.size(); i++) {
			item = items.get(i);
			if (item.getHandle() == oldObj) {
				item.setHandle(newObj);
			}
		}
	}

	@Override
	public void setImportVars(PortDesc[] vars) {
		boolean dirty;

		dirty = this.setVars(vars, DependDirection.IMPORT);
		if (dirty && (this.parent instanceof Function)) {
			Function func = (Function) this.parent;
			int len = this.parent.getLength();
			for (int k = 0; k < len; k++) {
				func.setUpdateRequired(k, true);
			}
			this.parent.refreshPartial();
		}
	}

	@Override
	public void setExportVars(PortDesc[] vars) {
		boolean dirty;

		dirty = this.setVars(vars, DependDirection.EXPORT);
		if (dirty && (this.parent instanceof Function)) {
			Function func = (Function) this.parent;
			int len = this.parent.getLength();
			for (int k = 0; k < len; k++) {
				func.setUpdateRequired(k, true);
			}
			this.parent.refreshPartial();
		}
	}

	/**
	 * Set variable-list of the dependency
	 * 
	 * @param vars
	 *            list of variables
	 * @param direction
	 *            flag for import/export
	 * @return whether anything has changed
	 */
	protected boolean setVars(PortDesc[] vars, DependDirection direction) {
		boolean dirty = false;
		DependItem item;
		int k = 0, m;
		AbstractList<DependItem> items;
		LinkedList<PortDesc> varList = new LinkedList<PortDesc>();

		if (vars != null) {
			varList.addAll(Arrays.asList(vars));
		}

		if (direction == DependDirection.IMPORT) {
			items = this.imports;
		} else {
			items = this.exports;
		}

		// delete reduce list
		while (k < items.size()) {
			item = items.get(k);
			m = 0;
			while ((m < varList.size()) && (item.getVarName()
					.compareTo(varList.get(m).getName()) != 0)) {
				m++;
			}
			if (m < varList.size()) {
				k++;
				item.setVarDesc(varList.remove(m));
			} else {
				items.remove(k);
				if (item.getHandle() != null) {
					dirty = true;
					if (direction == DependDirection.IMPORT) {
						item.getHandle().rmUsedBy(this.parent);
					} else {
						item.getHandle().rmSetBy(this.parent);
					}
				}
			}
		}

		// add to list
		for (int i = 0; i < varList.size(); i++) {
			items.add(new DependItem(varList.get(i)));
		}

		return dirty;
	}

	@Override
	public void setImportAssignments(String[][] vars) {
		this.setAssignments(vars, DependDirection.IMPORT);
		if (this.refreshDepend(false, false)) {
			this.parent.refreshPartial();
		}
	}

	@Override
	public void setImportAssignmentsMap(Map<String, String> vars) {
		this.setAssignments(vars, DependDirection.IMPORT);
		if (this.refreshDepend(false, false)) {
			this.parent.refreshPartial();
		}
	}

	@Override
	public void setExportAssignments(String[][] vars) {
		this.setAssignments(vars, DependDirection.EXPORT);
		if (this.refreshDepend(false, false)) {
			this.parent.refreshPartial();
		}
	}

	@Override
	public void setExportAssignmentsMap(Map<String, String> vars) {
		this.setAssignments(vars, DependDirection.EXPORT);
		if (this.refreshDepend(false, false)) {
			this.parent.refreshPartial();
		}
	}

	@Override
	public void setAssignment(String var, String param,
			DependDirection direction) {
		this.setAssignments(new String[][] { { var, param } }, direction);
		if (this.refreshDepend(false, false)) {
			this.parent.refreshPartial();
		}
	}

	/**
	 * Assign Parameter to the given variable list
	 * 
	 * @param vars
	 *            list of assignments
	 * @param direction
	 *            flag for import/export
	 */
	protected void setAssignments(Map<String, String> vars,
			DependDirection direction) {
		LinkedList<String[]> list;

		list = new LinkedList<String[]>();
		for (String key : vars.keySet()) {
			list.add(new String[] { key, vars.get(key) });
		}
		this.setAssignments(list.toArray(new String[0][0]), direction);

	}

	/**
	 * Assign Parameter to the given variable list
	 * 
	 * @param vars
	 *            list of assignments
	 * @param direction
	 *            flag for import/export
	 */
	protected void setAssignments(String[][] vars, DependDirection direction) {
		int i, k;
		i = 0;
		AbstractList<DependItem> items;

		if (direction == DependDirection.IMPORT) {
			items = this.imports;
		} else {
			items = this.exports;
		}

		if (vars != null) {
			while (i < vars.length) {
				k = 0;
				while ((k < items.size()) && (vars[i][0]
						.compareTo(items.get(k).getVarName()) != 0)) {
					k++;
				}
				if (k < items.size()) {
					if (vars[i][1].compareTo(items.get(k).getParam()) != 0) {
						items.get(k).setParam(vars[i][1]);
					}
				} else {
					this.parent.warning(
							"Failure by assigning Parameter: " + vars[i][0]);
				}
				i++;
			}
		}
	}

	/**
	 * Refresh dependencies. Check whether the references are up-do-date and
	 * refresh them.
	 * 
	 * @param dirty
	 *            flag to indicate if parameter already changed
	 * @param load
	 *            flag to indicate loading parameter (no modification)
	 * @return weather anything has changed
	 */
	boolean refreshDepend(boolean dirty, boolean load) {
		Parameter oldHandle, newHandle;
		Parameter param = this.parent;
		Parameter root = param.getRoot();
		AbstractList<ParameterPath> list;

		// imports
		if (this.imports != null) {
			for (int k = 0; k < this.imports.size(); k++) {
				// get object to import
				list = param.getObjectPaths(this.imports.get(k).getParam());
				while (!list.isEmpty()
						&& (list.get(0).getParameter() == this.parent)) {
					list.remove(0);
				}
				if (list.isEmpty()) {
					newHandle = null;
				} else {
					newHandle = list.get(0).getParameter();
				}

				oldHandle = this.imports.get(k).getHandle();
				if (oldHandle != newHandle) {
					dirty = true;
					this.imports.get(k).setHandle(newHandle);
					if (oldHandle != null) {
						oldHandle.rmUsedBy(this.parent);
					}
					if (newHandle != null) {
						newHandle.addUsedBy(this.parent);
					}
				}
			}
		}

		// exports
		if (this.exports != null) {
			for (int k = 0; k < this.exports.size(); k++) {
				newHandle = param.getObject(this.exports.get(k).getParam());
				oldHandle = this.exports.get(k).getHandle();
				if (oldHandle != newHandle) {
					dirty = true;
					this.exports.get(k).setHandle(newHandle);
					if (oldHandle != null) {
						oldHandle.rmSetBy(this.parent);
					}
					if (newHandle != null) {
						newHandle.addSetBy(this.parent);
					}
				}
			}
		}

		if (!load && dirty) {
			for (int k = 0; k < root.getLength(); k++) {
				if (param instanceof Function) {
					((Function) param).setUpdateRequired(k, true);
				} else {
					this.parent.setModified(k, true);
				}
			}
		}

		return dirty;
	}

	/**
	 * Replace imports with given list
	 * 
	 * @param depend
	 *            List of imports
	 */
	void replaceImports(DependItem[] depend) {
		this.clearImport();
		for (DependItem item : depend) {
			this.imports.add(item);
		}
	}

	/**
	 * Replace exports with given list
	 * 
	 * @param depend
	 *            List of exports
	 */
	void replaceExports(DependItem[] depend) {
		this.clearExport();
		for (DependItem item : depend) {
			this.exports.add(item);
		}
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws Exception
	 */
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;
		DependItem obj;

		xmlReader.next();

		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("Imports")) {
				xmlReader.next();
				this.clearImport();
				while (xmlReader
						.getEventType() == XMLStreamConstants.START_ELEMENT) {
					obj = new DependItem();
					this.imports.add(obj);
					obj.xml2obj(xmlReader);
				}
				xmlReader.next();
			} else if (element.equals("Exports")) {
				xmlReader.next();
				this.clearExport();
				while (xmlReader
						.getEventType() == XMLStreamConstants.START_ELEMENT) {
					obj = new DependItem();
					this.exports.add(obj);
					obj.xml2obj(xmlReader);
				}
				xmlReader.next();

			} else {
				throw new Exception(
						"Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws Exception
	 */
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("Depend");

		// Imports
		xmlWriter.writeStartElement("Imports");
		for (DependItem depend : this.imports) {
			xmlWriter.writeStartElement("entry");
			depend.obj2xml(xmlWriter);
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();

		// Exports
		xmlWriter.writeStartElement("Exports");
		for (DependItem depend : this.exports) {
			xmlWriter.writeStartElement("entry");
			depend.obj2xml(xmlWriter);
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();

	}

	/**
	 * Report import-dependencies
	 * 
	 * @param content
	 *            tex-writer
	 */
	public void reportImport(DocTexWriter content) {
		content.add("\\subsection*{Imports}\n");
		this.report(content, this.imports);
	}

	/**
	 * Report export-dependencies
	 * 
	 * @param content
	 *            tex-writer
	 */
	public void reportExport(DocTexWriter content) {
		content.add("\\subsection*{Exports}\n");
		this.report(content, this.exports);
	}

	/**
	 * Report dependencies
	 * 
	 * @param content
	 *            tex-writer
	 * @param depend
	 *            depend-list
	 */
	protected void report(DocTexWriter content,
			AbstractList<DependItem> depend) {
		DependItem item;

		content.add("\t\\ifdefined\\usehtml");
		content.add("\\begin{tabular}{|m{3.5cm}|m{3cm}|m{6cm}|}");
		content.add("\\else");
		content.add("\\begin{longtable}{|m{3.5cm}|m{3cm}|m{6cm}|}");
		content.add("\\fi\n");

		content.add("\t\\hline");
		content.add("\\rowcolor[RGB]{\\RGBrowcolor}");
		content.add("Variable & ");
		content.add("Parameter & ");
		content.add("Idtentifier \\\\\n");
		content.add("\t\t\\hline\n");
		content.add("\t\\ifdefined\\usehtml\\else");
		content.add("\\endhead");
		content.add("\\fi\n");

		for (int k = 0; k < depend.size(); k++) {
			item = depend.get(k);
			content.add("\t\t\\hline ");
			// var
			content.addString(item.getVarName());
			content.add(" & ");

			// param
			content.addString(item.getParam());
			content.add(" & ");

			// handle
			if (item.getHandle() != null) {
				content.add("\\hyperlink{");
				content.addName("D_" + item.getHandle().getIdentifier());
				content.add("}{");
				content.addString(item.getHandle().getIdentifier());
				content.add("}");
			}
			content.add(" \\\\\n");
		}
		content.add("\t\t\\hline\n");
		content.add("\t\\ifdefined\\usehtml");
		content.add("\\end{tabular}");
		content.add("\\else");
		content.add("\\end{longtable}");
		content.add("\\fi\n\n");

	}

	/**
	 * Clear list of imports
	 */
	protected void clearImport() {
		Parameter handle;

		if (this.imports != null) {
			for (int k = 0; k < this.imports.size(); k++) {
				handle = this.imports.get(k).getHandle();
				if (handle != null) {
					handle.rmUsedBy(this.parent);
				}
			}
			this.imports.clear();
		}

	}

	/**
	 * Clear list of exports
	 */
	protected void clearExport() {
		Parameter handle;

		if (this.exports != null) {
			for (int k = 0; k < this.exports.size(); k++) {
				handle = this.exports.get(k).getHandle();
				if (handle != null) {
					handle.rmSetBy(this.parent);
				}
			}
			this.exports.clear();
		}

	}

	/**
	 * Get RMI-adapter of depend
	 * 
	 * @return RMI-adapter
	 * @throws RemoteException
	 */
	public IDepend getRMIImpl() throws RemoteException {
		return new RDepend(this);
	}

	@Override
	public IDepend getClientImpl() {
		return new CDepend(this);
	}
}
