/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import magopt.utils.MopAccess;
import magopt.utils.MopFile;
import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_34.
 * 
 * - Optimizer split solver-settings and solver-state
 * 
 * @author Werner Koppelsttter
 */
public class Converter_34 implements MopAccess {

	/** Mop-file of project */
	MopFile mop;

	/** Node of load archive */
	XMLNode node = null;

	/**
	 * Convert the given node
	 * 
	 * @param mop
	 *            Mop-file with optimization
	 * @throws Exception
	 */
	static void convertOptimizer(MopFile mop) throws Exception {
		Converter_34 convert;
		String[] filelist;

		convert = new Converter_34(mop);
		filelist = mop.getFileNames();
		for (String name : filelist) {
			if (name.startsWith("GeneticOptimization/")
					&& name.endsWith(".xml")) {
				convert.convertOpt(name);
			}
		}
	}

	/**
	 * Create converter object
	 * 
	 * @param mop
	 *            Mop-file to convert
	 */
	private Converter_34(MopFile mop) {
		this.mop = mop;
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param name
	 *            Name of archive
	 * @throws Exception
	 */
	private void convertOpt(String name) throws Exception {
		this.node = null;
		this.mop.readStream(this, name);
		this.convert(this.node);
		this.mop.writeStream(this, name);
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param node
	 *            Archive to convert
	 */
	private void convert(XMLNode node) {
		XMLNode state, settings;
		String attr;

		for (XMLNode n : node.getNodes()) {
			if (n.getName().compareTo("field") == 0) {
				attr = n.getAttribute("name");
				if ((attr != null) && (attr.compareTo("solverSettings") == 0)) {
					state = new XMLNode("field");
					state.setAttribute("name", "solverState");
					state.addNode(new XMLNode("Struct"));
					node.addNode(state);
					settings = n.getNode(0);
					for (int k = 0; k < settings.getNodeSize(); k++) {
						if (settings.getNode(k).getAttribute("name")
								.compareTo("archive") == 0) {
							state.getNode(0).addNode(settings.rmNode(k));
						}
					}
				}
			}
			this.convert(n);
		}
	}

	@Override
	public void readStream(InputStream is) throws Exception {
		XMLStreamReader xmlReader = XMLInputFactory.newInstance()
				.createXMLStreamReader(is, "utf-8");
		while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
		}

		// read objects
		this.node = new XMLNode(xmlReader);

		if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
			throw new XMLStreamException("Undefined error by reading file!");
		}
		xmlReader.close();
	}

	@Override
	public void writeStream(OutputStream os) throws Exception {
		XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance()
				.createXMLStreamWriter(os, "utf-8");
		xmlWriter.writeStartDocument("utf-8", "1.0");

		// write object to File
		this.node.writeXML(xmlWriter);

		xmlWriter.writeEndDocument();
		xmlWriter.close();
	}
}
