/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.simulate;

import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Map;

import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.structure.simulate.SimulateEvent.EventType;
import jobmanager.JobDesc;

/**
 * Simulate all cases
 *
 * @author Werner Koppelsttter
 */
public class SimulateAll {

	/** Project for simulation */
	private Project proj;

	/** Status of simulation */
	private String status = "Idle";

	/** Simulation is started */
	private boolean start = false;

	/** Stop simulation */
	private boolean stop = false;

	/** Number of finished cases */
	private int finished = 0;

	/** Number of error cases */
	private int error = 0;

	/** Number of initialized cases */
	private int initialized = 0;

	/** Simulation thread */
	private RegenerateThread regenThread = null;

	private boolean singleSim = true;
	private int singleSimIndex = 0;

	/** Simulation arguments */
	private Map<String, Object> args = new Hashtable<String, Object>();

	/** List of listeners */
	private transient LinkedList<SimulateListener> listeners = null;

	/**
	 * Simulate thread
	 *
	 * @author Werner Koppelsttter
	 */
	private class RegenerateThread extends Thread {

		@Override
		public void run() {
			int index;
			boolean wait;
			String[] parameters;
			boolean regen;
			boolean init;

			SimulateAll.this.start = true;
			SimulateAll.this.setStatus("Running");
			SimulateAll.this.fireEvent(new SimulateEvent(RegenerateThread.this,
					EventType.STARTED));

			if (SimulateAll.this.initialized > 0) {
				SimulateAll.this.finished = 0;
				SimulateAll.this.error = 0;
				for (int k = 0; k < SimulateAll.this.initialized; k++) {
					if (SimulateAll.this.proj.getError(k)) {
						SimulateAll.this.error++;
					} else if (!SimulateAll.this.proj.getIncomplete(k)) {
						SimulateAll.this.finished++;
					}
				}
			}

			while (!SimulateAll.this.stop) {
				index = 0;
				wait = true;
				init = false;

				while (!SimulateAll.this.stop && !init
						&& (index < SimulateAll.this.proj.getLength())
						&& (index <= SimulateAll.this.initialized)) {

					// initialize case
					if (!SimulateAll.this.stop
							&& (index >= SimulateAll.this.initialized)) {
						SimulateAll.this
								.setStatus("Initialize #" + (index + 1));
						SimulateAll.this.proj.setIndex(index);
						SimulateAll.this.proj.refresh();
						SimulateAll.this.proj.setError(false);
						SimulateAll.this.initialized++;
						init = true;
						wait = false;

						if (SimulateAll.this.proj.getError(index)) {
							SimulateAll.this.error++;
						} else if (!SimulateAll.this.proj
								.getIncomplete(index)) {
							SimulateAll.this.finished++;
						}
					}

					// simulate incomplete case
					if (!SimulateAll.this.stop
							&& SimulateAll.this.proj.getIncomplete(index)
							&& !SimulateAll.this.proj.getError(index)) {
						// Check running jobs
						parameters = JobDesc.getParameterList(
								SimulateAll.this.proj.getFile(),
								SimulateAll.this.proj.getDsId(index));
						if ((parameters == null) || (parameters.length == 0)) {
							regen = true;
						} else {
							regen = false;
							for (String param : parameters) {
								regen |= JobDesc.allFinished(
										SimulateAll.this.proj.getFile(),
										SimulateAll.this.proj.getDsId(index),
										param);
							}
						}

						// simulate case
						if (regen) {
							SimulateAll.this.proj.setIndex(index);
							SimulateAll.this
									.setStatus("Regenerate #" + (index + 1));
							SimulateAll.this.proj.setIndex(index);
							SimulateAll.this.proj
									.regenerate(SimulateAll.this.args);
							wait = false;
							if (!SimulateAll.this.proj.getIncomplete(index)) {
								SimulateAll.this.finished++;
							}
							if (SimulateAll.this.proj.getError(index)) {
								SimulateAll.this.error++;
							}
						}
					}

					index++;

					SimulateAll.this.setStatus("Waiting");

				}

				// stop if all finished
				if (!SimulateAll.this.stop) {
					if ((SimulateAll.this.finished
							+ SimulateAll.this.error) == SimulateAll.this.proj
									.getLength()) {
						SimulateAll.this.stop();
					} else {
						// sleep if loop is to fast (relax communication)
						try {
							if (wait) {
								int k = 0;
								while (!SimulateAll.this.stop && (k < 10)) {
									Thread.sleep(500);
									k++;
								}
							} else {
								Thread.sleep(100);
							}
						} catch (InterruptedException e) {
						}
					}
				}
			}

			SimulateAll.this.start = false;
			SimulateAll.this.regenThread = null;
			SimulateAll.this.setStatus("Idle");

			SimulateAll.this.fireEvent(new SimulateEvent(RegenerateThread.this,
					EventType.STOPPED));

		}
	}

	/**
	 * Initilize SimulateAll
	 * 
	 * @param proj
	 *            Project to simulate
	 */
	public SimulateAll(Project proj) {
		this.proj = proj;
	}

	/**
	 * Get status sting
	 * 
	 * @return Status
	 */
	public String getStatus() {
		return this.status;
	}

	/**
	 * Set status string
	 * 
	 * @param status
	 *            Status
	 */
	protected void setStatus(String status) {
		this.status = status;
		SimulateAll.this
				.fireEvent(new SimulateEvent(this, EventType.STATUS_CHANGED));
	}

	/**
	 * Get run simulation flag
	 * 
	 * @return Run-simulation
	 */
	public boolean getRunSimulation() {
		Boolean runSim;

		runSim = (Boolean) this.args.get("RunSim");

		return runSim != null ? runSim : false;
	}

	/**
	 * Set run simulation flag
	 * 
	 * @param runSim
	 *            Run-simulation
	 */
	public void setRunSimulation(boolean runSim) {
		this.args.put("RunSim", runSim);
		this.fireEvent(new SimulateEvent(this, EventType.RUN_SIM_CHANGED));
	}

	/**
	 * Get remote compute flag
	 * 
	 * @return Remote-compute
	 */
	public boolean getRemoteCompute() {
		Boolean remoteCompute;

		remoteCompute = (Boolean) this.args.get("RemoteCompute");

		return remoteCompute != null ? remoteCompute : false;
	}

	/**
	 * Set remote compute flag
	 * 
	 * @param remoteCompute
	 *            Remote-compute
	 */
	public void setRemoteCompute(boolean remoteCompute) {
		this.args.put("RemoteCompute", remoteCompute);
		this.fireEvent(
				new SimulateEvent(this, EventType.REMOTE_COMPUTE_CHANGED));
	}

	/**
	 * Get optimize flag
	 * 
	 * @return Optimize
	 */
	public boolean getOptimize() {
		Boolean optimize;

		optimize = (Boolean) this.args.get("Optimize");

		return optimize != null ? optimize : false;
	}

	/**
	 * Set optimize flag
	 * 
	 * @param optimize
	 *            Optimize
	 */
	public void setOptimize(boolean optimize) {
		this.args.put("Optimize", optimize);
		this.fireEvent(new SimulateEvent(this, EventType.OPTIMIZE_CHANGED));
	}

	/**
	 * Get timeout in seconds
	 * 
	 * @return Timeout
	 */
	public long getTimeout() {
		Long timeout;

		timeout = (Long) this.args.get("Timeout");

		return timeout != null ? timeout : Long.MAX_VALUE;
	}

	/**
	 * Set timeout in seconds
	 * 
	 * @param timeout
	 *            Timeout
	 */
	public void setTimeout(long timeout) {
		this.args.put("Timeout", timeout);
		this.fireEvent(new SimulateEvent(this, EventType.TIMEOUT_CHANGED));
	}

	public boolean getSingleSim() {
		return this.singleSim;
	}

	public void setSingleSim(boolean singleSim) {
		this.singleSim = singleSim;
	}

	public int getSingleSimIndex() {
		return this.singleSimIndex;
	}

	public void setSingleSimIndex() {

	}

	/**
	 * Is simulation started
	 * 
	 * @return Started
	 */
	public boolean isStart() {
		return this.start;
	}

	/**
	 * Start simulation
	 */
	public void start() {
		this.fireEvent(new SimulateEvent(this, EventType.START));
		this.stop = false;
		if ((this.regenThread == null) || !this.regenThread.isAlive()) {
			this.regenThread = new RegenerateThread();
			this.regenThread.start();
		}
	}

	/**
	 * Is stop simulation is set
	 * 
	 * @return Stop
	 */
	public boolean isStop() {
		return this.stop;
	}

	/**
	 * Stop simulation
	 */
	public void stop() {
		this.fireEvent(new SimulateEvent(this, EventType.STOP));
		this.stop = true;
	}

	/**
	 * Get list of listeners
	 * 
	 * @return <code>SimulateListener</code> list
	 */
	private LinkedList<SimulateListener> getSimulateListenersList() {
		if (this.listeners == null) {
			this.listeners = new LinkedList<SimulateListener>();
		}
		return this.listeners;
	}

	/**
	 * Add simulate listener
	 * 
	 * @param listener
	 *            Listener to add
	 */
	public void addSimulateListener(SimulateListener listener) {
		LinkedList<SimulateListener> list;

		list = this.getSimulateListenersList();
		synchronized (list) {
			if (!list.contains(listener)) {
				list.add(listener);
			}
		}
	}

	/**
	 * Remove simulate listener
	 * 
	 * @param listener
	 *            Listener to remove
	 */
	public void removeSimulateListener(SimulateListener listener) {
		LinkedList<SimulateListener> list;

		list = this.getSimulateListenersList();
		synchronized (list) {
			list.remove(listener);
		}
	}

	/**
	 * Get list of simulate listener
	 * 
	 * @return List of listener
	 */
	public SimulateListener[] getSimulateListeners() {
		LinkedList<SimulateListener> list;

		list = this.getSimulateListenersList();
		synchronized (list) {
			return list.toArray(new SimulateListener[0]);
		}
	}

	/**
	 * Call to fire the event
	 * 
	 * @param event
	 *            Event that occurs
	 */
	private void fireEvent(SimulateEvent event) {
		LinkedList<SimulateListener> list;

		list = this.getSimulateListenersList();
		synchronized (list) {
			for (SimulateListener listener : list) {
				listener.simulateChanged(event);
			}
		}
	}

	/**
	 * Get number of finished cases
	 * 
	 * @return Number of finished
	 */
	public int getNumFinished() {
		return this.finished;
	}

	/**
	 * Get number of error cases
	 * 
	 * @return Number of errors
	 */
	public int getNumError() {
		return this.error;
	}

	/**
	 * Get number of incomplete caese
	 * 
	 * @return Number of incomplete
	 */
	public int getNumIncomplete() {
		return this.initialized - this.finished - this.error;
	}

	/**
	 * Get number of not initialized
	 * 
	 * @return Number not initialized
	 */
	public int getNumNotInitialized() {
		return this.proj.getLength() - this.initialized;
	}
}
