/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.console;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;

/**
 * Display input, output and error of console
 * 
 * @author Werner Koppelsttter
 */
public class Console extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Console display */
	JTextPane display;

	/** Scroll pane of display */
	JScrollPane scroll;

	/** Maximum number of lines */
	private int limitLines = 1000;

	/** PrintStream to print to console */
	private PrintStream out = null;

	/** PrintStream to print errors to console */
	private PrintStream err = null;

	/** Attributes for document (color) */
	private SimpleAttributeSet set = new SimpleAttributeSet();

	/** Popup-menu of display */
	protected JPopupMenu popup;

	/**
	 * Create console to print input, output and errors
	 */
	public Console() {
		this.setLayout(new BorderLayout());

		this.display = new JTextPane();
		// this.display.setAutoscrolls(true);
		this.display.setEditable(false);
		this.display.setBorder(new EmptyBorder(10, 10, 10, 10));

		this.popup = new JPopupMenu("Hallo");
		this.display.setComponentPopupMenu(this.popup);
		JMenuItem item = new JMenuItem("Clear display");
		this.popup.add(item);
		item.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				Console.this.clear();
			}
		});

		this.scroll = new JScrollPane(this.display);
		this.add(this.scroll, BorderLayout.CENTER);

		this.out = new PrintStream(new OutputStream() {
			StringBuilder line = new StringBuilder();

			@Override
			public void write(int b) throws IOException {
				this.line.append((char) b);

				if ((b == '\n') || (b == '\r')) {
					Console.this.writeOutput(this.line.toString());
					this.line.setLength(0);
				}
			}

			@Override
			public void flush() {
				if (this.line.length() > 0) {
					Console.this.writeOutput(this.line.toString());
					this.line.setLength(0);
				}
			}

			@Override
			public void close() throws IOException {
				this.flush();
				super.close();
			}
		});

		this.err = new PrintStream(new OutputStream() {
			StringBuilder line = new StringBuilder();

			@Override
			public void write(int b) throws IOException {
				this.line.append((char) b);

				if ((b == '\n') || (b == '\r')) {
					Console.this.writeError(this.line.toString());
					this.line.setLength(0);
				}
			}

			@Override
			public void flush() {
				if (this.line.length() > 0) {
					Console.this.writeOutput(this.line.toString());
					this.line.setLength(0);
				}
			}

			@Override
			public void close() throws IOException {
				this.flush();
				super.close();
			}
		});
	}

	/**
	 * Write input to console
	 * 
	 * @param msg
	 *            Message
	 */
	public void writeInput(String msg) {
		this.write(msg, Color.GREEN.darker());
	}

	/**
	 * Write output to console
	 * 
	 * @param msg
	 *            Message
	 */
	public void writeOutput(String msg) {
		this.write(msg, Color.BLACK);
	}

	/**
	 * Write error to console
	 * 
	 * @param msg
	 *            Message
	 */
	public void writeError(String msg) {
		this.write(msg, Color.RED);
	}

	/**
	 * Write to console
	 * 
	 * @param msg
	 *            Message
	 * @param color
	 *            Color of message
	 */
	private synchronized void write(final String msg, final Color color) {
		Document doc = Console.this.display.getDocument();

		// insert message
		StyleConstants.setForeground(Console.this.set, color);
		try {
			doc.insertString(doc.getLength(), msg, Console.this.set);
		} catch (BadLocationException e) {
			e.printStackTrace();
		}

		// trim number of lines
		this.trimLines();
	}

	/**
	 * Trim number of lines
	 */
	private void trimLines() {
		Document doc = this.display.getDocument();

		try {
			String text = doc.getText(0, doc.getLength());
			while (text.split("\n").length > this.limitLines) {
				doc.remove(0, text.indexOf("\n") + 1);
				text = doc.getText(0, doc.getLength());
			}
		} catch (BadLocationException e) {
			e.printStackTrace();
		}

		SwingUtilities.invokeLater(new Runnable() {

			@Override
			public void run() {
				// Console.this.scrollRectToVisible(
				// new Rectangle(0, Console.this.getHeight() - 2, 1, 1));
				Console.this.scroll.getVerticalScrollBar()
						.setValue(Console.this.scroll.getVerticalScrollBar()
								.getMaximum());
			}
		});
	}

	/**
	 * Get PrintStream for output
	 * 
	 * @return PrintStream
	 */
	public PrintStream getOutputPrintStream() {
		return this.out;
	}

	/**
	 * Get PrintStream for error
	 * 
	 * @return PrintStream
	 */
	public PrintStream getErrorPrintStream() {
		return this.err;
	}

	/**
	 * Get the maximum number of lines
	 * 
	 * @return Number of lines
	 */
	public int getLimitLines() {
		return this.limitLines;
	}

	/**
	 * Set the maximum number of lines
	 * 
	 * @param lines
	 *            Number of lines
	 */
	public void setLimitLines(int lines) {
		this.limitLines = lines;
		this.trimLines();
	}

	/**
	 * Clear console
	 */
	public void clear() {
		Document doc;

		doc = this.display.getDocument();
		try {
			doc.remove(0, doc.getLength());
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
	}
}
