/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.view.report;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.rmi.RemoteException;

import javax.swing.DefaultCellEditor;
import javax.swing.GroupLayout;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;

import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;
import at.lcm.bu21.magopt.structure.ReportParameter;
import at.lcm.bu21.magopt.structure.ReportParameter.Language;
import magopt.gui.view.ParamBaseScroll;

/**
 * View of report-parameter
 * 
 * @author Werner Koppelsttter
 */
public class ReportParameterView extends ParamBaseScroll {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Project of report-parameter */
	private IProject project = null;

	/** Layout of panel */
	private GroupLayout layout;

	/** Selection of styles */
	private JComboBox<String> boxStyle;

	/** Selection of languages */
	private JComboBox<Language> boxLang;

	/** Table of report-properties */
	private JTable table;

	/** Model of report-properties-table */
	private ReportPropertyModel model;

	/** Main panel */
	private JPanel mainPanel;

	/**
	 * Create and initialize view
	 */
	public ReportParameterView() {
		JLabel labelStyle, labelLang, labelReportParam;
		JScrollPane panelTable;

		this.mainPanel = new JPanel();
		this.mainPanel.setVisible(false);
		this.add(this.mainPanel);
		this.mainPanel.setPreferredSize(this.getPreferredSize());

		this.layout = new GroupLayout(this.mainPanel);
		this.mainPanel.setLayout(this.layout);

		// style
		labelStyle = new JLabel("Report style");
		this.mainPanel.add(labelStyle);

		this.boxStyle = new JComboBox<String>(ReportParameter.getStyles());
		this.boxStyle.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					ReportParameterView.this.project.getReportParameter()
							.setStyle((String) ReportParameterView.this.boxStyle.getSelectedItem());
				} catch (RemoteException e1) {
					e1.printStackTrace();
				}
				ReportParameterView.this.update();
			}
		});

		this.mainPanel.add(this.boxStyle);

		// language
		labelLang = new JLabel("Language");
		this.mainPanel.add(labelLang);

		this.boxLang = new JComboBox<Language>(ReportParameter.Language.values());
		this.boxLang.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					ReportParameterView.this.project.getReportParameter()
							.setLanguage((ReportParameter.Language) ReportParameterView.this.boxLang.getSelectedItem());
				} catch (RemoteException e1) {
					e1.printStackTrace();
				}
			}
		});
		this.mainPanel.add(this.boxLang);

		// report parameter
		labelReportParam = new JLabel("Report Parameter");
		this.mainPanel.add(labelReportParam);

		panelTable = new JScrollPane();
		this.mainPanel.add(panelTable);
		this.table = new JTable();
		panelTable.setViewportView(this.table);

		this.table.getTableHeader().setReorderingAllowed(false);
		this.table.setColumnSelectionAllowed(true);
		this.table.setRowSelectionAllowed(true);
		this.table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		this.model = new ReportPropertyModel();
		this.table.setModel(this.model);

		// Cell renderer
		this.table.setDefaultRenderer(Object.class, new ReportPropertyRenderer());

		// Select test by enter the editor
		this.table.setDefaultEditor(Object.class, new DefaultCellEditor(new JTextField()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row,
					int column) {
				Component c;
				c = super.getTableCellEditorComponent(table, value, isSelected, row, column);
				((JTextField) c).selectAll();
				return c;
			}
		});
		// TODO event
		// this.model.addTableModelListener(new TableModelListener() {
		// @Override
		// public void tableChanged(TableModelEvent e) {
		// ReportPropertiesView.this.fireParamChanged();
		// }
		// });

		// define layout
		this.layout.setAutoCreateGaps(true);
		this.layout.setAutoCreateContainerGaps(true);
		this.layout.setHorizontalGroup(this.layout.createSequentialGroup()
				.addGroup(this.layout.createParallelGroup(GroupLayout.Alignment.LEADING).addComponent(labelStyle)
						.addComponent(labelLang).addComponent(labelReportParam))
				.addGroup(this.layout.createParallelGroup(GroupLayout.Alignment.LEADING).addComponent(this.boxStyle)
						.addComponent(this.boxLang).addComponent(panelTable)));
		this.layout.setVerticalGroup(this.layout.createSequentialGroup()
				.addGroup(this.layout.createParallelGroup(GroupLayout.Alignment.BASELINE).addComponent(labelStyle)
						.addComponent(this.boxStyle))
				.addGroup(this.layout.createParallelGroup(GroupLayout.Alignment.BASELINE).addComponent(labelLang)
						.addComponent(this.boxLang))
				.addGroup(this.layout.createParallelGroup(GroupLayout.Alignment.BASELINE).addComponent(labelReportParam)
						.addComponent(panelTable)));
	}

	@Override
	protected void update() {
		try {
			if (this.project != null) {
				this.model.setReport(this.project.getReportParameter());
			} else {
				this.model.setReport(null);
			}
		} catch (RemoteException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Sets new Project.
	 * 
	 * @param param
	 *            Project parameter
	 */
	public void setProject(IParameter param) {
		if (this.table.isEditing()) {
			this.table.getCellEditor().cancelCellEditing();
		}
		if (param instanceof IProject) {
			this.project = (IProject) param;
		} else {
			this.project = null;
		}

		this.mainPanel.setVisible(this.project != null);

		try {
			if (this.project != null) {
				this.boxStyle.setSelectedItem(this.project.getReportParameter().getStyle());
				this.boxLang.setSelectedItem(this.project.getReportParameter().getLanguage());
			}
		} catch (RemoteException ex) {
			ex.printStackTrace();
		}
		this.update();
	}
}
