/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm;

import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;

import at.lcm.bu21.magopt.utils.MagOptConfig;
import at.lcm.bu21.optim.algorithm.decmo.SsaDecmo;
import at.lcm.bu21.optim.algorithm.grid.Grid;
import at.lcm.bu21.optim.algorithm.me_nsga2.GenME_Nsga2;
import at.lcm.bu21.optim.algorithm.me_nsga2.SsaME_Nsga2;
import at.lcm.bu21.optim.algorithm.nsga2.GenNsga2;
import at.lcm.bu21.optim.algorithm.nsga2.SsaNsga2;
import at.lcm.bu21.optim.algorithm.spea2.GenSpea2;
import at.lcm.bu21.optim.algorithm.spea2.SsaSpea2;
import at.lcm.bu21.optim.core.Problem;

/**
 * List of available Optimizer
 * 
 * @author Werner Koppelsttter
 */
public enum SolverList {
	/** Steady State Asynchronous DECMO */
	SSA_DECMO(SsaDecmo.class),
	// /** Generation-based DECMO */
	// GEN_DECMO(GenDecmo.class),
	/** Steady State Asynchronous Spea2 */
	SSA_SPEA2(SsaSpea2.class),
	/** Generation-based Spea2 */
	GEN_SPEA2(GenSpea2.class),
	/** Steady State Asynchronous Nsga2 */
	SSA_NSGA2(SsaNsga2.class),
	/** Generation-based Nsga2 */
	GEN_NSGA2(GenNsga2.class),
	/** Grid */
	GRID(Grid.class),
	/** Steady State Asynchronous ME-NSGA-II */
	SSA_ME_NSGA2(SsaME_Nsga2.class),
	/** Steady State Asynchronous ME-NSGA-II */
	GEN_ME_NSGA2(GenME_Nsga2.class);

	/** Class of solver */
	private Class<? extends Solver> cl;

	/**
	 * Get valid values
	 * 
	 * @return List of Solver
	 */
	public static SolverList[] getValues() {
		SolverList[] list;
		String config;

		list = SolverList.values();

		config = MagOptConfig.getConfig("OptimizerDebug");
		if ((config == null) || !Boolean.parseBoolean(config)) {
			list = Arrays.copyOfRange(list, 0, list.length - 2);
		}
		return list;
	}

	/**
	 * Get OptimizerSolver from enum-name
	 * 
	 * @param name
	 *            Name of OptimizerSolver
	 * @return Enumeration-entry
	 */
	public static SolverList getSolverListEntry(String name) {
		SolverList item = null;

		for (SolverList opt : SolverList.values()) {
			if (opt.getName().compareTo(name) == 0) {
				item = opt;
			}
		}

		return item;
	}

	/**
	 * Create an enumeration-entry
	 * 
	 * @param cl
	 */
	private SolverList(Class<? extends Solver> cl) {
		this.cl = cl;
	}

	/**
	 * Get instance of optimizer
	 * 
	 * @param problem
	 *            Problem for solver
	 * @return Optimizer
	 */
	public Solver getSolver(Problem problem) {
		Solver inst = null;
		try {
			inst = this.cl.getConstructor(Problem.class).newInstance(problem);
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		}
		return inst;
	}

	/**
	 * Get name of Optimizer
	 * 
	 * @return Name of Optimizer
	 */
	public String getName() {
		String val = this.getSolver((Problem) null).getOptimizerName();
		return val;
	}

	/**
	 * Is optimizer generation-based
	 * 
	 * @return True of optmizer is generation-based
	 */
	public boolean isGenerationBased() {
		return (GenerationBased.class.isAssignableFrom(this.cl));
	}

	@Override
	public String toString() {
		return this.getName();
	}

}
