/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 252 $
 * $Date: 2020-06-22 23:52:03 +0200 (Mon, 22 Jun 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * List of individuals which are already calculated
 * 
 * @author Werner Koppelsttter
 */
public class Memory implements Serializable {

	/** Archive of Memory */
	private Archive archive;

	/** List with all calculated individuals */
//	CopyOnWriteArrayList<Individual> individuals = new CopyOnWriteArrayList<Individual>(); // slow with many individuals
//	LinkedList<Individual> individuals = new LinkedList<Individual>();
//	List<Individual> individuals = Collections.synchronizedList(new LinkedList<Individual>());
//	ArrayList<Individual> individuals = new ArrayList<Individual>();
	List<Individual> individuals = Collections.synchronizedList(new ArrayList<Individual>());

	/**
	 * Create memory to archive
	 * 
	 * @param archive
	 *            Archive
	 */
	public Memory(Archive archive) {
		this.archive = archive;
	}

	/**
	 * Get archive of memory
	 * 
	 * @return Archive
	 */
	public Archive getArchive() {
		return this.archive;
	}

	/**
	 * Clear memory
	 */
	public void clear() {
		this.individuals.clear();
	}

	/**
	 * Is memory emtpy
	 * 
	 * @return Is memory emtpy
	 */
	public boolean isEmpty() {
		return this.individuals.isEmpty();
	}

	/**
	 * Get number of individuals in memory
	 * 
	 * @return Number of individuals
	 */
	public int getNumIndividuals() {
		return this.individuals.size();
	}

	/**
	 * Add individual to memory
	 * 
	 * @param ind
	 *            Individual to add
	 */
	public void addIndividual(Individual ind) {
		this.addIndividual(ind, true);
	}

	/**
	 * Add individual to memory
	 * 
	 * @param ind
	 *            Individual to add
	 * @param checkUnique
	 *            Check is individual unique (Attention! Duplicates variable-set are
	 *            not checked if checkUnique is false)
	 */
	public void addIndividual(Individual ind, boolean checkUnique) {
		int index;

		if (checkUnique) {
			index = this.individuals.indexOf(ind);
			if (index == -1) {
				this.individuals.add(ind);
			} else {
				this.individuals.set(index, ind);
			}
		} else {
			this.individuals.add(ind);
		}
	}

	/**
	 * Get Individual from memory (with results)
	 * 
	 * @param ind
	 *            Individual with variables
	 * @return Individual with results
	 */
	public Individual getIndividual(Individual ind) {
		int index = -1;
		Individual res = null;

		if (ind.getArchive() == this.archive) {
			index = this.individuals.indexOf(ind);
		}

		if (index != -1) {
			res = this.individuals.get(index);
		}

		return res;
	}

	/**
	 * Get individual of index
	 * 
	 * @param index
	 *            Index of individual
	 * @return Individual from archive
	 */
	public Individual getIndividual(int index) {
		Individual ind = null;

		if ((0 <= index) && (index < this.individuals.size())) {
			ind = this.individuals.get(index);
		}

		return ind;
	}

	/**
	 * Get individual-array of archive-index
	 * 
	 * @param index
	 *            Index-array of individual
	 * @return Individual-array from archive
	 */
	public Individual[] getIndividual(int[] index) {
		Individual[] ind;

		ind = new Individual[index.length];
		for (int k = 0; k < index.length; k++) {
			ind[k] = this.getIndividual(index[k]);
		}

		return ind;
	}

	/**
	 * Get archive-index of individual
	 * 
	 * @param ind
	 *            Individual
	 * @return Index of individual
	 */
	public int getIndexOf(Individual ind) {
		return this.individuals.indexOf(ind);
	}

	/**
	 * Get archive-index of individual-array
	 * 
	 * @param ind
	 *            Individual-array
	 * @return Index-array of individuals
	 */
	public int[] getIndexOf(Individual[] ind) {
		int[] index;

		index = new int[ind.length];
		for (int k = 0; k < ind.length; k++) {
			index[k] = this.getIndexOf(ind[k]);
		}

		return index;
	}

	/**
	 * Get array of all individuals
	 * 
	 * @return All individuals
	 */
	public Individual[] getIndividual() {
		return this.individuals.toArray(new Individual[0]);
	}

	/**
	 * Get array individuals for recompute
	 * 
	 * @return All individuals to recompute
	 */
	public Individual getNextRecomputeIndividual() {
		Individual ind;

		for (int k = 0; k < this.individuals.size(); k++) {
			ind = this.individuals.get(k);
			if (ind.isFromMapping() && !ind.isRecompute()) {
				return ind;
			}
		}

		return null;
	}

	/**
	 * Get array of individuals without errors
	 * 
	 * @return No error individuals
	 */
	public Individual[] getNoErrorIndividual() {
		LinkedList<Individual> list;
		Individual ind;

		list = new LinkedList<Individual>();

		for (int k = 0; k < this.individuals.size(); k++) {
			ind = this.individuals.get(k);
			if (!ind.isError()) {
				list.add(ind);
			}
		}

		return list.toArray(new Individual[0]);
	}

	/**
	 * Get array of individuals with errors
	 * 
	 * @return No error individuals
	 */
	public Individual[] getErrorIndividual() {
		LinkedList<Individual> list;
		Individual ind;

		list = new LinkedList<Individual>();

		for (int k = 0; k < this.individuals.size(); k++) {
			ind = this.individuals.get(k);
			if (ind.isError()) {
				list.add(ind);
			}
		}

		return list.toArray(new Individual[0]);
	}

	/**
	 * Get array of valid individuals
	 * 
	 * @param problem
	 *            Problem to check valid
	 * @return Valid individuals
	 */
	public Individual[] getValidIndividual(Problem problem) {
		LinkedList<Individual> list;
		Individual ind;

		list = new LinkedList<Individual>();

		for (int k = 0; k < this.individuals.size(); k++) {
			ind = this.individuals.get(k);
			if (ind.isValid(problem)) {
				list.add(ind);
			}
		}

		return list.toArray(new Individual[0]);
	}

	/**
	 * Get array of pareto individuals
	 * 
	 * @param problem
	 *            Problem to check pareto
	 * @return Pareto individuals
	 */
	public Individual[] getParetoIndividual(Problem problem) {
		return this.getParetoBySplitMerge(problem, this.individuals).toArray(new Individual[0]);
	}

	/**
	 * Calculate pareto-front by split and merge
	 * 
	 * @param problem
	 *            Problem to check pareto
	 * @param inds
	 *            List of individuals to check
	 * @return List of paretos
	 */
	private List<Individual> getParetoBySplitMerge(Problem problem, List<Individual> inds) {
		List<Individual> list;
		Individual ind;
		int comp;
		boolean dominated;

		if (inds.size() > 1000) {
			// split individuals
			list = this.getParetoBySplitMerge(problem, inds.subList(0, inds.size() / 2));
			inds = this.getParetoBySplitMerge(problem, inds.subList(inds.size() / 2, inds.size()));
		} else {
			list = new LinkedList<Individual>();
		}

		// insert inds to list (merge)
		for (int n = 0; n < inds.size(); n++) {
			ind = inds.get(n);
			if (ind.isValid(problem)) {
				dominated = false;
				for (int k = list.size() - 1; k >= 0; k--) {
					comp = ind.dominate(list.get(k), problem);
					if (comp == -1) {
						list.remove(k);
					} else if (comp == 1) {
						dominated = true;
					}
				}
				if (!dominated) {
					list.add(ind);
				}
			}
		}

		return list;
	}

	/**
	 * Get limits of valid objective values
	 * 
	 * @param problem
	 *            Problem to check valid
	 * @return Limits of valid objectives
	 */
	public double[][] getObjectivesLimitsValid(Problem problem) {
		Individual[] list;
		double[][] limits;
		double[] values;

		limits = new double[2][problem.getNumObjectives()];

		list = this.getValidIndividual(problem);
		for (int k = 0; k < limits[0].length; k++) {
			limits[0][k] = Double.POSITIVE_INFINITY;
			limits[1][k] = Double.NEGATIVE_INFINITY;
		}

		for (Individual ind : list) {
			values = ind.getObjectiveValues(problem);
			for (int k = 0; k < values.length; k++) {
				if (values[k] < limits[0][k]) {
					limits[0][k] = values[k];
				}

				if (values[k] > limits[1][k]) {
					limits[1][k] = values[k];
				}
			}
		}

		return limits;
	}

	/**
	 * Get limits of pareto objective values
	 * 
	 * @param problem
	 *            Problem to check valid
	 * @return Limits of pareto objectives
	 */
	public double[][] getObjectivesLimitsPareto(Problem problem) {
		Individual[] list;
		double[][] limits;
		double[] values;

		limits = new double[2][problem.getNumObjectives()];

		list = this.getParetoIndividual(problem);
		for (int k = 0; k < limits[0].length; k++) {
			limits[0][k] = Double.POSITIVE_INFINITY;
			limits[1][k] = Double.NEGATIVE_INFINITY;
		}

		for (Individual ind : list) {
			values = ind.getObjectiveValues(problem);
			for (int k = 0; k < values.length; k++) {
				if (values[k] < limits[0][k]) {
					limits[0][k] = values[k];
				}

				if (values[k] > limits[1][k]) {
					limits[1][k] = values[k];
				}
			}
		}

		return limits;
	}

	/**
	 * Get resume data
	 * 
	 * @return Resume data
	 * @throws Exception
	 */
	TArray getResumeData() throws Exception {
		TArray resumeData;

		resumeData = new TArray(1, this.individuals.size());
		for (int k = 0; k < this.individuals.size(); k++) {
			resumeData.set(0, k, this.individuals.get(k).getResumeData());
		}

		return resumeData;
	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume Data
	 * @throws TDataException
	 */
	void setResumeData(TArray resumeData) throws TDataException {
		if (resumeData != null) {
			this.clear();
			for (int k = 0; k < resumeData.size()[1]; k++) {
				this.individuals.add(new Individual(this.archive, (TStruct) resumeData.get(0, k)));
			}
		}
	}

	@Override
	public String toString() {
		return "Memory:\n\t" + this.getNumIndividuals() + " individuals computed\n";
	}
}
