/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.Serializable;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * Description of result. Range of valid results and is used as objective
 * 
 * @author Werner Koppelsttter
 */
public class ResultDesc implements Serializable {

	/** Name of result */
	private String name;

	/** Is result an objective */
	private ObjectiveType objective = ObjectiveType.NONE;

	/** Lower limit of result constraint */
	private double lb;

	/** Upper limit of result constraint */
	private double ub;

	/**
	 * Create a result from resumeData
	 * 
	 * @param resumeData
	 *            ResumeData
	 * @throws TDataException
	 */
	ResultDesc(TStruct resumeData) throws TDataException {
		this.setResumeData(resumeData);
	}

	/**
	 * Create a result for name
	 * 
	 * @param name
	 *            Name of result
	 */
	public ResultDesc(String name) {
		this(name, ObjectiveType.NONE, Double.NaN, Double.NaN);
	}

	/**
	 * Create a result as objective for name
	 * 
	 * @param name
	 *            Name of result
	 * @param objecitve
	 *            Is Objective
	 */
	public ResultDesc(String name, ObjectiveType objecitve) {
		this(name, objecitve, Double.NaN, Double.NaN);
	}

	/**
	 * Create a result as contratint for name
	 * 
	 * @param name
	 *            Name of result
	 * @param lb
	 *            Lower-boundary of constraint
	 * @param ub
	 *            Upper-boundary of constraint
	 */
	public ResultDesc(String name, double lb, double ub) {
		this(name, ObjectiveType.NONE, lb, ub);
	}

	/**
	 * Create a result as objective and constraint for name
	 * 
	 * @param name
	 *            Name of result
	 * @param objecitve
	 *            Is objective
	 * @param lb
	 *            Lower-boundary of constraint
	 * @param ub
	 *            Upper-boundary of constraint
	 */
	public ResultDesc(String name, ObjectiveType objecitve, double lb, double ub) {
		this.name = name;
		this.setObjective(objecitve);
		this.setBounds(lb, ub);
	}

	/**
	 * Get name of result
	 * 
	 * @return Name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Is result an objective
	 * 
	 * @return Is objective
	 */
	public boolean isObjective() {
		return this.objective != ObjectiveType.NONE;
	}

	/**
	 * Get objectibe type of result
	 * 
	 * @return Is objective
	 */
	public ObjectiveType getObjective() {
		return this.objective;
	}

	/**
	 * Set objective type of result
	 * 
	 * @param objecitve
	 *            Is Objective
	 */
	public void setObjective(ObjectiveType objecitve) {
		this.objective = objecitve;
	}

	/**
	 * Check if result is a constraint (lower-boundary and upper-boundary exists)
	 * 
	 * @return Is constraint
	 */
	public boolean isConstraint() {
		return !Double.isNaN(this.lb) && !Double.isNaN(this.ub) && (this.lb < this.ub);
	}

	/**
	 * Get lower boundary of contraint
	 * 
	 * @return Lower boundary
	 */
	public double getLB() {
		return this.lb;
	}

	/**
	 * Get upper boundary of contraint
	 * 
	 * @return Upper boundary
	 */
	public double getUB() {
		return this.ub;
	}

	/**
	 * Set boundary of constraints
	 * 
	 * @param lb
	 *            Lower limit of valid value
	 * @param ub
	 *            Upper limit of valid value
	 */
	public void setBounds(double lb, double ub) {
		this.lb = lb;
		this.ub = ub;
	}

	/**
	 * Is value valid. Is the value between the boundaries of constraint
	 * 
	 * @param val
	 *            Value to check
	 * @return Is value valid
	 */
	public boolean isValidValue(double val) {
		boolean valid;

		valid = !Double.isNaN(val);
		if (this.isConstraint()) {
			valid &= (this.lb <= val) && (val <= this.ub);
		}

		return valid;
	}

	/**
	 * Check if result definition is valid (completely defined)
	 * 
	 * @return Is valid
	 */
	public boolean isValid() {
		boolean valid;

		valid = Double.isNaN(this.lb) && Double.isNaN(this.ub);
		valid |= this.isConstraint();

		return valid;
	}

	/**
	 * Get resume data
	 * 
	 * @return Resume data
	 * @throws TDataException
	 */
	TStruct getResumeData() throws TDataException {
		TStruct resumeData;

		resumeData = new TStruct();
		resumeData.put("name", new TString(this.name));
		resumeData.put("obj", new TString(this.objective.name()));
		resumeData.put("lb", new TNumeric(this.lb));
		resumeData.put("ub", new TNumeric(this.ub));

		return resumeData;
	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume data
	 * @throws TDataException
	 */
	private void setResumeData(TStruct resumeData) throws TDataException {
		this.name = ((TString) resumeData.get("name")).getString();
		this.objective = ObjectiveType.valueOf(((TString) resumeData.get("obj")).getString());
		this.lb = ((TNumeric) resumeData.get("lb")).getReal(0, 0);
		this.ub = ((TNumeric) resumeData.get("ub")).getReal(0, 0);
	}

	@Override
	public String toString() {
		StringBuffer str;

		str = new StringBuffer();
		str.append(this.name);
		str.append(":\t");
		str.append(this.objective);
		if (this.isConstraint()) {
			str.append("\t");
			str.append(this.lb);
			str.append(" : ");
			str.append(this.ub);
		}
		str.append("\n");

		return str.toString();
	}
}
