/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.Date;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * Savepoint with current pareto-fronts
 * 
 * @author Werner Koppelsttter
 */
public class SavePoint implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Date-formatter for reading and writing date */
	protected static SimpleDateFormat sdf = new SimpleDateFormat("dd-MM-yyyy HH:mm");

	/**
	 * Type of savepoint
	 */
	public enum SavePointType {
		/** User defined */
		USER,
		/** Interval-based */
		INTERVAL,
		/** Time-based */
		TIME;
	}

	/** Problem of savepoint */
	private Problem problem;

	/** Data of savepoint */
	private Date date;

	/** Type of savepoint */
	private SavePointType type;

	/** Full pareto-front */
	private Individual[] fullPareto;

	/** Local pareto-front */
	private Individual[] localPareto;

	/** Number of valid individuals */
	private int nValid = 0;

	/** Number of invalid individuals */
	private int nInvalid = 0;

	/** Number of individuals with errors */
	private int nError = 0;

	/**
	 * Create new savepoint of project
	 * 
	 * @param type
	 *            Type of savepoint
	 * @param problem
	 *            Problem of savepoint
	 * @param nValid
	 *            Number of valid individuals
	 * @param nInvalid
	 *            Number of invalid individuals
	 * @param nError
	 *            Number of individuals with errors
	 * @throws Exception
	 */
	SavePoint(SavePointType type, Problem problem, int nValid, int nInvalid, int nError) throws Exception {
		this(type, problem, nValid, nInvalid, nError, true);
	}

	/**
	 * Create new savepoint of project
	 * 
	 * @param type
	 *            Type of savepoint
	 * @param problem
	 *            Problem of savepoint
	 * @param nValid
	 *            Number of valid individuals
	 * @param nInvalid
	 *            Number of invalid individuals
	 * @param nError
	 *            Number of individuals with errors
	 * @param calcPareto
	 *            Calulate pareto-front or use all valid individuals
	 * @throws Exception
	 */
	SavePoint(SavePointType type, Problem problem, int nValid, int nInvalid, int nError, boolean calcPareto)
			throws Exception {
		Memory mem;
		Individual[] inds;

		this.date = new Date();
		this.type = type;
		this.problem = problem;

		mem = problem.getArchive().getMemory();
		if (calcPareto) {
			this.fullPareto = mem.getParetoIndividual(problem);
		} else {
			this.fullPareto = mem.getValidIndividual(problem);
		}
		inds = problem.getSolver().getPareto();
		for (int k = 0; k < inds.length; k++) {
			inds[k] = mem.getIndividual(inds[k]);
		}
		this.localPareto = inds;
		this.nValid = nValid;
		this.nInvalid = nInvalid;
		this.nError = nError;
	}

	/**
	 * Resume savepoint
	 * 
	 * @param problem
	 *            Problem of savepoint
	 * @param resumeData
	 *            Resumedata of savepoint
	 * @throws Exception
	 */
	SavePoint(Problem problem, TStruct resumeData) throws Exception {
		this.problem = problem;
		this.setResumeData(resumeData);
	}

	/**
	 * Get resume data
	 * 
	 * @return Resume data
	 * @throws Exception
	 */
	TStruct getResumeData() throws Exception {
		TStruct resumeData;
		TData data;

		resumeData = new TStruct();

		// time
		if (this.date == null) {
			data = new TNull();
		} else {
			data = new TString(SavePoint.sdf.format(this.date));
		}
		resumeData.put("date", data);

		// type
		resumeData.put("type", new TString(this.type.name()));

		// full paretos
		if (this.fullPareto.length == 0) {
			data = new TNull();
		} else {
			data = new TNumeric(Archive.int2double(this.problem.getArchive().getMemory().getIndexOf(this.fullPareto)));
		}
		resumeData.put("full", data);

		// full paretos
		if (this.localPareto.length == 0) {
			data = new TNull();
		} else {
			data = new TNumeric(Archive.int2double(this.problem.getArchive().getMemory().getIndexOf(this.localPareto)));
		}
		resumeData.put("local", data);

		// number of valid
		resumeData.put("nValid", new TString(Integer.toString(this.nValid)));

		// number of invalid
		resumeData.put("nInvalid", new TString(Integer.toString(this.nInvalid)));

		// number of error
		resumeData.put("nError", new TString(Integer.toString(this.nError)));

		return resumeData;
	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume Data
	 * @throws Exception
	 */
	private void setResumeData(TStruct resumeData) throws Exception {
		TData data;

		// time
		data = resumeData.get("date");
		if (data instanceof TNull) {
			this.date = null;
		} else {
			this.date = SavePoint.sdf.parse(((TString) data).getString());
		}

		// type
		this.type = SavePointType.valueOf(((TString) resumeData.get("type")).getString());

		// full paretos
		data = resumeData.get("full");
		if (data instanceof TNull) {
			this.fullPareto = new Individual[0];
		} else {
			this.fullPareto = this.problem.getArchive().getMemory()
					.getIndividual(Archive.double2int(((TNumeric) data).getRealMatrix().getRow(0)));
		}

		// local paretos
		data = resumeData.get("local");
		if (data instanceof TNull) {
			this.localPareto = new Individual[0];
		} else {
			this.localPareto = this.problem.getArchive().getMemory()
					.getIndividual(Archive.double2int(((TNumeric) data).getRealMatrix().getRow(0)));
		}

		// number of valid
		data = resumeData.get("nValid");
		if (data != null) {
			this.nValid = Integer.parseInt(((TString) data).getString());
		}

		// number of invalid
		data = resumeData.get("nInvalid");
		if (data != null) {
			this.nInvalid = Integer.parseInt(((TString) data).getString());
		}

		// number of error
		data = resumeData.get("nError");
		if (data != null) {
			this.nError = Integer.parseInt(((TString) data).getString());
		}
	}

	/**
	 * Get date of savepoint
	 * 
	 * @return Date
	 */
	public Date getDate() {
		return this.date;
	}

	/**
	 * Get type of savepoint
	 * 
	 * @return Type
	 */
	public SavePointType getType() {
		return this.type;
	}

	/**
	 * Get full pareto-front
	 * 
	 * @return Full pareto
	 */
	public Individual[] getFullPareto() {
		return this.fullPareto;
	}

	/**
	 * Get local pareto-front
	 * 
	 * @return Local pareto
	 */
	public Individual[] getLocalPareto() {
		return this.localPareto;
	}

	/**
	 * Get number of valid individuals
	 * 
	 * @return Number of valids
	 */
	public int getNValid() {
		return this.nValid;
	}

	/**
	 * Get number of invalid individuals
	 * 
	 * @return Number of invalids
	 */
	public int getNInvalid() {
		return this.nInvalid;
	}

	/**
	 * Get number of individuals with errors
	 * 
	 * @return Number of errors
	 */
	public int getNError() {
		return this.nError;
	}

	@Override
	public String toString() {
		StringBuffer str;

		str = new StringBuffer();
		str.append("SavePoint:\n");
		str.append("\tType:\t " + this.type + "\n");
		str.append("\tDate:\t " + this.date + "\n");

		return str.toString();
	}
}
