/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.jobs;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Random;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants;
import jobmanager.JobDesc;
import jobmanager.JobStatus;
import jobmanager.JobType;

/**
 * Dispatcher class for mapping jobs
 * 
 * @author Werner Koppelsttter / Ciprian Zavoianu
 */
public class MappingJobDispatcher {

	private final String mappingJobDispatcher = "MappingJobDispatcher_";

	public boolean checkAllJobsFinished(String uniqueOptimizerRunId) {
		return JobDesc.allFinished(this.mappingJobDispatcher + uniqueOptimizerRunId, null, null);
	}

	public String createJob(String uniqueOptimizerRunId, File inputFile, String nrOfFolds, int jobID) {
		// create job

		int runID = (new Random((new Random()).nextLong())).nextInt(1000);

		String jobId = JobDesc.createJob(this.mappingJobDispatcher + uniqueOptimizerRunId, "", "",
				"MappingJob_" + jobID + "_" + runID, MappingJob.class.getName(), inputFile.getName(), nrOfFolds);
		try {
			JobDesc.addFile(jobId,
					FileUtils.getFile(Project.getInstallPath(), "Tools", "JobManager", "Java", "jobmanager.jar"));
			JobDesc.linkFile(jobId, FileUtils.getFile(Project.getInstallPath(), "System", "Java", "magopt.jar"));
			JobDesc.linkFile(jobId,
					FileUtils.getFile(Project.getInstallPath(), "System", "General", "Java", "General.jar"));
			JobDesc.linkFile(jobId,
					FileUtils.getFile(Project.getInstallPath(), "External", "javasysmon", "javasysmon-0.3.3.jar"));

			JobDesc.linkFile(jobId, FileUtils.getFile(Project.getInstallPath(), "System", "Optimizer", "External",
					"opencv", "opencv_extension.jar"));
			JobDesc.linkFile(jobId, FileUtils.getFile(Project.getInstallPath(), "System", "Optimizer", "External",
					"opencv", "opencv-249.jar"));

			JobDesc.linkFile(jobId,
					FileUtils.getFile(Project.getInstallPath(), "System", "Optimizer", "Java", "optimizer.jar"));

			// copy local test description file
			JobDesc.addFile(jobId, inputFile);

			File dir = new File(inputFile.getParent());
			if (!dir.isDirectory()) {
				throw new Exception("Error: input directory could not be found!");
			} else {
				FilenameFilter filefilter = new FilenameFilter() {

					@Override
					public boolean accept(File dir, String name) {
						return name.endsWith(".arff");
					}
				};
				String[] arffFilenames = dir.list(filefilter);
				for (String name : arffFilenames) {
					JobDesc.addFile(jobId, new File(dir.getAbsolutePath() + "\\" + name));
				}
			}

			// set java version requirement
			JobDesc.setRequirement(jobId, "JavaVersion >= \"1.8.0_\"");

			// job timeout 180 minutes
			JobDesc.setTimeout(jobId, 180 * 60);

			JobDesc.submit(jobId, JobType.CLUSTER);
		} catch (Exception e) {
			e.printStackTrace();
		}

		return jobId;
	}

	public void getAllJobResults(String jobId, File localTaskBatchFileName, File localRegressionModelFileName)
			throws IOException {
		JobDesc.getFile(jobId, TMappingConstants.REMOTE_TASK_BATCH_FILE_NAME, localTaskBatchFileName);
		JobDesc.getFile(jobId, TMappingConstants.REMOTE_MODEL_FILE_NAME, localRegressionModelFileName);
		JobDesc.remove(jobId);
	}

	public int[] getJobStatusCounts(String uniqueOptimizerRunId) {
		Hashtable<String, JobStatus> list;
		int[] result = { 0, 0, 0, 0 };

		list = JobDesc.getStatusList(this.mappingJobDispatcher + uniqueOptimizerRunId, null, null);

		if (list != null) {
			for (JobStatus status : list.values()) {
				if (status.equals(JobStatus.SUBMITTED)) {
					result[0]++;
				}
				if (status.equals(JobStatus.RUNNING)) {
					result[1]++;
				}
				if (status.equals(JobStatus.COMPLETED) || status.equals(JobStatus.ERROR)
						|| status.equals(JobStatus.TIMEOUT)) {
					result[2]++;
				}
				if (status.equals(JobStatus.HELD)) {
					result[3]++;
				}
			}
		}

		return result;
	}

	public void getJobTestSetResults(String jobId, File localTaskBatchFileName) throws IOException {
		JobDesc.getFile(jobId, TMappingConstants.REMOTE_TASK_BATCH_FILE_NAME, localTaskBatchFileName);
		JobDesc.remove(jobId);
	}

	public boolean isJobCompleted(String jobIdentifier) {
		return (JobStatus.COMPLETED == JobDesc.getStatus(jobIdentifier));
	}

	public boolean isJobFinished(String jobIdentifier) {
		return ((JobStatus.COMPLETED == JobDesc.getStatus(jobIdentifier))
				|| (JobStatus.TIMEOUT == JobDesc.getStatus(jobIdentifier))
				|| (JobStatus.ERROR == JobDesc.getStatus(jobIdentifier))
				|| (JobStatus.REMOVED == JobDesc.getStatus(jobIdentifier)));
	}

	public void removeAllRemainingJobs(String uniqueOptimizerRunId) {
		Hashtable<String, JobStatus> list = JobDesc.getStatusList(this.mappingJobDispatcher + uniqueOptimizerRunId,
				null, null);
		if (list != null) {
			for (String jobId : list.keySet()) {
				JobDesc.remove(jobId);
			}
		}
	}

	public void restartLocalAllIdleCondorJobs(String uniqueOptimizerRunId) {
		Hashtable<String, JobStatus> list = JobDesc.getStatusList(this.mappingJobDispatcher + uniqueOptimizerRunId,
				null, null);
		// int count = 0;
		for (String s : list.keySet()) {
			if (list.get(s).equals(JobStatus.SUBMITTED)) {
				JobDesc.restart(s, JobType.LOCAL);
				// count++;
			}
		}
		// System.out.println("Restarted on Local Machine " + count
		// + " idle jobs.");
	}

	public void restartOnCondorAllHeldJobs(String uniqueOptimizerRunId) {
		Hashtable<String, JobStatus> list;
		list = JobDesc.getStatusList(this.mappingJobDispatcher + uniqueOptimizerRunId, null, null);
		// int count = 0;
		for (String s : list.keySet()) {
			if (list.get(s).equals(JobStatus.SUBMITTED)) {
				JobDesc.restart(s, JobType.CLUSTER);
				// count++;
			}
		}
		// System.out.println("Restarted on Condor " + count + " held jobs.");
	}
}
