#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9896 $
$LastChangedDate:: 2020-06-30 08:41:38 +0200#$

This file is part of SyMSpace.

Copyright (c) 2019, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import pickle
from symspace.parampythonfnc import ParamPythonFnc


def debugexternal(Cls: ParamPythonFnc):
    """
    Decorator for debugging of SyMSpace functions (ParamPythonFnc)
    """

    class DebugWrapper(Cls):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def run(self, *args, **kwargs) -> int:
            """ Run method for debugging """
            if self.debug_mode:
                self.importValues = self.importValuesDebug
                self.exportValues = self.exportValuesDebug
                self.info = self.infoDebug
                self.warning = self.warningDebug
                self.error = self.errorDebug
                return super().run(*args, **kwargs)
            else:
                class_path = self._structure.getClassPath()
                class_name = self._structure.getClassName()
                module_name = self._structure.getModuleName()
                fnc_debug = 'debug_' + module_name.replace('fnc_', '') + '.py'
                fnc_debug_path = os.path.join(class_path, fnc_debug)

                if not os.path.isfile(fnc_debug_path):
                    run_args = args[0] if len(args) == 1 else {}
                    remote_compute = run_args['RemoteCompute'] if 'RemoteCompute' in run_args else False
                    run_sim = run_args['RunSim'] if 'RunSim' in run_args else False
                    optimize = run_args['Optimize'] if 'Optimize' in run_args else False

                    with open(fnc_debug_path, 'w') as fh:
                        '''write debug starter file'''
                        fh.write(f'from {module_name} import {class_name} as DebugFnc\n\n')
                        fh.write('dbg = DebugFnc(structure=None)\n')
                        fh.write('dbg.debug_mode = True\n')
                        fh.write(f'run_args = dict(RemoteCompute={remote_compute}, '
                                 f'RunSim={run_sim}, '
                                 f'Optimize={optimize})\n')
                        fh.write('dbg.run(run_args)\n')

                pklfile = os.path.join(self._structure.getClassPath(), 'inputParameter.p')
                with open(pklfile, 'wb') as fh:
                    pickle.dump(self.importValues(), fh)

                self.info('Open {} for debugging with Python debugger'.format(fnc_debug_path))

                return super().run(*args, **kwargs)

        def importValuesDebug(self):
            pklfile = 'inputParameter.p'
            if os.path.isfile(pklfile):
                with open(pklfile, 'rb') as fh:
                    data = pickle.load(fh)
            else:
                print('Pickle file not found')
                data = {}
            return data

        def exportValuesDebug(self, d):
            self.return_values = d

        def infoDebug(self, message):
            print(message)

        def warningDebug(self, message):
            print(message)

        def errorDebug(self, message):
            raise Exception(message)

    return DebugWrapper


def remotecompute(requires):
    """
    Decorator to distribute run methode to remote instance

    Parameters
    ----------
    requires : str
        Example:
        - @remotecompute('(PYTHON >= \"3.5.3\")') requires Python >= 3.5.3 at remote instance
        or with Python in java
        - @remotecompute('(PYTHON >= \"3.5.3\") && (JEP >= \"3.6.3\")') requires Python >= 3.5.3 and JEP >= 3.6.3 at remote instance

    """

    def wrap(Cls: ParamPythonFnc):

        class RemoteWrapper(Cls):

            def __init__(self, *args, **kwargs):
                self.data = {}
                super().__init__(*args, **kwargs)

            def isDistributable(self) -> bool:
                """Return distributable flag"""
                return True

            # def getUniverse(self):
            #    return 'DOCKER'

            def getJobRequirement(self):
                """Get Condor compatible string for remote machine requirements"""
                return requires

            def getFile(self) -> str:
                """Return current Python script"""
                class_path = self._structure.getClassPath()
                module_name = self._structure.getModuleName()
                file = os.path.join(class_path, module_name + '.py')
                return file

            def getAdditionalModules(self):
                """List of additional files to add"""
                lstModules = super().getAdditionalModules()
                if lstModules is None:
                    lstModules = []
                return lstModules

            def importValuesRemote(self):
                return self.data

            def exportValuesRemote(self, d):
                for key in d:
                    self.data[key] = d[key]

            def runJob(self, d, runargs) -> int:
                if hasattr(Cls, 'runJob'):
                    status = super().runJob(d, runargs)
                else:
                    self.importValues = self.importValuesRemote
                    self.exportValues = self.exportValuesRemote
                    self.data = d
                    status = super().run(runargs)
                return status

            def run(self, runargs):
                return ParamPythonFnc.run(self, runargs)

        return RemoteWrapper

    return wrap


def remotedockercompute(image):
    """
    Decorator to distribute run methode to remote docker instance

    Parameters
    ----------
    image : str
        Example:
        - @remotecompute('python:3.7-openjdk13') docker image with python at remote instance

    """

    def wrapDocker(Cls: ParamPythonFnc):

        class RemoteDockerWrapper(Cls):

            def __init__(self, *args, **kwargs):
                self.data = {}
                super().__init__(*args, **kwargs)
                print('init Docker')

            def isDistributable(self) -> bool:
                """Return distributable flag"""
                return True

            def getUniverse(self) -> str:
                return 'DOCKER'

            def getDockerImage(self) -> str:
                return image

            def getJobRequirement(self) -> str:
                """Get Condor compatible string for remote machine requirements"""
                return '(TARGET.OPSYS == "LINUX")'

            def getFile(self) -> str:
                """Return current Python script"""
                class_path = self._structure.getClassPath()
                module_name = self._structure.getModuleName()
                file = os.path.join(class_path, module_name + '.py')
                return file

            def getAdditionalModules(self):
                """List of additional files to add"""
                lstModules = super().getAdditionalModules()
                if lstModules is None:
                    lstModules = []
                return lstModules

            def importValuesRemote(self):
                return self.data

            def exportValuesRemote(self, d):
                for key in d:
                    self.data[key] = d[key]

            def runJob(self, d, runargs) -> int:
                if hasattr(Cls, 'runJob'):
                    status = super().runJob(d, runargs)
                else:
                    # ToDo: if remote compute then ...
                    self.importValues = self.importValuesRemote
                    self.exportValues = self.exportValuesRemote
                    self.data = d
                    status = super().run(runargs)
                return status

            def run(self, runargs) -> int:
                return ParamPythonFnc.run(self, runargs)

        return RemoteDockerWrapper

    return wrapDocker


if __name__ == '__main__':
    pass
