#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 9874 $
$LastChangedDate:: 2020-06-12 18:50:19 +0200#$

This file is part of SyMSpace.

Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
from threading import Thread
import symspace.gui.qt
symspace.gui.qt.set_QT_PLUGIN_PATH()
from qtpy import QtGui
from qtpy.QtWidgets import (QApplication, QComboBox, QDialog,
                            QDialogButtonBox, QFormLayout, QGroupBox, QHBoxLayout,
                            QLabel, QLineEdit, QPushButton,
                            QVBoxLayout, QCheckBox, QFileDialog, QMessageBox, QScrollArea)
from symspace.osutils.pathutils import symspace_home


class FormDialog(QDialog):

    def __init__(self, title, form_elements, width=640, height=480):
        """
        Create a generic form dialog

        Parameters
        ----------
        title : string
            Title of the form
        form_elements : list
            Form elements
            list_elements: list
                - index 0 : string
                    Label
                - index 1 : string
                    Type : ['label', 'line', 'combobox', 'checkbox', 'browse_file', 'browse_folder', 'save_file]
                - index 2 : string, list or bool
                    Default value

        width : int
            Width of the form, 640 (Default)
        height: int
            Height of the form, 480 (Default)
        """
        super(FormDialog, self).__init__()
        self.title = title
        self.width = width
        self.height = height
        self.form_elements = form_elements
        self.entry = {}
        self.formGroupBox = QGroupBox("Form layout")
        self.initUI()

    def initUI(self):
        """Initialize user interface"""
        self.setWindowTitle(self.title)
        self.resize(self.width, self.height)

        self.__addFormLayout()
        scroll_area = self.__createScrollArea()
        buttonBox = self.__createButtonBox()

        mainLayout = QVBoxLayout()
        mainLayout.addWidget(scroll_area)
        mainLayout.addWidget(buttonBox)
        self.setLayout(mainLayout)

    def __createScrollArea(self):
        scroll_area = QScrollArea()
        scroll_area.setWidget(self.formGroupBox)
        scroll_area.setWidgetResizable(True)
        return scroll_area

    def __createButtonBox(self):
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        return buttonBox

    def __addFormLayout(self):

        form_layout = QFormLayout()

        for row, elem in enumerate(self.form_elements):
            key = elem[0]
            option = elem[1]
            if len(elem) == 2:
                default_value = ''
            else:
                default_value = elem[2]

            if option == 'label':
                self.entry[key] = QLabel(default_value)
                widget = self.entry[key]
            elif option == 'line':
                self.entry[key] = QLineEdit(default_value)
                widget = self.entry[key]
            elif option == 'combobox':
                self.entry[key] = QComboBox()
                self.entry[key].addItems(default_value)
                widget = self.entry[key]
            elif option == 'browse' or option == 'browse_folder':
                self.entry[key] = QLineEdit(default_value)
                browse_button = QPushButton('Browse')
                browse_button.clicked.connect(lambda state, button_key=key: self.browseDirectory(button_key))

                hbox = QHBoxLayout()
                hbox.addWidget(self.entry[key])
                hbox.addWidget(browse_button)
                widget = hbox
            elif option == 'browse_file':
                file_filter = 'All Files(*)'
                if len(elem) == 4:
                    file_filter = elem[3] + ';;' + file_filter

                self.entry[key] = QLineEdit(default_value)
                browse_button = QPushButton('Browse')
                browse_button.clicked.connect(
                    lambda state, button_key=key, ffilter=file_filter: self.browseFile(button_key, ffilter))

                hbox = QHBoxLayout()
                hbox.addWidget(self.entry[key])
                hbox.addWidget(browse_button)
                widget = hbox
            elif option == 'save_file':
                file_filter = 'All Files(*)'
                if len(elem) == 4:
                    file_filter = elem[3] + ';;' + file_filter

                self.entry[key] = QLineEdit(default_value)
                browse_button = QPushButton('Browse')
                browse_button.clicked.connect(
                    lambda state, button_key=key, ffilter=file_filter: self.saveFile(button_key, ffilter))

                hbox = QHBoxLayout()
                hbox.addWidget(self.entry[key])
                hbox.addWidget(browse_button)
                widget = hbox
            elif option == 'checkbox':
                self.entry[key] = QCheckBox()
                if default_value:
                    self.entry[key].setChecked(True)
                widget = self.entry[key]

            else:
                raise Exception('Option unknown:' + str(option))

            label = QLabel(key)
            # if len(elem) == 4 and elem[3] != '':
            #     label.setToolTip(elem[3])
            form_layout.addRow(label, widget)

        self.formGroupBox.setLayout(form_layout)

    def browseDirectory(self, button_key):
        """Browse directory button"""
        previous_dir = self.entry[button_key].text()
        dir_name = QFileDialog.getExistingDirectory(self, "Select Directory", previous_dir,
                                                    QFileDialog.ShowDirsOnly)
        if str(dir_name) != '':
            self.entry[button_key].setText(os.path.abspath(dir_name))

    def browseFile(self, button_key, file_filter):
        """Browse file button"""
        previous_file = self.entry[button_key].text()
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        file_name, _ = QFileDialog.getOpenFileName(self, "Select File", previous_file, file_filter)
        if str(file_name) != '':
            self.entry[button_key].setText(os.path.abspath(file_name))

    def saveFile(self, button_key, file_filter):
        """Save file button"""
        previous_file = self.entry[button_key].text()
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        file_name, _ = QFileDialog.getSaveFileName(self, "Select File", previous_file, file_filter)
        if str(file_name) != '':
            self.entry[button_key].setText(os.path.abspath(file_name))

    def getReturnValues(self):
        """Get the return values of the widgets"""
        result = {}
        for row, elem in enumerate(self.form_elements):
            key = elem[0]
            option = elem[1]
            if option == 'checkbox':
                result[key] = False if self.entry[key].checkState() == 0 else True
            elif option == 'combobox':
                result[key] = self.entry[key].currentText()
            else:
                result[key] = self.entry[key].text()
        return result


class InfoDialog(QMessageBox):

    def __init__(self, title, msg_elements, width=640, height=480, *args, **kwargs):
        """
        Create a generic info dialog

        Parameters
        ----------
        title : string
            Title of the form
        msg_elements : dict
            Message elements
            Keywords :
                - icon : string
                    ['info', 'question', 'warning', 'critical']
                - text : string
                - info_text : string
                - detail_text : string
                - buttons : string
                    ['yes_no', 'ok_cancel', 'ok']
        width : int
            Width of the form, 640 (Default)
        height: int
            Height of the form, 480 (Default)

        """
        super(InfoDialog, self).__init__()
        self.title = title
        self.width = width
        self.height = height
        self.msg_elements = msg_elements
        # self.entry = {}
        self.yesAnswer = QPushButton()
        self.noAnswer = QPushButton()
        self.initUI()

    def initUI(self):
        """Initialize user interface"""
        self.setWindowTitle(self.title)
        self.resize(self.width, self.height)

        icon = self.msg_elements['icon'] if 'icon' in self.msg_elements else 'info'
        text = self.msg_elements['text'] if 'text' in self.msg_elements else None
        info_text = self.msg_elements['info_text'] if 'info_text' in self.msg_elements else None
        detail_text = self.msg_elements['detail_text'] if 'detail_text' in self.msg_elements else None
        buttons = self.msg_elements['buttons'] if 'buttons' in self.msg_elements else 'ok'

        if icon == 'info':
            self.setIcon(QMessageBox.Information)
        elif icon == 'question':
            self.setIcon(QMessageBox.Question)
        elif icon == 'warning':
            self.setIcon(QMessageBox.Warning)
        elif icon == 'critical':
            self.setIcon(QMessageBox.Critical)

        if text is not None:
            self.setText(text)
        if info_text is not None:
            self.setInformativeText(info_text)
        if detail_text is not None:
            self.setDetailedText(detail_text)

        if 'ok_cancel' == buttons:
            self.yesAnswer = QPushButton('OK')
            self.addButton(self.yesAnswer, QMessageBox.YesRole)
            self.noAnswer = QPushButton("Cancel")
            self.addButton(self.noAnswer, QMessageBox.NoRole)
        elif 'yes_no' == buttons:
            self.yesAnswer = QPushButton('Yes')
            self.addButton(self.yesAnswer, QMessageBox.YesRole)
            self.noAnswer = QPushButton("No")
            self.addButton(self.noAnswer, QMessageBox.NoRole)
        else:
            self.yesAnswer = QPushButton('OK')
            self.addButton(self.yesAnswer, QMessageBox.YesRole)

    def exec_(self):
        QMessageBox.exec_(self)
        '''ignore strange exec_ return value'''
        return self.clickedButton() == self.yesAnswer


class GUIThread(Thread):
    def __init__(self, title, *args, **kwargs):
        """Generic graphical user interface thread

        Parameters
        ----------
        title : string
        args : list
        kwargs : dict
        """
        super().__init__()
        self.title = title
        self.accept = False
        self.icon_path = os.path.join(symspace_home, 'System', 'Icon', 'Icon_128x128.png')

    def start(self, wait=True):
        super().start()
        while self.is_alive() and wait:
            pass


class GUIFormThread(GUIThread):
    def __init__(self, title, form_elements, *args, **kwargs):
        """Graphical user interface thread for a form dialog

        Parameters
        ----------
        title : string
        form_elements : list
        args : list
        kwargs : dict

        Example
        -------

        >>> form_elements = [('Function name', 'line'),
                             ('Parameter name', 'line'),
                             ('Select directory', 'browse_folder', 'X:/'),
                             ('Combobox', 'combobox', ['a', 'b', 'c']),
                             ('Open in editor', 'checkbox', True)]
        >>> form = GUIFormThread('Form', form_elements)
        >>> form.start()
        >>> print(form.accept)

        """
        super().__init__(title, *args, **kwargs)
        names = [element[0] for element in form_elements]
        if len(names) != len(set(names)):
            raise Exception("Duplicate names in: " + str(names))
        self.form_elements = form_elements
        self.result = {}

    def run(self):
        app = QApplication([])
        app.setWindowIcon(QtGui.QIcon(self.icon_path))
        dialog = FormDialog(self.title, self.form_elements)

        if dialog.exec_():
            self.result = dialog.getReturnValues()
            self.accept = True


class GUIMessageThread(GUIThread):
    def __init__(self, title, msg_elements, *args, **kwargs):
        """Graphical user interface thread for a message dialog

        Parameters
        ----------
        title : string
        form_elements : dict
        args : list
        kwargs : dict

        Example
        -------

        >>> msg_elements = {'icon': 'warning', 'text': 'test', 'buttons': 'yes_no'}
        >>> msg = GUIMessageThread('Warning', msg_elements)
        >>> msg.start()
        >>> print(msg.accept)

        """

        super().__init__(title, *args, **kwargs)
        self.msg_elements = msg_elements

    def run(self):
        app = QApplication([])
        app.setWindowIcon(QtGui.QIcon(self.icon_path))
        dialog = InfoDialog(self.title, self.msg_elements)

        retval = dialog.exec_()

        self.accept = retval


if __name__ == "__main__":
    _form_title = 'Create Python Function'

    _form_elements = [('Function name', 'line'),
                      ('Parameter name', 'line'),
                      ('Select directory1', 'browse_folder', 'X:/'),
                      ('Select directory2', 'browse_folder', 'D:/'),
                      ('Select file1', 'browse_file'),
                      ('Select file2', 'browse_file', 'X:/test.txt',
                       'Python Files (*.py);;Text Files (*.txt);;Group Files (*.txt *.py);;Doc Files (*.rst *.tex)'),
                      ('Save file1', 'save_file'),
                      ('Save file2', 'save_file', 'X:/test.txt',
                       'Python Files (*.py);;Text Files (*.txt);;Group Files (*.txt *.py);;Doc Files (*.rst *.tex)'),
                      ('Combobox', 'combobox', ['a', 'b', 'c']),
                      ('Open in editor', 'checkbox', True),
                      ]
    # _form_elements = [('a{}'.format(i), 'line') for i in range(60)]  # scrollbar test
    _form = GUIFormThread(_form_title, _form_elements)
    _form.start()
    print(_form.accept)
    if _form.accept:
        print(_form.result)
    else:
        _msg_title = 'Warning'
        _msg_elements = {'icon': 'warning', 'text': 'test', 'buttons': 'yes_no'}
        _msg = GUIMessageThread(_msg_title, _msg_elements)
        _msg.start()
        print(_msg.accept)
