"""
Created on 29.11.2016

@author: wkoppelstaetter, mkitzberger
"""

from abc import abstractmethod
from symspace.jobinterface.portDescWrapper import SymJobDesc
import sys
import traceback
import os
import shutil
import pickle
from symspace.osutils.pathutils import symspace_home

try:
    import jep
    import numpy as np
    from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
    from at.lcm.bu21.magopt.structure import PortDesc
    from java.lang import StackTraceElement
    from java.io import File as jFile
    from jep import jarray
    from jobmanager import JobDesc
    from jobmanager import JobType

except Exception as e:
    print(e)


class ParamPythonFnc(object):
    """
    Base class for |sym|'s Python functions which are chained
    into the tree structure.
    """

    def __init__(self, structure):
        """
        Initialize.
        """
        self._structure = structure
        self.debug_mode = False

    #        self.jobIDs = []

    @classmethod
    def getFunctionID(cls):
        return __name__ + '_' + cls.__name__

    @classmethod
    def getDefaultDescription(cls):
        return cls.__doc__.splitlines()[0].strip()

    @abstractmethod
    def importVars(self):
        """Description of the import variables

        Returns a list of tuples, which defines the import variables.
        Each tuple consists of multiple entries in the following order:

        1. Variable Name : string

        2. Variable Description : string

        3. Variable Unit : string

        4. Optional : bool

            The boolean value 'Optional' defines if the variable has to be set (False) or can be left unset (True).

        5. Type / Dimension : string

            The string 'Type / Dimension' defines the type and the dimension / size of the variable:

                * 'Null / 1'

                * 'Bool / 1'

                * 'String / 1'

                * 'Numeric / [n_dim1 x n_dim2]'

                * 'Periodic / [n_dim1 x n_dim2]'

                * 'Array / [n_dim1 x n_dim2]'

                * 'Struct / n_fields'

            If ndim1, ndim2 or n_fields is unknown in advance use ? instead. E.g. '[1 x ?]'

        6. Struct : tuple

            The tuple 'Struct' defines the inner structure of the variable.

                * No entry or empty list '[]' for all types except of 'Struct'

                *  If 'Type' is 'Struct' all variables, which are known in advance to be contained need to be described.
                The description of sub-level variables is done in the same way as the top layer description.


        return [('ImportVar1', 'Description of import variable 1', 'Unit1', Optional1, 'Type1 / Dimension1', Struct 1),
                ('ImportVar2', 'Description of import variable 2', 'Unit2', Optional2, 'Type2 / Dimension2', Struct 2)]

        Returns
        -------
        list
            List of import variables

        """
        pass

    @abstractmethod
    def exportVars(self):
        pass

    @abstractmethod
    def getJobRequirement(self):
        pass

    @abstractmethod
    def getAdditionalModules(self):
        pass

    @abstractmethod
    def isDistributable(self):
        return False

    @abstractmethod
    def getFile(self):
        pass

    def getUniverse(self) -> str:
        return 'JAVA'

    def getDockerImage(self) -> str:
        raise NotImplementedError

    def getProjectPath(self) -> str:
        proj = self._structure.getRoot()
        moProjectPath = os.path.dirname(os.path.normpath(proj.getFile()))
        return moProjectPath

    def runJob(self, d, runargs) -> int:
        # overwritten in subclass
        return 1

    def run(self, runargs) -> int:

        infileName = 'inputParameter.p'
        outfileName = 'outputParameter.p'

        d = self.importValues()
        remoteCompute = runargs['RemoteCompute']
        runSim = runargs['RunSim']
        # optimize = runargs['Optimize']

        timeout = sys.maxsize
        if runargs.get('Timeout') is not None:
            timeout = runargs['Timeout']

        state = 0
        status = 0
        jstate = self._structure.getTmpDSData("State")
        if jstate is not None:
            state = int(jstate)

        if self.isDistributable() and not runSim:
            return 1

        if state == 0:
            if self.isDistributable():
                # function is distributable sumbmit via jobmanager

                proj = self._structure.getRoot()

                # Add runargs to the input dictionary
                # On remote machine python starter will pass this value to run Job
                d['__RunArgs'] = dict(PyJDataInterface.convert_J2Py(runargs))

                # create job
                if 'jep' in self.getJobRequirement().lower():
                    job = 'jobmanager.python.CLILocalJepJob'
                else:
                    job = 'jobmanager.python.CLILocalJob'

                if self.getUniverse().upper() == 'DOCKER':
                    jobID = SymJobDesc.createDockerJob(proj.getFile(), proj.getDsId(), self._structure.getPId(),
                                                       self._structure.getClassName(), self.getDockerImage(), 'java',
                                                       ['jobmanager.server.JobStarter', job, 'symstarter.py',
                                                        self._structure.getModuleName(), self._structure.getClassName(),
                                                        infileName, outfileName])
                else:
                    jobID = SymJobDesc.createJavaJob(proj.getFile(), proj.getDsId(), self._structure.getPId(),
                                                     self._structure.getClassName(),
                                                     ['jobmanager.server.JobStarter', job, 'symstarter.py',
                                                      self._structure.getModuleName(), self._structure.getClassName(),
                                                      infileName, outfileName])

                jobPath = SymJobDesc.getPath(jobID)

                infilePath = os.path.normpath(os.path.join(jobPath, infileName))
                infile = open(infilePath, 'wb')
                pickle.dump(d, infile)
                infile.close()
                SymJobDesc.linkFile(jobID, infilePath)

                # archive python packages
                symspacePackage = os.path.join(symspace_home, 'System', 'Python', 'src', 'symspace')
                shutil.make_archive(os.path.join(jobPath, 'symspace_package'), 'zip', symspacePackage)
                SymJobDesc.linkFile(jobID, os.path.join(jobPath, 'symspace_package.zip'))

                lcmPackage = os.path.join(symspace_home, 'System', 'General', 'Python', 'src', 'lcm')
                shutil.make_archive(os.path.join(jobPath, 'lcm_package'), 'zip', lcmPackage)
                SymJobDesc.linkFile(jobID, os.path.join(jobPath, 'lcm_package.zip'))

                scriptPath = os.path.join(self._structure.getRoot().getAllFilesDirectory().getAbsolutePath(),
                                          '_Script')
                if os.path.exists(scriptPath):
                    shutil.make_archive(os.path.join(jobPath, 'packages'), 'zip', scriptPath)
                    SymJobDesc.linkFile(jobID, os.path.join(jobPath, 'packages.zip'))

                # add default files
                SymJobDesc.addFile(jobID, os.path.join(symspace_home, 'Tools', 'JobManager', 'Java', 'jobmanager.jar'))
                SymJobDesc.linkFile(jobID, os.path.join(symspace_home, 'System', 'Java', 'magopt.jar'))
                SymJobDesc.linkFile(jobID, os.path.join(symspace_home, 'System', 'General', 'Java', 'general.jar'))
                SymJobDesc.linkFile(jobID, os.path.join(symspace_home, 'System', 'General', 'External', 'commons',
                                                        'commons-math3-3.0.jar'))
                SymJobDesc.linkFile(jobID,
                                    os.path.join(symspace_home, 'External', 'javasysmon', 'javasysmon-0.3.3.jar'))

                # Add python starter
                SymJobDesc.addFile(jobID, os.path.join(symspace_home, 'System', 'Python', 'src', 'symspace',
                                                       'jobinterface', 'starter', 'symstarter.py'))

                # Add function file
                SymJobDesc.addFile(jobID, self.getFile())

                # Add dummy ParamPythonFnc
                SymJobDesc.addFile(jobID, os.path.join(symspace_home, 'System', 'Python', 'src', 'symspace',
                                                       'jobinterface', 'starter', 'dummy_parampythonfnc.py'))

                # Add additional files
                for file in self.getAdditionalModules():
                    if not os.path.exists(file):
                        self._structure.error('Additional file {} not found'.format(os.path.basename(file)))
                        return 1
                    else:
                        SymJobDesc.linkFile(jobID, file)

                # Set timeouts and requirements
                SymJobDesc.setTimeout(jobID, timeout)
                SymJobDesc.setRequirement(jobID, self.getJobRequirement())

                if remoteCompute:
                    SymJobDesc.submit(jobID, destination='cluster')
                else:
                    SymJobDesc.submit(jobID, destination='local')

                state = 1
                self._structure.info('Job sent to JobManager')
                self._structure.setTmpDSData("State", state)
                self._structure.setTmpDSData("JobID", jobID)
                status = 1

            else:
                # function is not distributable -> run without job manager
                status = self.runJob(d, runargs)

        elif state == 1:
            # process single job
            jobID = str(self._structure.getTmpDSData("JobID"))
            if SymJobDesc.isFinished(jobID):
                if SymJobDesc.getStatus(jobID) == 'COMPLETED' or SymJobDesc.getStatus(jobID) == 'ERROR':
                    status = 0
                    fileList = SymJobDesc.getFileList(jobID)
                    fileListBasenames = [os.path.basename(file) for file in fileList]

                    if 'symstarter_error.log' in fileListBasenames and status == 0:
                        '''Fail, if sym starter error log was created'''
                        index = fileListBasenames.index('symstarter_error.log')
                        stack = []
                        with open(fileList[index], 'r') as f:
                            _type = None
                            msg = None
                            for line in f:
                                line = line.strip()
                                if _type is None:
                                    _type = line
                                elif msg is None:
                                    msg = line
                                else:
                                    exc = line.split(',')
                                    stack.insert(0, StackTraceElement('', exc[0], exc[1], int(exc[2])))

                        self._structure.errorWithStack('Job: ' + msg + ': ' + exc[3], stack)
                        # self.error('Error file symstarter_error.log created by python starter')
                        status = 1

                    elif 'starter_error.log' in fileListBasenames and status == 0:
                        '''Fail, if java starter error log was created'''
                        # index = fileListBasenames.index('starter_error.log')
                        # with open(fileList[index], 'r') as f:
                        #     msg = f.read()
                        self.error('Error file starter_error.log created by java starter')
                        status = 1

                    if 'symstarter_debug.log' in fileListBasenames and status == 0:
                        '''Read runJob status'''
                        run_job_status = None
                        index = fileListBasenames.index('symstarter_debug.log')
                        with open(fileList[index], 'r') as f:
                            for line in f:
                                if 'status runJob:' in line:
                                    run_job_status = line.split('status runJob:')[-1].strip()
                                    break
                        if run_job_status != '0':
                            '''Fail, if job status is not 0'''
                            status = 1
                            self.error('runJob returned status: {}'.format(run_job_status))

                    if 'starter_debug.log' in fileListBasenames and status == 0:
                        '''Read cli status'''
                        cliStatus = None
                        index = fileListBasenames.index('starter_debug.log')
                        with open(fileList[index], 'r') as f:
                            for line in f:
                                line = line.strip()
                                if line.startswith('status'):
                                    cliStatus = line.split(':')[1].strip()
                                    break
                        if cliStatus != '0':
                            '''Fail, if cli status is not 0'''
                            status = 1
                            self.error('Local commandline interface returned with status: {}'.format(cliStatus))

                    elif 'starter_debug.log' not in fileListBasenames and status == 0:
                        '''Fail, if file is missing'''
                        self.error('Missing Java starter debug logfile')
                        status = 1

                    if outfileName not in fileListBasenames and status == 0:
                        '''Fail, if file is missing'''
                        self.error('Missing outfile of python subprocess')
                        status = 1

                    elif outfileName in fileListBasenames and status == 0:
                        '''Succeed and read results'''
                        index = fileListBasenames.index(outfileName)
                        with open(fileList[index], 'rb') as f:
                            d = pickle.load(f)

                elif SymJobDesc.getStatus(jobID) == 'ERROR':
                    print('Job error')
                    status = 1

                # remove job from jobmanager
                SymJobDesc.remove(jobID)
                # reset statemachine
                self._structure.rmTmpDSData("State")
                self._structure.rmTmpDSData("JobID")

            else:
                status = 2

        # ==============================================================================
        #         elif state == 2:
        #             # process multiple jobs
        #             print('Parsing multiple jobs')
        #             jobIDs = self._structure.getTmpDSData("JobID")
        #             finished = [False for item in jobIDs]
        #             status_arr = [1 for item in jobIDs]
        #             d_id = [None for item in jobIDs]
        #
        #             for j, jobID in enumerate(jobIDs):
        #                 if SymJobDesc.isFinished(jobID):
        #                     finished[j] = True
        #                     if SymJobDesc.getStatus(jobID) == 'COMPLETED' or SymJobDesc.getStatus(jobID) == 'ERROR':
        #                         status_arr[j] = 0
        #                         fileList = SymJobDesc.getFileList(jobID)
        #                         fileListBasenames = [os.path.basename(file) for file in fileList]
        #
        #                         # Check if sym starter error log was created
        #                         if ('symstarter_error.log') in fileListBasenames:
        #                             index = fileListBasenames.index('symstarter_error.log')
        #                             stack = []
        #                             with open(fileList[index], 'r') as f:
        #                                 for line in f:
        #                                     line = line.strip()
        #                                     exc = line.split(',')
        #                                     stack.insert(0, StackTraceElement('', exc[0], exc[1], int(exc[2])))
        #
        #                             self._structure.errorWithStack(exc[3], stack)
        #                             self.error('Error file symstarter_error.log created by python starter')
        #                             status_arr[j] = 1
        #
        #                         # Check if java starter error log was created
        #                         if ('starter_error.log') in fileListBasenames:
        #                             index = fileListBasenames.index('starter_error.log')
        #                             with open(fileList[index], 'r') as f:
        #                                 msg = f.read()
        #                             self.error('Error file starter_error.log created by java starter')
        #                             status_arr[j] = 1
        #
        #                         if ('starter_debug.log') in fileListBasenames:
        #                             # Read cli status
        #                             cliStatus = None
        #                             index = fileListBasenames.index('starter_debug.log')
        #                             with open(fileList[index], 'r') as f:
        #                                 for line in f:
        #                                     line = line.strip()
        #                                     if line.startswith('status'):
        #                                         cliStatus = line.split(':')[1].strip()
        #                             if cliStatus != '0':
        #                                 status_arr[j] = 1
        #                                 self.error('Local commandline interface returned with status != 0')
        #
        #                         else:
        #                             self.error('Missing Java starter debug logfile')
        #                             status_arr[j] = 1
        #
        #                         if outfileName in fileListBasenames:
        #                             index = fileListBasenames.index(outfileName)
        #                             with open(fileList[index], 'rb') as f:
        #                                 dic = pickle.load(f)
        #                                 dic.pop('__RunArgs')
        #                                 d_id[j] = dic
        #
        #                         else:
        #                             self.error('Missing outfile of python subprocess')
        #                             status_arr[j] = 1
        #
        #                     elif SymJobDesc.getStatus(jobID) == 'ERROR':
        #                         print('Job error')
        #                         status_arr[j] = 1
        #
        #                     if not False in finished:
        #                         # reset statemachine
        #                         print('reset statemachine')
        #                         state = 0
        #                         self._structure.setTmpDSData("State", state)
        #                         # remove jobs from jobmanager
        #                         for jobID in jobIDs:
        #                             SymJobDesc.remove(jobID)
        #
        #                 else:
        #                     status_arr[j] = 2
        #
        #             if 2 in status_arr:
        #                 status = 2
        #             elif 1 in status_arr:
        #                 status = 1
        #             else:
        #                 status = 0
        #                 d['c'] = 'Resulting values stored in variants'
        #                 d['variants'] = d_id
        #
        # ==============================================================================

        if status == 0:
            self.exportValues(d)

        return status

    def packageJob(self):
        pass

    def submitJob(self):
        pass

    def refreshDependency(self):
        self.setImportVars(self.importVars())
        self.setExportVars(self.exportVars())

    def generateDocFiles(self, path):
        pass

    def reportTex(self, template):
        return template

    def reset(self):
        # remove job from jobmanager
        jobID = self._structure.getTmpDSData("JobID")
        if jobID is not None:
            SymJobDesc.remove(str(jobID))

        # reset statemachine
        self._structure.rmTmpDSData("State")
        self._structure.rmTmpDSData("JobID")

    def getImports(self):
        imp = []
        depend = self._structure.getDepend().getImports()
        for dep in depend:
            imp.append((dep.getVarName(), dep.getParam(), dep.getHandle()))
        return imp

    def getExports(self):
        exp = []
        depend = self._structure.getDepend().getExports()
        for dep in depend:
            exp.append((dep.getVarName(), dep.getParam(), dep.getHandle()))
        return exp

    def getUpdateRequired(self):
        return self._structure.getUpdateRequired()

    def __getPortDesc(self, var):
        item = None
        length = len(var)
        if length == 1:
            item = PortDesc(var[0], var[0], "", False)
        elif length == 2:
            item = PortDesc(var[0], var[1], "", False)
        elif length == 3:
            item = PortDesc(var[0], var[1], var[2], False)
        elif length == 4:
            item = PortDesc(var[0], var[1], var[2], var[3])
        elif length == 5:
            item = PortDesc(var[0], var[1], var[2], var[3], var[4])
        elif length == 6:
            # print(type(var[5]))
            if type(var[5]) is list:
                subItem = []
                for sub in var[5]:
                    subItem.append(self.__getPortDesc(sub))
                item = PortDesc(var[0], var[1], var[2], var[3], var[4])
                item.setStructureList(subItem)
            else:
                subItem = self.__getPortDesc(var[5])
                item = PortDesc(var[0], var[1], var[2], var[3], var[4], subItem)
        else:
            self.warning("Undefined port-description")

        return item

    def setImportVars(self, varList):
        items = []
        for var in varList:
            item = self.__getPortDesc(var)
            if item is not None:
                items.append(item)

        jarr = jarray(len(items), PortDesc)
        for i in range(len(items)):
            jarr[i] = items[i]

        self._structure.getDepend().setImportVars(jarr)

    def setExportVars(self, varList):
        items = []
        for var in varList:
            item = self.__getPortDesc(var)
            if item is not None:
                items.append(item)

        jarr = jarray(len(items), PortDesc)
        for i in range(len(items)):
            jarr[i] = items[i]

        self._structure.getDepend().setExportVars(jarr)

    def setImportAssignments(self, **kwargs):
        self._structure.getDepend().setImportAssignmentsMap(kwargs)

    def setExportAssignments(self, **kwargs):
        self._structure.getDepend().setExportAssignmentsMap(kwargs)

    def setUpdateRequired(self, val):
        self._structure.setUpdateRequired(val)

    def info(self, message):
        stack = []
        for trace in traceback.extract_stack():
            stack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))
        self._structure.infoWithStack(message, stack)

    def warning(self, message):
        stack = []
        for trace in traceback.extract_stack():
            stack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))
        self._structure.warningWithStack(message, stack)

    def error(self, message):
        stack = []
        for trace in traceback.extract_stack():
            stack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))
        self._structure.errorWithStack(message, stack)

    def importValues(self):
        d = {}
        imp = self._structure.getDepend().getImports()

        for im in imp:
            name = im.getVarName()
            data = self._structure.getImportValue(name)

            if data is not None:
                d[name] = PyJDataInterface.convert_J2Py(data)

        return d

    def importRawValues(self):
        d = {}
        imp = self._structure.getDepend().getImports()

        for im in imp:
            name = im.getVarName()
            data = self._structure.getImportValue(name)

            if data is not None:
                d[name] = data

        return d

    def exportValues(self, d):
        exp = self._structure.getDepend().getExports()

        for ex in exp:
            name = ex.getVarName()
            if name in d:
                self._structure.setExportValue(name, PyJDataInterface.convert_Py2J(d[name]))
            else:
                raise Exception("Export value for '" + name + "' doesn't exist")

    def revertExportValues(self, d):
        exp = self._structure.getDepend().getExports()

        for ex in exp:
            name = ex.getVarName()
            data = self._structure.getExportValue(name)

            if data is not None:
                d[name] = PyJDataInterface.convert_J2Py(data)

    def revertExportRawValues(self, d):
        exp = self._structure.getDepend().getExports()

        for ex in exp:
            name = ex.getVarName()
            data = self._structure.getExportValue(name)

            if data is not None:
                d[name] = data

    def getImportUnit(self, name):
        return self._structure.getImportUnit(name)

    def getExportUnit(self, name):
        return self._structure.getExportUnit(name)

    def registerDocFile(self, name, path, desc):
        return self._structure.registerDocFile(name, path, desc)

    def getUniqueDocFileName(self, directory):
        return self._structure.getUniqueDocFileName(directory)

    def getDocFiles(self):
        docs = []
        docFiles = self._structure.getDocFiles()
        for doc in docFiles:
            docs.append((doc.getName(), doc.getDescription(), doc.getFile()))
        return docs

    def getDocFile(self, name):
        doc = self._structure.getDocFile(name)
        if doc is not None:
            return doc.getName(), doc.getDescription(), doc.getFile()
        else:
            return None

    def getTmpData(self, key):
        self._structure.getTmpData(key)

    def setTmpData(self, key, data):
        self._structure.setTmpData(key, data)

    def rmTmpData(self, key):
        self._structure.rmTmpData(key)

    def clearTmpData(self, key):
        self._structure.clearTmpData(key)

    def getTmpDSData(self, key):
        self._structure.getTmpDSData(key)

    def setTmpDSData(self, key, data):
        self._structure.setTmpDSData(key, data)

    def rmTmpDSData(self, key):
        self._structure.rmTmpDSData(key)

    def clearTmpDSData(self, key):
        self._structure.clearTmpDSData(key)

    def getParamDirectory(self):
        return self._structure.getParamDirectory().getPath()

    def hasPreview(self):
        return hasattr(self, 'preview') and callable(getattr(self, 'preview'))

    def hasShow(self):
        return hasattr(self, 'show') and callable(getattr(self, 'show'))

    def getModule(self):
        import inspect
        return inspect.getfile(self.__class__)
