#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9469 $
$LastChangedDate:: 2019-11-26 12:53:44 +0100#$

This file is part of SyMSpace.

Copyright (c) 2019, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import numpy as np
import pandas as pd
import lcm.utils.plot_fnc as pf


class ContainerComparison:

    @classmethod
    def fromSymContainer(cls, ref_fnc, cur_fnc) -> 'ContainerComparison':
        """
        Initialize a LoadpointComparison from SyMSpace functions

        Parameters
        ----------
        ref_fnc : SymContainer
        cur_fnc : SymContainer

        Returns
        -------
        ContainerComparison
        """

        ref_lp = ref_fnc.toDict()
        cur_lp = cur_fnc.toDict()
        new_cls = cls(ref_lp, cur_lp, ref_fnc.getName(), cur_fnc.getName())
        return new_cls

    def __init__(self, ref_lp: dict, cur_lp: dict, ref_name: str = 'Ref', cur_name: str = 'Cur'):
        """
        Loadpoint comparison with dictionaries

        Parameters
        ----------
        ref_lp : dict
            Reference loadpoint
        cur_lp : dict
            Other loadpoint
        ref_name : str
            Name of the reference loadpoint
        cur_name : str
            Name of the other loadpoint
        """
        self.ref_lp = ref_lp
        self.cur_lp = cur_lp
        self.ref_name = ref_name
        self.cur_name = cur_name

    @staticmethod
    def getUniformResults(lp: dict) -> dict:
        """
        Normalize the loadpoint data for comparison

        Parameters
        ----------
        lp : dict
            loadpoint results

        Returns
        -------
        new_dict : dict
            normalized loadpoint results
        """
        new_dict = dict()

        for key, value in lp.items():
            if isinstance(value, str):
                if value == '[]':
                    new_dict[key] = np.nan
                elif value == "'not evaluated'":
                    new_dict[key] = np.nan
                elif value == "Inf":
                    new_dict[key] = np.inf
                elif value == "-Inf":
                    new_dict[key] = -np.inf
                elif value == "NaN":
                    new_dict[key] = np.nan
                elif value == "true":
                    new_dict[key] = True
                elif value == "false":
                    new_dict[key] = False
                else:
                    try:
                        new_val = eval(value)
                        if isinstance(new_val, int) or isinstance(new_val, float) or isinstance(new_val, np.ndarray):
                            new_dict[key] = new_val
                        else:
                            new_dict[key] = np.nan
                    except:
                        new_dict[key] = np.nan
            else:
                new_dict[key] = value
        return new_dict

    def getDataFrame(self):
        """
        Get a pandas dataframe from loadpoint data

        Returns
        -------

        """
        ref_name = self.ref_name
        cur_name = self.cur_name

        if ref_name == cur_name:
            ref_name += '_Ref'

        ref_results = self.getUniformResults(self.ref_lp)
        cur_results = self.getUniformResults(self.cur_lp)

        d = {}
        ref_keys = set(list(ref_results.keys()))
        cur_keys = set((list(cur_results.keys())))
        keys = ref_keys.intersection(cur_keys)

        for key in keys:
            if key not in ['__error__', '__updaterequired__']:
                try:
                    ref_value = ref_results[key]
                    cur_value = cur_results[key]
                    error = np.abs(ref_value - cur_value) / np.abs(ref_value)

                    if 1e-8 < ref_value < 1e-4:
                        ref_value_str = '{:.2e}'.format(ref_value)
                        cur_value_str = '{:.2e}'.format(cur_value)
                    else:
                        ref_value_str = '{:.4f}'.format(ref_value)
                        cur_value_str = '{:.4f}'.format(cur_value)
                    error_str = '{:.2f}'.format(error * 100)
                    d[key] = [ref_value_str, cur_value_str, error_str]
                except:
                    d[key] = ['-', '-', '-']
        table = pd.DataFrame.from_dict(d, orient='index', columns=[ref_name, cur_name, 'Error %'])
        table = table.sort_index()
        return table

    def printTable(self):
        table = self.getDataFrame()
        print(table.to_string())

    def showHTMLTable(self):

        def colorValue(value):
            """
            Colors elements in a dateframe
            green if error is small and red if
            error is high. Does not color NaN
            values.
            """
            limit1 = 1.0
            limit2 = 10.0
            try:
                value = float(eval(value))
                if limit1 < value < limit2:
                    color = 'orange'
                elif value >= limit2:
                    color = 'red'
                else:
                    color = 'green'
            except:
                color = 'blue'

            return 'color: {}'.format(color)

        table = self.getDataFrame()
        print(table.index, table.index.is_unique)
        print(table.columns, table.columns.is_unique)
        try:
            html = (table.style
                    .applymap(colorValue, subset=['Error %'])
                    .set_table_attributes('border="1" class="dataframe table table-hover table-bordered"')
                    .render())
        except:
            html = table.to_html()

        pf.showHTML(html)
