/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9720 $
 * $Date: 2020-03-19 15:59:49 +0100 (Thu, 19 Mar 2020) $
 */
package viewer3d;

import java.io.FileNotFoundException;
import java.io.IOException;

import javax.media.j3d.BranchGroup;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

import com.sun.j3d.loaders.IncorrectFormatException;
import com.sun.j3d.loaders.ParsingErrorException;
import com.sun.j3d.loaders.Scene;
import com.sun.j3d.loaders.objectfile.ObjectFile;

/**
 * Transformation group with scale, transform and rotate
 * 
 * @author Werner Koppelsttter
 */
public class Transform {

	/**
	 * Scale object
	 */
	protected TransformGroup scale;

	/**
	 * Translation object
	 */
	protected TransformGroup translation;

	/**
	 * Rotation object
	 */
	protected TransformGroup rotate;

	/**
	 * Constructor of transformation group
	 */
	public Transform() {

		// Scale object
		this.scale = new TransformGroup();
		this.scale.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		this.scale.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		this.scale.setTransform(new Transform3D());

		// translate
		this.translation = new TransformGroup();
		this.translation.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		this.translation.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		this.translation.setTransform(new Transform3D());
		this.scale.addChild(this.translation);

		// rotate stator coordinate system
		this.rotate = new TransformGroup();
		this.rotate.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		this.rotate.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		this.rotate.setTransform(new Transform3D());
		this.translation.addChild(this.rotate);

		this.rotate.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
	}

	/**
	 * Get BranchGroup of this transformation group
	 * 
	 * @return branch group
	 */
	BranchGroup getBranchGroup() {
		BranchGroup group = new BranchGroup();
		group.addChild(scale);
		return group;
	}

	/**
	 * Add BranchGroup to this transformation group
	 * 
	 * @param trans
	 *            branch group
	 */
	void addBranchGroup(BranchGroup trans) {
		rotate.addChild(trans);
	}

	/**
	 * 
	 * Add model to scene-graph
	 * 
	 * @param filename
	 *            filename of model
	 */
	public void addModel(String filename) {

		ObjectFile f = new ObjectFile();
		Scene s = null;
		try {
			MaterialConverter.convertMaterial(filename);
			s = f.load(filename);
		} catch (FileNotFoundException e) {
			System.err.println(e);
			System.exit(1);
		} catch (IOException e) {
			System.err.println(e);
			System.exit(1);
		} catch (ParsingErrorException e) {
			System.err.println(e);
			System.exit(1);
		} catch (IncorrectFormatException e) {
			System.err.println(e);
			System.exit(1);
		}
		this.addBranchGroup(s.getSceneGroup());

	}

	/**
	 * Add Transform to scene-graph
	 * 
	 * @param trans
	 *            Transform object
	 */
	public void addTransform(Transform trans) {
		this.addBranchGroup(trans.getBranchGroup());
	}

	/**
	 * Scale model
	 * 
	 * @param scale
	 *            scale-factor
	 */
	public void scale(double scale) {
		Transform3D trans;

		trans = new Transform3D();
		trans.setScale(scale);
		this.scale.setTransform(trans);

	}

	/**
	 * Translate model
	 * 
	 * @param x
	 *            shift-value in x-axis
	 * @param y
	 *            shift-value in y-axis
	 * @param z
	 *            shift-value in z-axis
	 */
	public void translate(double x, double y, double z) {
		Transform3D trans;

		trans = new Transform3D();
		trans.setTranslation(new Vector3d(x, y, z));
		this.translation.setTransform(trans);

	}

	/**
	 * Rotate model
	 * 
	 * @param phiX
	 *            rotation-angle on x-axis (radiant)
	 * @param phiY
	 *            rotation-angle on y-axis (radiant)
	 * @param phiZ
	 *            rotation-angle on z-axis (radiant)
	 */
	public void rotate(double phiX, double phiY, double phiZ) {
		Transform3D trans;
		Matrix4d mat;

		trans = new Transform3D();

		mat = new Matrix4d(Math.cos(phiY) * Math.cos(phiZ), -Math.cos(phiY)
				* Math.sin(phiZ), Math.sin(phiY), 0.0, Math.sin(phiX)
				* Math.sin(phiY) * Math.cos(phiZ) + Math.cos(phiX)
				* Math.sin(phiZ), -Math.sin(phiX) * Math.sin(phiY)
				* Math.sin(phiZ) + Math.cos(phiX) * Math.cos(phiZ),
				-Math.sin(phiX) * Math.cos(phiY), 0.0, -Math.cos(phiX)
						* Math.sin(phiY) * Math.cos(phiZ) + Math.sin(phiX)
						* Math.sin(phiZ), Math.cos(phiX) * Math.sin(phiY)
						* Math.sin(phiZ) + Math.sin(phiX) * Math.cos(phiZ),
				Math.cos(phiX) * Math.cos(phiY), 0.0, 0.0, 0.0, 0.0, 1.0);

		trans.set(mat);
		this.rotate.setTransform(trans);
	}

}
